# Overview of the Framework 

## Overview of the code 

- `correlations` - Top-level directory of the code.  You can simply
  copy this directory (and sub-directories) to some other project or
  include the path to this in your header path. 
  
  In this directory, you will find the main interface classes, such as 
  
  - `correlations::QVector`
  - `correlations::Correlator` 
  - `correlations::Result` 
  
  Here, you will also find some generic interfaces for correlator
  [doc/Methods.md](methods), such as 
  
  - `correlations::NestedLoops` 
  - `correlations::RecursiveLoops` 
  - `correlations::Closed` 
  - `correlations::Recursive` 
  - `correlations::Recurrence` 
  
  and type definitions. 
  
  - `correlations::Real` 
  - `correlations::Size`
  - `correlations::Harmonic`
  - `correlations::Power`
  - `correlations::RealVector`
  - `correlations::Harmonics` 
  
  Sub-directories of this directory deal with specifics.  First, we
  have sub-directories for each [method](doc/Methods.md)
  
  - `correlations/Closed.hh` 
  
    The class `correlations::Closed` calculates the correlators using
    the fully expanded, closed-form expression for
    $`C_{\mathrm{h}}\{n\}`$ up to $`n=8`$ and $`C'_{\mathrm{h}}\{n\}`$
    up to $`n=6`$
	
  - `correlations/Recurrence.hh` 
  
    The class `correlations::Recurrence` calculates the both the
    integrated and differential correlators, $`C_{\mathrm{h}}\{n\}`$
    and $`C'_{\mathrm{h}}\{n\}`$ to any $`n`$.  The algorithm uses the
    recurrence relation of the correlators to express higher order
    correlators in terms of lower order correlators.
	
  - `correlations/Recursive.hh` 
  
     The class `correlations::Recursive` calculates the integrated and
     differential correlators  $`C_{\mathrm{h}}\{n\}`$
     and $`C'_{\mathrm{h}}\{n\}`$ to any order $`n`$ using
     $`Q`$-vectors.  The algorithm uses a recursive method to
     implicitly remove autocorrelations and the Bell sequence. 
	 
   - `correlations/RecursiveLoops.hh`
   
     The class `correlations::RecursiveLoops` uses recursive calls to
     calculate the integrated correlator $`C_{\mathrm{h}}\{n\}`$
     limited only by recursion depth and computing time. 
	
  - `correlations/symbolic` 
  
    Provides code to expand expression of the $`m`$-particle
    correlators into formats understood by various Computer-aided
    Algebra Systems such as 
     
    - [SymPy](https://sympy.org)
    - [Sage](http://www.sagemath.org)
    - [Maxima](http://maxima.sourceforge.net/)
    - [Maple](https://www.maplesoft.com)
	 
	For SymPy and Sage the code supports writing to a
    [Jupyter](https://jupyter.org) notebook. 
	 
  - `correlations/stat` 
  
    Statistics tools to calculate means, variances, and covariances on
    the fly.  This uses algorithms developed by Welford and West to
    calculate the means, variances, and covariance on-line (i.e., as
    the data becomes available) while minimizing rounding errors and
    without the need for additional storage.  This is particularly
    useful when adding large and small numbers and when looping over
    lots of data. 
	
  - `correlations/flow` 
  
    Reference implementations of using this framework for
    [flow](doc/GF.md) analyses.  The code in this directory provides
    simple examples of using the framework to estimate the integrated
    and differential flow, $`v_n\{m\}`$ and $`v'_n\{m\}`$,
    respectively, and can serve as a basis for implementing flow
    analyses elsewhere.
	
  - `correlations/progs` 
  
    Example programs 
	
    - `symbolic.cc` Expand expressions for correlators symbolically 
    - `benchmark.cc` Calculate correators using the different
      [methods](doc/Methods.md).
    - `writer.cc`  Write data files
	- `integrated.cc` Do integrated flow analysis
	- `differential.cc` Do differential flow analysis
	

## Main interface 

The main user components of this framework are 

 - correlations::Correlator 
 - correlations::Result 

The `correlations::Correlator` defines the interface for calculating
multi-particle correlations in any combinations of harmonics, while
`correlations::Result` represent the result of these calculations. 

## Calculating the correlators 

The member function `correlations::Correlator::calculate` takes a
list of harmonics and calculates the $`m`$-particle correlator,
where $`m`$ is the number of harmonics in the passed list. For
example,

	```
	correlations::Correlator c = ...;
	...
	correlations::HarmonicVector h = { 2, 3, 4, 5 };
	correlations::Result r = c.calculate(h);
	```
	
will calculate the 4-particle correlator (as there are 4 elements in
`h`) given by

```math
C_{2,3,4,5}\{4\}=
\left\langle\exp\left[i(2\phi_1+3\phi_2+4\phi_3+5\phi_4)\right]\right\rangle
```

where the mean represents the average over all combinations of 
$`\phi_1,\phi_2,\phi_3,\phi_4`$ in an event. 
 
This means, to calculate the standard 2-particle correlator
corresponding to the $`n`$ flow-harmonic, one would do 

	correlations::HarmonicVector h = { n, -n };
	correlations::Result r = c.calculate(h);

## Include in another project 

To use this code in another project located at say `/some/directory`
do 

    make install PREFIX=/some/directory 
	
from the project, and then make sure that `/some/directory` is in the
header search path of the compiler (e.g., `-I/some/directory`).  One
can also copy the sub-directory `correlations` to the project and use
directly from there (please observe the license conditions). 


## Choosing the method of calculation 

This framework, as outlined [elsewhere](doc/Methods.md), provides a
number of ways to "feed" the calculations of the correlators.  The
optimum solution depends on the problem at hand:  If only a few
correlators of low order is needed, then one can use the
`correlations::Closed` algorithm.  If one has a generally low event
multiplicity, then the `correlations::NestedLoops` algorithms may
suffice.  The $`Q`$-vector algoritms are useful if many correlators
are needed. 

## Example programs 

See [here](doc/Tests.md)


	

