#ifndef CORRELATIONS_CLOSED_FROMQVECTOR_HH
#define CORRELATIONS_CLOSED_FROMQVECTOR_HH
/**
 * @file   correlations/Closed.hh
 * @author Christian Holm Christensen <cholm@nbi.dk>
 * @date   Thu Oct 24 23:45:40 2013
 *
 * @brief  Cumulant using closed-form expressions
 */
/*
 * Multi-particle correlations
 * Copyright (C) 2013 K.Gulbrandsen, A.Bilandzic, C.H. Christensen.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses.
 */
#include <correlations/FromQVector.hh>

/** 
 * @defgroup correlations_closed Closed-form implementations 
 *
 * The code in this module corresponds to closed form (i.e., full
 * expanded) calculations of the 1 to 6-particle correlator
 * (optionally, 7 and 8-particle correlators may be enabled but these
 * can take a long time to compile - if indeed the compilation ever
 * finishes).  These are by far the fastest calculations but is
 * limited in the number of particles one may want to correlate.
 * 
 * @ingroup correlations 
 */
namespace correlations
{
  //____________________________________________________________________
  /**
   * Structure to calculate Cumulants of up to 8th order and power
   * from a given @f$ Q@f$-vector.
   *
   * This implementation used closed-form expression to evaluate up
   * to @f$ C\{8\}@f$.
   *
   * @tparam M Maximum size of correlator to compile.  This template
   * argument selects the largest possible correlator to compile in.
   * Correlators larger than this are specialised to throw a runtime
   * exception.
   *
   * @tparam Harmonics Vector of harmonics to use 
   *
   * @note This implementation is probably one of the fastest, but
   * it is not very flexible and it taked a very long time to
   * compile.  A better option would probably be to use pre-defined
   * functions with recurssion - at least up to some order
   *
   * @headerfile ""  <correlations/Closed.hh>
   * @ingroup correlations_closed
   */
  template <typename Harmonics=HarmonicVector, int M=5>
  struct ClosedM : public FromQVector<Harmonics>
  {
    /** Base type */
    using Base=FromQVector<Harmonics>;
    /** The type of vectors of harmonics to use */
    using HarmonicVector=typename Base::HarmonicVector;
    /** The harmonics to use */
    using HarmonicElement=typename Base::HarmonicElement;
    /** The type of Q-vector storage */
    using QStore=typename Base::QStore;
    /** The Q-vector to use */
    using QVector=typename Base::QVector;
    /** The complex type */
    using Complex=typename Base::Complex;

    /**
     * Constructor
     *
     * @param s Q-vector storage to use
     */
    ClosedM(const QStore& s) : Base(s) {}
    /** Move constructor */
    ClosedM(ClosedM&& o) : Base(std::move(o)) {}
    /** Deleted copy constructor */
    ClosedM(const ClosedM&) = delete;
    /** Deleted assignment operator */
    ClosedM& operator=(const ClosedM&) = delete;
    /**
     * @return Name of the correlator
     */
    const char* name() const { return "Closed form correlator"; }
    /** 
     * Maximum number of particles we can correlate 
     * 
     * @return 0 (indicating infinity)
     */
    Size maxSize() const { return M; }
  protected:
    using Base::_q;
    using Base::_r;
    using Base::_p;
    /**
     * Calculate the numerator or denominator of @f$ m@f$-particle
     * correlation using harmonics @f$\mathbf{h}@f$.
     *
     * For integrated measurements, either of 
     *
     * @f[ N_{\mathbf{h}} \quad\mathrm{or}
     \quad{D_{\mathbf{h}}=N_{\mathbf{0}}\quad,@f]
     *
     * and for differential measurements 
     *
     * @f[ N'_{\mathbf{h}} \quad\mathrm{or}
     \quad{D'_{\mathbf{h}}=N'_{\mathbf{0}}\quad.@f]
     *
     * @param m @f$ m@f$ How many particles to correlate
     * @param h @f$\mathbf{h}@f$ - Harmonic of each term
     *
     * @return The numerator or denominator correlator and the
     * summed weights
     */
    Complex eval(const Size m, const HarmonicVector& h) const
    {
      switch (m) {
      case 1: return c1(h);  break;
      case 2: return c2(h);  break;
      case 3: return c3(h);  break;
      case 4: return c4(h);  break;
      case 5: return c5(h);  break;
      case 6: return c6(h);  break;
      case 7: return c7(h);  break;
      case 8: return c8(h);  break;
      }
	
      throw std::runtime_error("Number of particles too big");
      return Complex();
    }
    Complex c1(const HarmonicVector& h) const { return c1(h[0]); }
    Complex c2(const HarmonicVector& h) const { return c2(h[0],h[1]); }
    Complex c3(const HarmonicVector& h) const { return c3(h[0],h[1],h[2]); }
    Complex c4(const HarmonicVector& h) const { return c4(h[0],h[1],h[2],
							  h[3]); }
    Complex c5(const HarmonicVector& h) const { return c5(h[0],h[1],h[2],
							  h[3],h[4]); }
    Complex c6(const HarmonicVector& h) const { return c6(h[0],h[1],h[2],
							  h[3],h[4],h[5]); }
    Complex c7(const HarmonicVector& h) const { return c7(h[0],h[1],h[2],
							  h[3],h[4],h[5],
							  h[6]); }
    Complex c8(const HarmonicVector& h) const { return c8(h[0],h[1],h[2],
							  h[3],h[4],h[5],
							  h[6],h[7]); }
    /**
     * Generic 1-particle correlation.
     *
     * @f[
     * QC\{1\} = \langle\exp[i(\sum_j^1 h_j\phi_j)]\rangle
     * @f]
     *
     * @param n1 Harmonic @f$ h_1@f$
     *
     * @return @f$ QC\{1\}@f$
     */
    Complex c1(const HarmonicElement n1) const
    {
      return _p(n1, 1);
    }
    /**
     * Do the 2-particle calculation.
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     *
     * @return the correlator
     */
    Complex c2(const HarmonicElement n1, const HarmonicElement n2) const
    {
      return _p(n1,1) * _r(n2,1) - _q(n1+n2,2);
    }
    /**
     * Do the 3-particle calculation.
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     * @param n3 3rd Harmonic
     *
     * @return the correlator
     */
    Complex c3(const HarmonicElement n1,
	       const HarmonicElement n2,
	       const HarmonicElement n3) const
    {
      const Real k2 = 2;
      return (_p(n1,1)           * _r(n2,1) * _r(n3,1)
	      -      _q(n1+n2,2) * _r(n3,1)
	      -      _q(n1+n3,2) * _r(n2,1)
	      -      _p(n1,1)    * _r(n2+n3,2)
	      + k2 * _q(n1+n2+n3,3));
    }
    /**
     * Do the 4-particle calculation.
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     * @param n3 3rd Harmonic
     * @param n4 4th Harmonic
     *
     * @return the correlator
     */
    Complex c4(const HarmonicElement n1,
	       const HarmonicElement n2,
	       const HarmonicElement n3,
	       const HarmonicElement n4) const
    {
      const Real k2 = 2;
      const Real k6 = 6;
      return (_p(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)
	      -      _q(n1+n2,2)    * _r(n3,1)       * _r(n4,1)
	      -      _r(n2,1)       * _q(n1+n3,2)    * _r(n4,1)
	      -      _p(n1,1)       * _r(n2+n3,2)    * _r(n4,1)
	      + k2 * _q(n1+n2+n3,3) * _r(n4,1)
	      -      _r(n2,1)       * _r(n3,1)       * _q(n1+n4,2)
	      +      _r(n2+n3,2)    * _q(n1+n4,2)
	      -      _p(n1,1)       * _r(n3,1)       * _r(n2+n4,2)
	      +      _q(n1+n3,2)    * _r(n2+n4,2)
	      + k2 * _r(n3,1)       * _q(n1+n2+n4,3)
	      -      _p(n1,1)       * _r(n2,1)       * _r(n3+n4,2)
	      +      _q(n1+n2,2)    * _r(n3+n4,2)
	      + k2 * _r(n2,1)       * _q(n1+n3+n4,3)
	      + k2 * _p(n1,1)       * _r(n2+n3+n4,3)
	      - k6 * _q(n1+n2+n3+n4,4));
    }
    /**
     * Do the 5-particle calculation.
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     * @param n3 3rd Harmonic
     * @param n4 4th Harmonic
     * @param n5 5th Harmonic
     *
     * @return The correlator
     */
    Complex c5(const HarmonicElement n1,
	       const HarmonicElement n2,
	       const HarmonicElement n3,
	       const HarmonicElement n4,
	       const HarmonicElement n5) const
    {
      const Real k2  = 2;
      const Real k6  = 6;
      const Real k24 = 24;
      return (_p(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)
	      -      _q(n1+n2,2)    * _r(n3,1)    * _r(n4,1)    * _r(n5,1)
	      -      _r(n2,1)       * _q(n1+n3,2) * _r(n4,1)    * _r(n5,1)
	      -      _p(n1,1)       * _r(n2+n3,2) * _r(n4,1)    * _r(n5,1)
	      + k2 * _q(n1+n2+n3,3) * _r(n4,1)    * _r(n5,1)
	      -      _r(n2,1)       * _r(n3,1)    * _q(n1+n4,2) * _r(n5,1)
	      +      _r(n2+n3,2)    * _q(n1+n4,2) * _r(n5,1)
	      -      _p(n1,1)       * _r(n3,1)    * _r(n2+n4,2) * _r(n5,1)
	      +      _q(n1+n3,2)    * _r(n2+n4,2) * _r(n5,1)
	      + k2 * _r(n3,1)       * _q(n1+n2+n4,3)* _r(n5,1)
	      -      _p(n1,1)       * _r(n2,1)* _r(n3+n4,2)     * _r(n5,1)
	      +      _q(n1+n2,2)    * _r(n3+n4,2) * _r(n5,1)
	      + k2 * _r(n2,1)       * _q(n1+n3+n4,3)* _r(n5,1)
	      + k2 * _p(n1,1)       * _r(n2+n3+n4,3)* _r(n5,1)
	      - k6 * _q(n1+n2+n3+n4,4)* _r(n5,1)
	      -      _r(n2,1)       * _r(n3,1)    * _r(n4,1)    * _q(n1+n5,2)
	      +      _r(n2+n3,2)    * _r(n4,1)    * _q(n1+n5,2)
	      +      _r(n3,1)       * _r(n2+n4,2) * _q(n1+n5,2)
	      +      _r(n2,1)       * _r(n3+n4,2) * _q(n1+n5,2)
	      - k2 * _r(n2+n3+n4,3) * _q(n1+n5,2)
	      -      _p(n1,1)       * _r(n3,1)    * _r(n4,1)    * _r(n2+n5,2)
	      +      _q(n1+n3,2)    * _r(n4,1)    * _r(n2+n5,2)
	      +      _r(n3,1)       * _q(n1+n4,2) * _r(n2+n5,2)
	      +      _p(n1,1)       * _r(n3+n4,2) * _r(n2+n5,2)
	      - k2 * _q(n1+n3+n4,3) * _r(n2+n5,2)
	      + k2 * _r(n3,1)       * _r(n4,1)    * _q(n1+n2+n5,3)
	      - k2 * _r(n3+n4,2)    * _q(n1+n2+n5,3)
	      -      _p(n1,1)       * _r(n2,1)    * _r(n4,1)    * _r(n3+n5,2)
	      +      _q(n1+n2,2)    * _r(n4,1)    * _r(n3+n5,2)
	      +      _r(n2,1)       * _q(n1+n4,2) * _r(n3+n5,2)
	      +      _p(n1,1)       * _r(n2+n4,2) * _r(n3+n5,2)
	      - k2 * _q(n1+n2+n4,3) * _r(n3+n5,2)
	      + k2 * _r(n2,1)       * _r(n4,1)    * _q(n1+n3+n5,3)
	      - k2 * _r(n2+n4,2)    * _q(n1+n3+n5,3)
	      + k2 * _p(n1,1)       * _r(n4,1)    * _r(n2+n3+n5,3)
	      - k2 * _q(n1+n4,2)    * _r(n2+n3+n5,3)
	      - k6 * _r(n4,1)       * _q(n1+n2+n3+n5,4)
	      -      _p(n1,1)       * _r(n2,1)    * _r(n3,1)    * _r(n4+n5,2)
	      +      _q(n1+n2,2)    * _r(n3,1)    * _r(n4+n5,2)
	      +      _r(n2,1)       * _q(n1+n3,2) * _r(n4+n5,2)
	      +      _p(n1,1)       * _r(n2+n3,2) * _r(n4+n5,2)
	      - k2 * _q(n1+n2+n3,3) * _r(n4+n5,2)
	      + k2 * _r(n2,1)       * _r(n3,1)    * _q(n1+n4+n5,3)
	      - k2 * _r(n2+n3,2)    * _q(n1+n4+n5,3)
	      + k2 * _p(n1,1)       * _r(n3,1)    * _r(n2+n4+n5,3)
	      - k2 * _q(n1+n3,2)    * _r(n2+n4+n5,3)
	      - k6 * _r(n3,1)       * _q(n1+n2+n4+n5,4)
	      + k2 * _p(n1,1)       * _r(n2,1)    * _r(n3+n4+n5,3)
	      - k2 * _q(n1+n2,2)    * _r(n3+n4+n5,3)
	      - k6 * _r(n2,1)       * _q(n1+n3+n4+n5,4)
	      - k6 * _p(n1,1)       * _r(n2+n3+n4+n5,4)
	      + k24* _q(n1+n2+n3+n4+n5,5));
    }
#ifdef __GNUC__
#pragma GCC optimize ("no-var-tracking")
#endif 
    /**
     * Do the 6-particle calculation.
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     * @param n3 3rd Harmonic
     * @param n4 4th Harmonic
     * @param n5 5th Harmonic
     * @param n6 6th Harmonic
     *
     * @return The correlator
     */
    Complex c6(const HarmonicElement n1,
	       const HarmonicElement n2,
	       const HarmonicElement n3,
	       const HarmonicElement n4,
	       const HarmonicElement n5,
	       const HarmonicElement n6) const
    {
      const Real k2   = 2;
      const Real k4   = 4;
      const Real k6   = 6;
      const Real k24  = 24;
      const Real k120 = 120;
      return
	(_p(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)
	 -       _q(n1+n2,2)    * _r(n3,1)    * _r(n4,1) * _r(n5,1) * _r(n6,1)
	 -       _r(n2,1)       * _q(n1+n3,2) * _r(n4,1) * _r(n5,1) * _r(n6,1)
	 -       _p(n1,1)       * _r(n2+n3,2) * _r(n4,1) * _r(n5,1) * _r(n6,1)
	 + k2  * _q(n1+n2+n3,3) * _r(n4,1)    * _r(n5,1)       * _r(n6,1)
	 -       _r(n2,1)       * _r(n3,1) * _q(n1+n4,2) * _r(n5,1) * _r(n6,1)
	 +       _r(n2+n3,2)    * _q(n1+n4,2) * _r(n5,1)       * _r(n6,1)
	 -       _p(n1,1)       * _r(n3,1) * _r(n2+n4,2) * _r(n5,1) * _r(n6,1)
	 +       _q(n1+n3,2)    * _r(n2+n4,2) * _r(n5,1)       * _r(n6,1)
	 + k2  * _r(n3,1)       * _q(n1+n2+n4,3) * _r(n5,1)       * _r(n6,1)
	 -       _p(n1,1)       * _r(n2,1) * _r(n3+n4,2) * _r(n5,1) * _r(n6,1)
	 +       _q(n1+n2,2)    * _r(n3+n4,2) * _r(n5,1)       * _r(n6,1)
	 + k2  * _r(n2,1)       * _q(n1+n3+n4,3) * _r(n5,1)    * _r(n6,1)
	 + k2  * _p(n1,1)       * _r(n2+n3+n4,3) * _r(n5,1)    * _r(n6,1)
	 - k6  * _q(n1+n2+n3+n4,4)            * _r(n5,1)       * _r(n6,1)
	 -       _r(n2,1)       * _r(n3,1) * _r(n4,1) * _q(n1+n5,2) * _r(n6,1)
	 +       _r(n2+n3,2)    * _r(n4,1)    * _q(n1+n5,2)    * _r(n6,1)
	 +       _r(n3,1)       * _r(n2+n4,2) * _q(n1+n5,2)    * _r(n6,1)
	 +       _r(n2,1)       * _r(n3+n4,2) * _q(n1+n5,2)    * _r(n6,1)
	 - k2  * _r(n2+n3+n4,3) * _q(n1+n5,2) * _r(n6,1)
	 -       _p(n1,1)       * _r(n3,1) * _r(n4,1) * _r(n2+n5,2) * _r(n6,1)
	 +       _q(n1+n3,2)    * _r(n4,1)    * _r(n2+n5,2)    * _r(n6,1)
	 +       _r(n3,1)       * _q(n1+n4,2) * _r(n2+n5,2)    * _r(n6,1)
	 +       _p(n1,1)       * _r(n3+n4,2) * _r(n2+n5,2)    * _r(n6,1)
	 - k2  * _q(n1+n3+n4,3) * _r(n2+n5,2) * _r(n6,1)
	 + k2  * _r(n3,1)       * _r(n4,1)    * _q(n1+n2+n5,3) * _r(n6,1)
	 - k2  * _r(n3+n4,2)    * _q(n1+n2+n5,3)               * _r(n6,1)
	 -       _p(n1,1)       * _r(n2,1) * _r(n4,1) * _r(n3+n5,2) * _r(n6,1)
	 +       _q(n1+n2,2)    * _r(n4,1)    * _r(n3+n5,2)    * _r(n6,1)
	 +       _r(n2,1)       * _q(n1+n4,2) * _r(n3+n5,2)    * _r(n6,1)
	 +       _p(n1,1)       * _r(n2+n4,2) * _r(n3+n5,2)    * _r(n6,1)
	 - k2  * _q(n1+n2+n4,3) * _r(n3+n5,2) * _r(n6,1)
	 + k2  * _r(n2,1)       * _r(n4,1)    * _q(n1+n3+n5,3) * _r(n6,1)
	 - k2  * _r(n2+n4,2)    * _q(n1+n3+n5,3)               * _r(n6,1)
	 + k2  * _p(n1,1)       * _r(n4,1)    * _r(n2+n3+n5,3) * _r(n6,1)
	 - k2  * _q(n1+n4,2)    * _r(n2+n3+n5,3)               * _r(n6,1)
	 - k6  * _r(n4,1)       * _q(n1+n2+n3+n5,4)            * _r(n6,1)
	 -       _p(n1,1)       * _r(n2,1) * _r(n3,1) * _r(n4+n5,2) * _r(n6,1)
	 +       _q(n1+n2,2)    * _r(n3,1)    * _r(n4+n5,2)    * _r(n6,1)
	 +       _r(n2,1)       * _q(n1+n3,2) * _r(n4+n5,2)    * _r(n6,1)
	 +       _p(n1,1)       * _r(n2+n3,2) * _r(n4+n5,2)    * _r(n6,1)
	 - k2  * _q(n1+n2+n3,3) * _r(n4+n5,2) * _r(n6,1)
	 + k2  * _r(n2,1)       * _r(n3,1)    * _q(n1+n4+n5,3) * _r(n6,1)
	 - k2  * _r(n2+n3,2)    * _q(n1+n4+n5,3)               * _r(n6,1)
	 + k2  * _p(n1,1)       * _r(n3,1)    * _r(n2+n4+n5,3) * _r(n6,1)
	 - k2  * _q(n1+n3,2)    * _r(n2+n4+n5,3) * _r(n6,1)
	 - k6  * _r(n3,1)       * _q(n1+n2+n4+n5,4) * _r(n6,1)
	 + k2  * _p(n1,1)       * _r(n2,1)    * _r(n3+n4+n5,3) * _r(n6,1)
	 - k2  * _q(n1+n2,2)    * _r(n3+n4+n5,3)               * _r(n6,1)
	 - k6  * _r(n2,1)       * _q(n1+n3+n4+n5,4)            * _r(n6,1)
	 - k6  * _p(n1,1)       * _r(n2+n3+n4+n5,4)            * _r(n6,1)
	 + k24 * _q(n1+n2+n3+n4+n5,5) * _r(n6,1)
	 -       _r(n2,1)       * _r(n3,1) * _r(n4,1) * _r(n5,1) * _q(n1+n6,2)
	 +       _r(n2+n3,2)    * _r(n4,1)    * _r(n5,1)       * _q(n1+n6,2)
	 +       _r(n3,1)       * _r(n2+n4,2) * _r(n5,1)       * _q(n1+n6,2)
	 +       _r(n2,1)       * _r(n3+n4,2) * _r(n5,1)       * _q(n1+n6,2)
	 - k2  * _r(n2+n3+n4,3) * _r(n5,1)    * _q(n1+n6,2)
	 +       _r(n3,1)       * _r(n4,1)    * _r(n2+n5,2)    * _q(n1+n6,2)
	 -       _r(n3+n4,2)    * _r(n2+n5,2) * _q(n1+n6,2)
	 +       _r(n2,1)       * _r(n4,1)    * _r(n3+n5,2)    * _q(n1+n6,2)
	 -       _r(n2+n4,2)    * _r(n3+n5,2) * _q(n1+n6,2)
	 - k2  * _r(n4,1)       * _r(n2+n3+n5,3)               * _q(n1+n6,2)
	 +       _r(n2,1)       * _r(n3,1)    * _r(n4+n5,2)    * _q(n1+n6,2)
	 -       _r(n2+n3,2)    * _r(n4+n5,2) * _q(n1+n6,2)
	 - k2  * _r(n3,1)       * _r(n2+n4+n5,3)               * _q(n1+n6,2)
	 - k2  * _r(n2,1)       * _r(n3+n4+n5,3)               * _q(n1+n6,2)
	 + k6  * _r(n2+n3+n4+n5,4) * _q(n1+n6,2)
	 -       _p(n1,1)       * _r(n3,1) * _r(n4,1) * _r(n5,1) * _r(n2+n6,2)
	 +       _q(n1+n3,2)    * _r(n4,1)    * _r(n5,1)       * _r(n2+n6,2)
	 +       _r(n3,1)       * _q(n1+n4,2) * _r(n5,1)       * _r(n2+n6,2)
	 +       _p(n1,1)       * _r(n3+n4,2) * _r(n5,1)       * _r(n2+n6,2)
	 - k2  * _q(n1+n3+n4,3) * _r(n5,1)    * _r(n2+n6,2)
	 +       _r(n3,1)       * _r(n4,1)    * _q(n1+n5,2)    * _r(n2+n6,2)
	 -       _r(n3+n4,2)    * _q(n1+n5,2) * _r(n2+n6,2)
	 +       _p(n1,1)       * _r(n4,1)    * _r(n3+n5,2)    * _r(n2+n6,2)
	 -       _q(n1+n4,2)    * _r(n3+n5,2) * _r(n2+n6,2)
	 - k2  * _r(n4,1)       * _q(n1+n3+n5,3)               * _r(n2+n6,2)
	 +       _p(n1,1)       * _r(n3,1)    * _r(n4+n5,2)    * _r(n2+n6,2)
	 -       _q(n1+n3,2)    * _r(n4+n5,2) * _r(n2+n6,2)
	 - k2  * _r(n3,1)       * _q(n1+n4+n5,3)               * _r(n2+n6,2)
	 - k2  * _p(n1,1)       * _r(n3+n4+n5,3)               * _r(n2+n6,2)
	 + k6  * _q(n1+n3+n4+n5,4)            * _r(n2+n6,2)
	 + k2  * _r(n3,1)       * _r(n4,1)    * _r(n5,1)     * _q(n1+n2+n6,3)
	 - k2  * _r(n3+n4,2)    * _r(n5,1)    * _q(n1+n2+n6,3)
	 - k2  * _r(n4,1)       * _r(n3+n5,2) * _q(n1+n2+n6,3)
	 - k2  * _r(n3,1)       * _r(n4+n5,2) * _q(n1+n2+n6,3)
	 + k4  * _r(n3+n4+n5,3) * _q(n1+n2+n6,3)
	 -       _p(n1,1)       * _r(n2,1) * _r(n4,1) * _r(n5,1) * _r(n3+n6,2)
	 +       _q(n1+n2,2)    * _r(n4,1)    * _r(n5,1)       * _r(n3+n6,2)
	 +       _r(n2,1)       * _q(n1+n4,2) * _r(n5,1)       * _r(n3+n6,2)
	 +       _p(n1,1)       * _r(n2+n4,2) * _r(n5,1)       * _r(n3+n6,2)
	 - k2  * _q(n1+n2+n4,3) * _r(n5,1)    * _r(n3+n6,2)
	 +       _r(n2,1)       * _r(n4,1)    * _q(n1+n5,2)    * _r(n3+n6,2)
	 -       _r(n2+n4,2)    * _q(n1+n5,2) * _r(n3+n6,2)
	 +       _p(n1,1)       * _r(n4,1)    * _r(n2+n5,2)    * _r(n3+n6,2)
	 -       _q(n1+n4,2)    * _r(n2+n5,2) * _r(n3+n6,2)
	 - k2  * _r(n4,1)       * _q(n1+n2+n5,3) * _r(n3+n6,2)
	 +       _p(n1,1)       * _r(n2,1)    * _r(n4+n5,2)    * _r(n3+n6,2)
	 -       _q(n1+n2,2)    * _r(n4+n5,2) * _r(n3+n6,2)
	 - k2  * _r(n2,1)       * _q(n1+n4+n5,3)               * _r(n3+n6,2)
	 - k2  * _p(n1,1)       * _r(n2+n4+n5,3)               * _r(n3+n6,2)
	 + k6  * _q(n1+n2+n4+n5,4) * _r(n3+n6,2)
	 + k2  * _r(n2,1)       * _r(n4,1)    * _r(n5,1)     * _q(n1+n3+n6,3)
	 - k2  * _r(n2+n4,2)    * _r(n5,1)    * _q(n1+n3+n6,3)
	 - k2  * _r(n4,1)       * _r(n2+n5,2) * _q(n1+n3+n6,3)
	 - k2  * _r(n2,1)       * _r(n4+n5,2) * _q(n1+n3+n6,3)
	 + k4  * _r(n2+n4+n5,3) * _q(n1+n3+n6,3)
	 + k2  * _p(n1,1)       * _r(n4,1)    * _r(n5,1)     * _r(n2+n3+n6,3)
	 - k2  * _q(n1+n4,2)    * _r(n5,1)    * _r(n2+n3+n6,3)
	 - k2  * _r(n4,1)       * _q(n1+n5,2) * _r(n2+n3+n6,3)
	 - k2  * _p(n1,1)       * _r(n4+n5,2) * _r(n2+n3+n6,3)
	 + k4  * _q(n1+n4+n5,3) * _r(n2+n3+n6,3)
	 - k6  * _r(n4,1)       * _r(n5,1)    * _q(n1+n2+n3+n6,4)
	 + k6  * _r(n4+n5,2)    * _q(n1+n2+n3+n6,4)
	 -       _p(n1,1)       * _r(n2,1) * _r(n3,1) * _r(n5,1) * _r(n4+n6,2)
	 +       _q(n1+n2,2)    * _r(n3,1)    * _r(n5,1)       * _r(n4+n6,2)
	 +       _r(n2,1)       * _q(n1+n3,2) * _r(n5,1)       * _r(n4+n6,2)
	 +       _p(n1,1)       * _r(n2+n3,2) * _r(n5,1)       * _r(n4+n6,2)
	 - k2  * _q(n1+n2+n3,3) * _r(n5,1)    * _r(n4+n6,2)
	 +       _r(n2,1)       * _r(n3,1)    * _q(n1+n5,2)    * _r(n4+n6,2)
	 -       _r(n2+n3,2)    * _q(n1+n5,2) * _r(n4+n6,2)
	 +       _p(n1,1)       * _r(n3,1)    * _r(n2+n5,2)    * _r(n4+n6,2)
	 -       _q(n1+n3,2)    * _r(n2+n5,2) * _r(n4+n6,2)
	 - k2  * _r(n3,1)       * _q(n1+n2+n5,3)               * _r(n4+n6,2)
	 +       _p(n1,1)       * _r(n2,1)    * _r(n3+n5,2)    * _r(n4+n6,2)
	 -       _q(n1+n2,2)    * _r(n3+n5,2) * _r(n4+n6,2)
	 - k2  * _r(n2,1)       * _q(n1+n3+n5,3)               * _r(n4+n6,2)
	 - k2  * _p(n1,1)       * _r(n2+n3+n5,3)               * _r(n4+n6,2)
	 + k6  * _q(n1+n2+n3+n5,4)                             * _r(n4+n6,2)
	 + k2  * _r(n2,1)       * _r(n3,1)    * _r(n5,1)     * _q(n1+n4+n6,3)
	 - k2  * _r(n2+n3,2)    * _r(n5,1)    * _q(n1+n4+n6,3)
	 - k2  * _r(n3,1)       * _r(n2+n5,2) * _q(n1+n4+n6,3)
	 - k2  * _r(n2,1)       * _r(n3+n5,2) * _q(n1+n4+n6,3)
	 + k4  * _r(n2+n3+n5,3) * _q(n1+n4+n6,3)
	 + k2  * _p(n1,1)       * _r(n3,1)    * _r(n5,1)     * _r(n2+n4+n6,3)
	 - k2  * _q(n1+n3,2)    * _r(n5,1)    * _r(n2+n4+n6,3)
	 - k2  * _r(n3,1)       * _q(n1+n5,2) * _r(n2+n4+n6,3)
	 - k2  * _p(n1,1)       * _r(n3+n5,2) * _r(n2+n4+n6,3)
	 + k4  * _q(n1+n3+n5,3) * _r(n2+n4+n6,3)
	 - k6  * _r(n3,1)       * _r(n5,1)    * _q(n1+n2+n4+n6,4)
	 + k6  * _r(n3+n5,2)    * _q(n1+n2+n4+n6,4)
	 + k2  * _p(n1,1)       * _r(n2,1)    * _r(n5,1)     * _r(n3+n4+n6,3)
	 - k2  * _q(n1+n2,2)    * _r(n5,1)    * _r(n3+n4+n6,3)
	 - k2  * _r(n2,1)       * _q(n1+n5,2) * _r(n3+n4+n6,3)
	 - k2  * _p(n1,1)       * _r(n2+n5,2) * _r(n3+n4+n6,3)
	 + k4  * _q(n1+n2+n5,3) * _r(n3+n4+n6,3)
	 - k6  * _r(n2,1)       * _r(n5,1)    * _q(n1+n3+n4+n6,4)
	 + k6  * _r(n2+n5,2)    * _q(n1+n3+n4+n6,4)
	 - k6  * _p(n1,1)       * _r(n5,1)    * _r(n2+n3+n4+n6,4)
	 + k6  * _q(n1+n5,2)    * _r(n2+n3+n4+n6,4)
	 -       _p(n1,1)       * _r(n2,1) * _r(n3,1) * _r(n4,1) * _r(n5+n6,2)
	 +       _q(n1+n2,2)    * _r(n3,1)    * _r(n4,1)       * _r(n5+n6,2)
	 +       _r(n2,1)       * _q(n1+n3,2) * _r(n4,1)       * _r(n5+n6,2)
	 +       _p(n1,1)       * _r(n2+n3,2) * _r(n4,1)       * _r(n5+n6,2)
	 - k2  * _q(n1+n2+n3,3) * _r(n4,1)    * _r(n5+n6,2)
	 +       _r(n2,1)       * _r(n3,1)    * _q(n1+n4,2)    * _r(n5+n6,2)
	 -       _r(n2+n3,2)    * _q(n1+n4,2) * _r(n5+n6,2)
	 +       _p(n1,1)       * _r(n3,1)    * _r(n2+n4,2)    * _r(n5+n6,2)
	 -       _q(n1+n3,2)    * _r(n2+n4,2) * _r(n5+n6,2)
	 - k2  * _r(n3,1)       * _q(n1+n2+n4,3)               * _r(n5+n6,2)
	 +       _p(n1,1)       * _r(n2,1)    * _r(n3+n4,2)    * _r(n5+n6,2)
	 -       _q(n1+n2,2)    * _r(n3+n4,2) * _r(n5+n6,2)
	 - k2  * _r(n2,1)       * _q(n1+n3+n4,3) * _r(n5+n6,2)
	 - k2  * _p(n1,1)       * _r(n2+n3+n4,3) * _r(n5+n6,2)
	 + k6  * _q(n1+n2+n3+n4,4)                             * _r(n5+n6,2)
	 + k2  * _r(n2,1)       * _r(n3,1)    * _r(n4,1)     * _q(n1+n5+n6,3)
	 - k2  * _r(n2+n3,2)    * _r(n4,1)    * _q(n1+n5+n6,3)
	 - k2  * _r(n3,1)       * _r(n2+n4,2) * _q(n1+n5+n6,3)
	 - k2  * _r(n2,1)       * _r(n3+n4,2) * _q(n1+n5+n6,3)
	 + k4  * _r(n2+n3+n4,3) * _q(n1+n5+n6,3)
	 + k2  * _p(n1,1)       * _r(n3,1)    * _r(n4,1)     * _r(n2+n5+n6,3)
	 - k2  * _q(n1+n3,2)    * _r(n4,1)    * _r(n2+n5+n6,3)
	 - k2  * _r(n3,1)       * _q(n1+n4,2) * _r(n2+n5+n6,3)
	 - k2  * _p(n1,1)       * _r(n3+n4,2) * _r(n2+n5+n6,3)
	 + k4  * _q(n1+n3+n4,3) * _r(n2+n5+n6,3)
	 - k6  * _r(n3,1)       * _r(n4,1)    * _q(n1+n2+n5+n6,4)
	 + k6  * _r(n3+n4,2)    * _q(n1+n2+n5+n6,4)
	 + k2  * _p(n1,1)       * _r(n2,1)    * _r(n4,1)      * _r(n3+n5+n6,3)
	 - k2  * _q(n1+n2,2)    * _r(n4,1)    * _r(n3+n5+n6,3)
	 - k2  * _r(n2,1)       * _q(n1+n4,2) * _r(n3+n5+n6,3)
	 - k2  * _p(n1,1)       * _r(n2+n4,2) * _r(n3+n5+n6,3)
	 + k4  * _q(n1+n2+n4,3) * _r(n3+n5+n6,3)
	 - k6  * _r(n2,1)       * _r(n4,1)    * _q(n1+n3+n5+n6,4)
	 + k6  * _r(n2+n4,2)    * _q(n1+n3+n5+n6,4)
	 - k6  * _p(n1,1)       * _r(n4,1)    * _r(n2+n3+n5+n6,4)
	 + k6  * _q(n1+n4,2)    * _r(n2+n3+n5+n6,4)
	 + k2  * _p(n1,1)       * _r(n2,1)    * _r(n3,1)      * _r(n4+n5+n6,3)
	 - k2  * _q(n1+n2,2)    * _r(n3,1)    * _r(n4+n5+n6,3)
	 - k2  * _r(n2,1)       * _q(n1+n3,2) * _r(n4+n5+n6,3)
	 - k2  * _p(n1,1)       * _r(n2+n3,2) * _r(n4+n5+n6,3)
	 + k4  * _q(n1+n2+n3,3) * _r(n4+n5+n6,3)
	 - k6  * _r(n2,1)       * _r(n3,1)    * _q(n1+n4+n5+n6,4)
	 + k6  * _r(n2+n3,2)    * _q(n1+n4+n5+n6,4)
	 - k6  * _p(n1,1)       * _r(n3,1)    * _r(n2+n4+n5+n6,4)
	 + k6  * _q(n1+n3,2)    * _r(n2+n4+n5+n6,4)
	 - k6  * _p(n1,1)       * _r(n2,1)    * _r(n3+n4+n5+n6,4)
	 + k6  * _q(n1+n2,2)    * _r(n3+n4+n5+n6,4)
	 + k24 * _r(n5,1)       * _q(n1+n2+n3+n4+n6,5)
	 + k24 * _r(n4,1)       * _q(n1+n2+n3+n5+n6,5)
	 + k24 * _r(n3,1)       * _q(n1+n2+n4+n5+n6,5)
	 + k24 * _r(n2,1)       * _q(n1+n3+n4+n5+n6,5)
	 + k24 * _p(n1,1)       * _r(n2+n3+n4+n5+n6,5)
	 - k120* _q(n1+n2+n3+n4+n5+n6,6));
    }
#ifdef __GNUC__
#pragma GCC optimize ("O0")
#endif 
    /**
     * Do the 7-particle calculation
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     * @param n3 3rd Harmonic
     * @param n4 4th Harmonic
     * @param n5 5th Harmonic
     * @param n6 6th Harmonic
     * @param n7 7th Harmonic
     *
     * @return The correlator
     */
    Complex c7(const HarmonicElement n1,
	       const HarmonicElement n2,
	       const HarmonicElement n3,
	       const HarmonicElement n4,
	       const HarmonicElement n5,
	       const HarmonicElement n6,
	       const HarmonicElement n7) const
    {
      const Real k2   = 2;
      const Real k4   = 4;
      const Real k6   = 6;
      const Real k12  = 12;
      const Real k24  = 24;
      const Real k120 = 120;
      const Real k720 = 720;
      return (_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      +       _r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      +       _r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      +       _r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      - k6*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n6,1)*_r(n7,1)
	      -       _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)
	      -       _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)
	      -       _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)
	      - k6*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n7,1)
	      -       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)
	      +       _r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)
	      - k6*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)
	      - k2*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)
	      - k6*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n7,1)
	      - k6*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n7,1)
	      + k24*_r(n1+n2+n3+n4+n5,5)*_r(n6,1)*_r(n7,1)
	      -       _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)
	      +       _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)
	      +       _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)
	      - k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)
	      +       _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)
	      -       _r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)
	      -       _r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)
	      - k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)
	      -       _r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n7,1)
	      + k6*_r(n2+n3+n4+n5,4)*_r(n1+n6,2)*_r(n7,1)
	      -       _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)
	      +       _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)
	      +       _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)
	      - k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)
	      +       _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)
	      -       _r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)
	      -       _r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)
	      - k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)
	      -       _r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n7,1)
	      + k6*_r(n1+n3+n4+n5,4)*_r(n2+n6,2)*_r(n7,1)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)
	      - k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)
	      - k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n7,1)
	      + k4*_r(n3+n4+n5,3)*_r(n1+n2+n6,3)*_r(n7,1)
	      -       _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)
	      +       _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)
	      - k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)
	      -       _r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)
	      -       _r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)
	      - k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)
	      -       _r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n7,1)
	      + k6*_r(n1+n2+n4+n5,4)*_r(n3+n6,2)*_r(n7,1)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)
	      - k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)
	      - k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n7,1)
	      + k4*_r(n2+n4+n5,3)*_r(n1+n3+n6,3)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)
	      - k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)
	      - k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n7,1)
	      + k4*_r(n1+n4+n5,3)*_r(n2+n3+n6,3)*_r(n7,1)
	      - k6*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n7,1)
	      + k6*_r(n4+n5,2)*_r(n1+n2+n3+n6,4)*_r(n7,1)
	      -       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)
	      +       _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)
	      - k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)
	      -       _r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)
	      -       _r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)
	      -       _r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n7,1)
	      + k6*_r(n1+n2+n3+n5,4)*_r(n4+n6,2)*_r(n7,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)
	      - k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n7,1)
	      + k4*_r(n2+n3+n5,3)*_r(n1+n4+n6,3)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)
	      - k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n7,1)
	      + k4*_r(n1+n3+n5,3)*_r(n2+n4+n6,3)*_r(n7,1)
	      - k6*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n7,1)
	      + k6*_r(n3+n5,2)*_r(n1+n2+n4+n6,4)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)
	      - k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n7,1)
	      + k4*_r(n1+n2+n5,3)*_r(n3+n4+n6,3)*_r(n7,1)
	      - k6*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n7,1)
	      + k6*_r(n2+n5,2)*_r(n1+n3+n4+n6,4)*_r(n7,1)
	      - k6*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n7,1)
	      + k6*_r(n1+n5,2)*_r(n2+n3+n4+n6,4)*_r(n7,1)
	      + k24*_r(n5,1)*_r(n1+n2+n3+n4+n6,5)*_r(n7,1)
	      -       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)
	      +       _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)
	      +       _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)
	      -       _r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)
	      -       _r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n7,1)
	      +       _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)
	      -       _r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n7,1)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n5+n6,2)*_r(n7,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n7,1)
	      + k4*_r(n2+n3+n4,3)*_r(n1+n5+n6,3)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n7,1)
	      + k4*_r(n1+n3+n4,3)*_r(n2+n5+n6,3)*_r(n7,1)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n7,1)
	      + k6*_r(n3+n4,2)*_r(n1+n2+n5+n6,4)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n7,1)
	      + k4*_r(n1+n2+n4,3)*_r(n3+n5+n6,3)*_r(n7,1)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n7,1)
	      + k6*_r(n2+n4,2)*_r(n1+n3+n5+n6,4)*_r(n7,1)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n7,1)
	      + k6*_r(n1+n4,2)*_r(n2+n3+n5+n6,4)*_r(n7,1)
	      + k24*_r(n4,1)*_r(n1+n2+n3+n5+n6,5)*_r(n7,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n7,1)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n7,1)
	      + k4*_r(n1+n2+n3,3)*_r(n4+n5+n6,3)*_r(n7,1)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n7,1)
	      + k6*_r(n2+n3,2)*_r(n1+n4+n5+n6,4)*_r(n7,1)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n7,1)
	      + k6*_r(n1+n3,2)*_r(n2+n4+n5+n6,4)*_r(n7,1)
	      + k24*_r(n3,1)*_r(n1+n2+n4+n5+n6,5)*_r(n7,1)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n7,1)
	      + k6*_r(n1+n2,2)*_r(n3+n4+n5+n6,4)*_r(n7,1)
	      + k24*_r(n2,1)*_r(n1+n3+n4+n5+n6,5)*_r(n7,1)
	      + k24*_r(n1,1)*_r(n2+n3+n4+n5+n6,5)*_r(n7,1)
	      - k120*_r(n1+n2+n3+n4+n5+n6,6)*_r(n7,1)
	      -       _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)
	      - k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)
	      -       _r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)
	      -       _r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)
	      - k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)
	      -       _r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)
	      - k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n7,2)
	      - k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n7,2)
	      + k6*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n1+n7,2)
	      +       _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)
	      -       _r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)
	      -       _r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n7,2)
	      -       _r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n7,2)
	      + k2*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n1+n7,2)
	      +       _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)
	      -       _r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)
	      -       _r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n7,2)
	      -       _r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n7,2)
	      + k2*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n1+n7,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n7,2)
	      + k2*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n1+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)
	      -       _r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)
	      -       _r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n7,2)
	      -       _r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n7,2)
	      + k2*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n1+n7,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n7,2)
	      + k2*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n1+n7,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n7,2)
	      + k2*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n1+n7,2)
	      + k6*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n1+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)
	      -       _r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)
	      -       _r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n7,2)
	      -       _r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n7,2)
	      + k2*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n1+n7,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n7,2)
	      + k2*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n1+n7,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n7,2)
	      + k2*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n1+n7,2)
	      + k6*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n1+n7,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n7,2)
	      + k2*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n1+n7,2)
	      + k6*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n1+n7,2)
	      + k6*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n1+n7,2)
	      - k24*_r(n2+n3+n4+n5+n6,5)*_r(n1+n7,2)
	      -       _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)
	      - k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)
	      -       _r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)
	      -       _r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)
	      - k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)
	      -       _r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)
	      - k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n7,2)
	      - k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n2+n7,2)
	      + k6*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n2+n7,2)
	      +       _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)
	      -       _r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)
	      -       _r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n7,2)
	      -       _r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n7,2)
	      + k2*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n2+n7,2)
	      +       _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)
	      -       _r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)
	      -       _r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n7,2)
	      -       _r(n1,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n2+n7,2)
	      + k2*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n2+n7,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n7,2)
	      + k2*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n2+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)
	      -       _r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)
	      -       _r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n7,2)
	      -       _r(n1,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n2+n7,2)
	      + k2*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n2+n7,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n7,2)
	      + k2*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n2+n7,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n2+n7,2)
	      + k2*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n2+n7,2)
	      + k6*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n2+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)
	      -       _r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)
	      -       _r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n7,2)
	      -       _r(n1,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n2+n7,2)
	      + k2*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n2+n7,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n7,2)
	      + k2*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n2+n7,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n2+n7,2)
	      + k2*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n2+n7,2)
	      + k6*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n2+n7,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n2+n7,2)
	      + k2*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n2+n7,2)
	      + k6*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n2+n7,2)
	      + k6*_r(n1,1)*_r(n3+n4+n5+n6,4)*_r(n2+n7,2)
	      - k24*_r(n1+n3+n4+n5+n6,5)*_r(n2+n7,2)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)
	      - k2*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)
	      - k2*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n7,3)
	      - k2*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n7,3)
	      + k4*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n2+n7,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n7,3)
	      + k2*_r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n2+n7,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n7,3)
	      + k2*_r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n2+n7,3)
	      + k4*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n2+n7,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n7,3)
	      + k2*_r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n2+n7,3)
	      + k4*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n2+n7,3)
	      + k4*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n2+n7,3)
	      - k12*_r(n3+n4+n5+n6,4)*_r(n1+n2+n7,3)
	      -       _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)
	      - k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)
	      -       _r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)
	      -       _r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)
	      - k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)
	      -       _r(n1+n2,2)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)
	      - k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n3+n7,2)
	      - k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n3+n7,2)
	      + k6*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n3+n7,2)
	      +       _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)
	      -       _r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)
	      -       _r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n7,2)
	      -       _r(n2,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n3+n7,2)
	      + k2*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n3+n7,2)
	      +       _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)
	      -       _r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)
	      -       _r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n7,2)
	      -       _r(n1,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n3+n7,2)
	      + k2*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n3+n7,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n7,2)
	      + k2*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n3+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)
	      -       _r(n1+n2,2)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)
	      -       _r(n2,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n3+n7,2)
	      -       _r(n1,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n3+n7,2)
	      + k2*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n3+n7,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n3+n7,2)
	      + k2*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n3+n7,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n3+n7,2)
	      + k2*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n3+n7,2)
	      + k6*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n3+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)
	      -       _r(n1+n2,2)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)
	      -       _r(n2,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n3+n7,2)
	      -       _r(n1,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n3+n7,2)
	      + k2*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n3+n7,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n3+n7,2)
	      + k2*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n3+n7,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n3+n7,2)
	      + k2*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n3+n7,2)
	      + k6*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n3+n7,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5+n6,3)*_r(n3+n7,2)
	      + k2*_r(n1+n2,2)*_r(n4+n5+n6,3)*_r(n3+n7,2)
	      + k6*_r(n2,1)*_r(n1+n4+n5+n6,4)*_r(n3+n7,2)
	      + k6*_r(n1,1)*_r(n2+n4+n5+n6,4)*_r(n3+n7,2)
	      - k24*_r(n1+n2+n4+n5+n6,5)*_r(n3+n7,2)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)
	      - k2*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)
	      - k2*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n7,3)
	      - k2*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n3+n7,3)
	      + k4*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n3+n7,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n7,3)
	      + k2*_r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n3+n7,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n3+n7,3)
	      + k2*_r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n3+n7,3)
	      + k4*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n3+n7,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n3+n7,3)
	      + k2*_r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n3+n7,3)
	      + k4*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n3+n7,3)
	      + k4*_r(n2,1)*_r(n4+n5+n6,3)*_r(n1+n3+n7,3)
	      - k12*_r(n2+n4+n5+n6,4)*_r(n1+n3+n7,3)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)
	      - k2*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)
	      - k2*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n7,3)
	      - k2*_r(n1,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n3+n7,3)
	      + k4*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n3+n7,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n7,3)
	      + k2*_r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n3+n7,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n3+n7,3)
	      + k2*_r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n3+n7,3)
	      + k4*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n3+n7,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n3+n7,3)
	      + k2*_r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n3+n7,3)
	      + k4*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n3+n7,3)
	      + k4*_r(n1,1)*_r(n4+n5+n6,3)*_r(n2+n3+n7,3)
	      - k12*_r(n1+n4+n5+n6,4)*_r(n2+n3+n7,3)
	      - k6*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n7,4)
	      + k6*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n3+n7,4)
	      + k6*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n3+n7,4)
	      + k6*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n3+n7,4)
	      - k12*_r(n4+n5+n6,3)*_r(n1+n2+n3+n7,4)
	      -       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)
	      - k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)
	      -       _r(n2+n3,2)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)
	      -       _r(n1+n3,2)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)
	      - k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)
	      -       _r(n1+n2,2)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)
	      - k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n4+n7,2)
	      - k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n4+n7,2)
	      + k6*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n4+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)
	      -       _r(n2+n3,2)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)
	      -       _r(n3,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n4+n7,2)
	      -       _r(n2,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n4+n7,2)
	      + k2*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n4+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)
	      -       _r(n1+n3,2)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)
	      -       _r(n3,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n4+n7,2)
	      -       _r(n1,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n4+n7,2)
	      + k2*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n4+n7,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n4+n7,2)
	      + k2*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n4+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)
	      -       _r(n1+n2,2)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)
	      -       _r(n2,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n4+n7,2)
	      -       _r(n1,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n4+n7,2)
	      + k2*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n4+n7,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n4+n7,2)
	      + k2*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n4+n7,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n4+n7,2)
	      + k2*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n4+n7,2)
	      + k6*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n4+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)
	      -       _r(n1+n2,2)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)
	      -       _r(n2,1)*_r(n1+n3,2)*_r(n5+n6,2)*_r(n4+n7,2)
	      -       _r(n1,1)*_r(n2+n3,2)*_r(n5+n6,2)*_r(n4+n7,2)
	      + k2*_r(n1+n2+n3,3)*_r(n5+n6,2)*_r(n4+n7,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5+n6,3)*_r(n4+n7,2)
	      + k2*_r(n2+n3,2)*_r(n1+n5+n6,3)*_r(n4+n7,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5+n6,3)*_r(n4+n7,2)
	      + k2*_r(n1+n3,2)*_r(n2+n5+n6,3)*_r(n4+n7,2)
	      + k6*_r(n3,1)*_r(n1+n2+n5+n6,4)*_r(n4+n7,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5+n6,3)*_r(n4+n7,2)
	      + k2*_r(n1+n2,2)*_r(n3+n5+n6,3)*_r(n4+n7,2)
	      + k6*_r(n2,1)*_r(n1+n3+n5+n6,4)*_r(n4+n7,2)
	      + k6*_r(n1,1)*_r(n2+n3+n5+n6,4)*_r(n4+n7,2)
	      - k24*_r(n1+n2+n3+n5+n6,5)*_r(n4+n7,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)
	      - k2*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)
	      - k2*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n4+n7,3)
	      - k2*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n4+n7,3)
	      + k4*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n4+n7,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n4+n7,3)
	      + k2*_r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n4+n7,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n4+n7,3)
	      + k2*_r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n4+n7,3)
	      + k4*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n4+n7,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n1+n4+n7,3)
	      + k2*_r(n2+n3,2)*_r(n5+n6,2)*_r(n1+n4+n7,3)
	      + k4*_r(n3,1)*_r(n2+n5+n6,3)*_r(n1+n4+n7,3)
	      + k4*_r(n2,1)*_r(n3+n5+n6,3)*_r(n1+n4+n7,3)
	      - k12*_r(n2+n3+n5+n6,4)*_r(n1+n4+n7,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)
	      - k2*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)
	      - k2*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n4+n7,3)
	      - k2*_r(n1,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n4+n7,3)
	      + k4*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n4+n7,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n4+n7,3)
	      + k2*_r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n4+n7,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n4+n7,3)
	      + k2*_r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n4+n7,3)
	      + k4*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n4+n7,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5+n6,2)*_r(n2+n4+n7,3)
	      + k2*_r(n1+n3,2)*_r(n5+n6,2)*_r(n2+n4+n7,3)
	      + k4*_r(n3,1)*_r(n1+n5+n6,3)*_r(n2+n4+n7,3)
	      + k4*_r(n1,1)*_r(n3+n5+n6,3)*_r(n2+n4+n7,3)
	      - k12*_r(n1+n3+n5+n6,4)*_r(n2+n4+n7,3)
	      - k6*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n4+n7,4)
	      + k6*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n4+n7,4)
	      + k6*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n4+n7,4)
	      + k6*_r(n3,1)*_r(n5+n6,2)*_r(n1+n2+n4+n7,4)
	      - k12*_r(n3+n5+n6,3)*_r(n1+n2+n4+n7,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)
	      - k2*_r(n1+n2,2)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)
	      - k2*_r(n2,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n4+n7,3)
	      - k2*_r(n1,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n4+n7,3)
	      + k4*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n4+n7,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n4+n7,3)
	      + k2*_r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n4+n7,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n4+n7,3)
	      + k2*_r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n4+n7,3)
	      + k4*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n4+n7,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5+n6,2)*_r(n3+n4+n7,3)
	      + k2*_r(n1+n2,2)*_r(n5+n6,2)*_r(n3+n4+n7,3)
	      + k4*_r(n2,1)*_r(n1+n5+n6,3)*_r(n3+n4+n7,3)
	      + k4*_r(n1,1)*_r(n2+n5+n6,3)*_r(n3+n4+n7,3)
	      - k12*_r(n1+n2+n5+n6,4)*_r(n3+n4+n7,3)
	      - k6*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n4+n7,4)
	      + k6*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n4+n7,4)
	      + k6*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n4+n7,4)
	      + k6*_r(n2,1)*_r(n5+n6,2)*_r(n1+n3+n4+n7,4)
	      - k12*_r(n2+n5+n6,3)*_r(n1+n3+n4+n7,4)
	      - k6*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n4+n7,4)
	      + k6*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n4+n7,4)
	      + k6*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n4+n7,4)
	      + k6*_r(n1,1)*_r(n5+n6,2)*_r(n2+n3+n4+n7,4)
	      - k12*_r(n1+n5+n6,3)*_r(n2+n3+n4+n7,4)
	      + k24*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n4+n7,5)
	      - k24*_r(n5+n6,2)*_r(n1+n2+n3+n4+n7,5)
	      -       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)
	      -       _r(n2+n3,2)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)
	      -       _r(n1+n3,2)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)
	      -       _r(n1+n2,2)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n6,1)*_r(n5+n7,2)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n6,1)*_r(n5+n7,2)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n6,1)*_r(n5+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)
	      -       _r(n2+n3,2)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)
	      -       _r(n3,1)*_r(n2+n4,2)*_r(n1+n6,2)*_r(n5+n7,2)
	      -       _r(n2,1)*_r(n3+n4,2)*_r(n1+n6,2)*_r(n5+n7,2)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n6,2)*_r(n5+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)
	      -       _r(n1+n3,2)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)
	      -       _r(n3,1)*_r(n1+n4,2)*_r(n2+n6,2)*_r(n5+n7,2)
	      -       _r(n1,1)*_r(n3+n4,2)*_r(n2+n6,2)*_r(n5+n7,2)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n6,2)*_r(n5+n7,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n6,3)*_r(n5+n7,2)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n6,3)*_r(n5+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)
	      -       _r(n1+n2,2)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)
	      -       _r(n2,1)*_r(n1+n4,2)*_r(n3+n6,2)*_r(n5+n7,2)
	      -       _r(n1,1)*_r(n2+n4,2)*_r(n3+n6,2)*_r(n5+n7,2)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n6,2)*_r(n5+n7,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n6,3)*_r(n5+n7,2)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n6,3)*_r(n5+n7,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n6,3)*_r(n5+n7,2)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n6,3)*_r(n5+n7,2)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n6,4)*_r(n5+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)
	      -       _r(n1+n2,2)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)
	      -       _r(n2,1)*_r(n1+n3,2)*_r(n4+n6,2)*_r(n5+n7,2)
	      -       _r(n1,1)*_r(n2+n3,2)*_r(n4+n6,2)*_r(n5+n7,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n6,2)*_r(n5+n7,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n6,3)*_r(n5+n7,2)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n6,3)*_r(n5+n7,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n6,3)*_r(n5+n7,2)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n6,3)*_r(n5+n7,2)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n6,4)*_r(n5+n7,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n6,3)*_r(n5+n7,2)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n6,3)*_r(n5+n7,2)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n6,4)*_r(n5+n7,2)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n6,4)*_r(n5+n7,2)
	      - k24*_r(n1+n2+n3+n4+n6,5)*_r(n5+n7,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n1+n5+n7,3)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n1+n5+n7,3)
	      + k4*_r(n2+n3+n4,3)*_r(n6,1)*_r(n1+n5+n7,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n1+n5+n7,3)
	      + k2*_r(n3+n4,2)*_r(n2+n6,2)*_r(n1+n5+n7,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n1+n5+n7,3)
	      + k2*_r(n2+n4,2)*_r(n3+n6,2)*_r(n1+n5+n7,3)
	      + k4*_r(n4,1)*_r(n2+n3+n6,3)*_r(n1+n5+n7,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n1+n5+n7,3)
	      + k2*_r(n2+n3,2)*_r(n4+n6,2)*_r(n1+n5+n7,3)
	      + k4*_r(n3,1)*_r(n2+n4+n6,3)*_r(n1+n5+n7,3)
	      + k4*_r(n2,1)*_r(n3+n4+n6,3)*_r(n1+n5+n7,3)
	      - k12*_r(n2+n3+n4+n6,4)*_r(n1+n5+n7,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n2+n5+n7,3)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n6,1)*_r(n2+n5+n7,3)
	      + k4*_r(n1+n3+n4,3)*_r(n6,1)*_r(n2+n5+n7,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n2+n5+n7,3)
	      + k2*_r(n3+n4,2)*_r(n1+n6,2)*_r(n2+n5+n7,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n6,2)*_r(n2+n5+n7,3)
	      + k2*_r(n1+n4,2)*_r(n3+n6,2)*_r(n2+n5+n7,3)
	      + k4*_r(n4,1)*_r(n1+n3+n6,3)*_r(n2+n5+n7,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n6,2)*_r(n2+n5+n7,3)
	      + k2*_r(n1+n3,2)*_r(n4+n6,2)*_r(n2+n5+n7,3)
	      + k4*_r(n3,1)*_r(n1+n4+n6,3)*_r(n2+n5+n7,3)
	      + k4*_r(n1,1)*_r(n3+n4+n6,3)*_r(n2+n5+n7,3)
	      - k12*_r(n1+n3+n4+n6,4)*_r(n2+n5+n7,3)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n2+n5+n7,4)
	      + k6*_r(n3+n4,2)*_r(n6,1)*_r(n1+n2+n5+n7,4)
	      + k6*_r(n4,1)*_r(n3+n6,2)*_r(n1+n2+n5+n7,4)
	      + k6*_r(n3,1)*_r(n4+n6,2)*_r(n1+n2+n5+n7,4)
	      - k12*_r(n3+n4+n6,3)*_r(n1+n2+n5+n7,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n6,1)*_r(n3+n5+n7,3)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n6,1)*_r(n3+n5+n7,3)
	      + k4*_r(n1+n2+n4,3)*_r(n6,1)*_r(n3+n5+n7,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n6,2)*_r(n3+n5+n7,3)
	      + k2*_r(n2+n4,2)*_r(n1+n6,2)*_r(n3+n5+n7,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n6,2)*_r(n3+n5+n7,3)
	      + k2*_r(n1+n4,2)*_r(n2+n6,2)*_r(n3+n5+n7,3)
	      + k4*_r(n4,1)*_r(n1+n2+n6,3)*_r(n3+n5+n7,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n6,2)*_r(n3+n5+n7,3)
	      + k2*_r(n1+n2,2)*_r(n4+n6,2)*_r(n3+n5+n7,3)
	      + k4*_r(n2,1)*_r(n1+n4+n6,3)*_r(n3+n5+n7,3)
	      + k4*_r(n1,1)*_r(n2+n4+n6,3)*_r(n3+n5+n7,3)
	      - k12*_r(n1+n2+n4+n6,4)*_r(n3+n5+n7,3)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n1+n3+n5+n7,4)
	      + k6*_r(n2+n4,2)*_r(n6,1)*_r(n1+n3+n5+n7,4)
	      + k6*_r(n4,1)*_r(n2+n6,2)*_r(n1+n3+n5+n7,4)
	      + k6*_r(n2,1)*_r(n4+n6,2)*_r(n1+n3+n5+n7,4)
	      - k12*_r(n2+n4+n6,3)*_r(n1+n3+n5+n7,4)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n2+n3+n5+n7,4)
	      + k6*_r(n1+n4,2)*_r(n6,1)*_r(n2+n3+n5+n7,4)
	      + k6*_r(n4,1)*_r(n1+n6,2)*_r(n2+n3+n5+n7,4)
	      + k6*_r(n1,1)*_r(n4+n6,2)*_r(n2+n3+n5+n7,4)
	      - k12*_r(n1+n4+n6,3)*_r(n2+n3+n5+n7,4)
	      + k24*_r(n4,1)*_r(n6,1)*_r(n1+n2+n3+n5+n7,5)
	      - k24*_r(n4+n6,2)*_r(n1+n2+n3+n5+n7,5)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n6,1)*_r(n4+n5+n7,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n6,1)*_r(n4+n5+n7,3)
	      + k4*_r(n1+n2+n3,3)*_r(n6,1)*_r(n4+n5+n7,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n6,2)*_r(n4+n5+n7,3)
	      + k2*_r(n2+n3,2)*_r(n1+n6,2)*_r(n4+n5+n7,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n6,2)*_r(n4+n5+n7,3)
	      + k2*_r(n1+n3,2)*_r(n2+n6,2)*_r(n4+n5+n7,3)
	      + k4*_r(n3,1)*_r(n1+n2+n6,3)*_r(n4+n5+n7,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n6,2)*_r(n4+n5+n7,3)
	      + k2*_r(n1+n2,2)*_r(n3+n6,2)*_r(n4+n5+n7,3)
	      + k4*_r(n2,1)*_r(n1+n3+n6,3)*_r(n4+n5+n7,3)
	      + k4*_r(n1,1)*_r(n2+n3+n6,3)*_r(n4+n5+n7,3)
	      - k12*_r(n1+n2+n3+n6,4)*_r(n4+n5+n7,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n1+n4+n5+n7,4)
	      + k6*_r(n2+n3,2)*_r(n6,1)*_r(n1+n4+n5+n7,4)
	      + k6*_r(n3,1)*_r(n2+n6,2)*_r(n1+n4+n5+n7,4)
	      + k6*_r(n2,1)*_r(n3+n6,2)*_r(n1+n4+n5+n7,4)
	      - k12*_r(n2+n3+n6,3)*_r(n1+n4+n5+n7,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n2+n4+n5+n7,4)
	      + k6*_r(n1+n3,2)*_r(n6,1)*_r(n2+n4+n5+n7,4)
	      + k6*_r(n3,1)*_r(n1+n6,2)*_r(n2+n4+n5+n7,4)
	      + k6*_r(n1,1)*_r(n3+n6,2)*_r(n2+n4+n5+n7,4)
	      - k12*_r(n1+n3+n6,3)*_r(n2+n4+n5+n7,4)
	      + k24*_r(n3,1)*_r(n6,1)*_r(n1+n2+n4+n5+n7,5)
	      - k24*_r(n3+n6,2)*_r(n1+n2+n4+n5+n7,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n3+n4+n5+n7,4)
	      + k6*_r(n1+n2,2)*_r(n6,1)*_r(n3+n4+n5+n7,4)
	      + k6*_r(n2,1)*_r(n1+n6,2)*_r(n3+n4+n5+n7,4)
	      + k6*_r(n1,1)*_r(n2+n6,2)*_r(n3+n4+n5+n7,4)
	      - k12*_r(n1+n2+n6,3)*_r(n3+n4+n5+n7,4)
	      + k24*_r(n2,1)*_r(n6,1)*_r(n1+n3+n4+n5+n7,5)
	      - k24*_r(n2+n6,2)*_r(n1+n3+n4+n5+n7,5)
	      + k24*_r(n1,1)*_r(n6,1)*_r(n2+n3+n4+n5+n7,5)
	      - k24*_r(n1+n6,2)*_r(n2+n3+n4+n5+n7,5)
	      - k120*_r(n6,1)*_r(n1+n2+n3+n4+n5+n7,6)
	      -       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)
	      -       _r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)
	      -       _r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)
	      -       _r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6+n7,2)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6+n7,2)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n6+n7,2)
	      +       _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)
	      -       _r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)
	      -       _r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6+n7,2)
	      -       _r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6+n7,2)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n6+n7,2)
	      +       _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)
	      -       _r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)
	      -       _r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6+n7,2)
	      -       _r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6+n7,2)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n6+n7,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6+n7,2)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n6+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)
	      -       _r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)
	      -       _r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6+n7,2)
	      -       _r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6+n7,2)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n6+n7,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6+n7,2)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n6+n7,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6+n7,2)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n6+n7,2)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n6+n7,2)
	      +       _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)
	      -       _r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6+n7,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6+n7,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n6+n7,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6+n7,2)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n6+n7,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6+n7,2)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n6+n7,2)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n6+n7,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6+n7,2)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n6+n7,2)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n6+n7,2)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n6+n7,2)
	      - k24*_r(n1+n2+n3+n4+n5,5)*_r(n6+n7,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6+n7,3)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6+n7,3)
	      + k4*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n6+n7,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6+n7,3)
	      + k2*_r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n6+n7,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6+n7,3)
	      + k2*_r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n6+n7,3)
	      + k4*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n6+n7,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6+n7,3)
	      + k2*_r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n6+n7,3)
	      + k4*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n6+n7,3)
	      + k4*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n6+n7,3)
	      - k12*_r(n2+n3+n4+n5,4)*_r(n1+n6+n7,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6+n7,3)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6+n7,3)
	      + k4*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n6+n7,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6+n7,3)
	      + k2*_r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n6+n7,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6+n7,3)
	      + k2*_r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n6+n7,3)
	      + k4*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n6+n7,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6+n7,3)
	      + k2*_r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n6+n7,3)
	      + k4*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n6+n7,3)
	      + k4*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n6+n7,3)
	      - k12*_r(n1+n3+n4+n5,4)*_r(n2+n6+n7,3)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6+n7,4)
	      + k6*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n6+n7,4)
	      + k6*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n6+n7,4)
	      + k6*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n6+n7,4)
	      - k12*_r(n3+n4+n5,3)*_r(n1+n2+n6+n7,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6+n7,3)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6+n7,3)
	      + k4*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n6+n7,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6+n7,3)
	      + k2*_r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n6+n7,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6+n7,3)
	      + k2*_r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n6+n7,3)
	      + k4*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n6+n7,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6+n7,3)
	      + k2*_r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n6+n7,3)
	      + k4*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n6+n7,3)
	      + k4*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n6+n7,3)
	      - k12*_r(n1+n2+n4+n5,4)*_r(n3+n6+n7,3)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6+n7,4)
	      + k6*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n6+n7,4)
	      + k6*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n6+n7,4)
	      + k6*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n6+n7,4)
	      - k12*_r(n2+n4+n5,3)*_r(n1+n3+n6+n7,4)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6+n7,4)
	      + k6*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n6+n7,4)
	      + k6*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n6+n7,4)
	      + k6*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n6+n7,4)
	      - k12*_r(n1+n4+n5,3)*_r(n2+n3+n6+n7,4)
	      + k24*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n6+n7,5)
	      - k24*_r(n4+n5,2)*_r(n1+n2+n3+n6+n7,5)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6+n7,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6+n7,3)
	      + k4*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n6+n7,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6+n7,3)
	      + k2*_r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n6+n7,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6+n7,3)
	      + k2*_r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n6+n7,3)
	      + k4*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n6+n7,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6+n7,3)
	      + k2*_r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n6+n7,3)
	      + k4*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n6+n7,3)
	      + k4*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n6+n7,3)
	      - k12*_r(n1+n2+n3+n5,4)*_r(n4+n6+n7,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6+n7,4)
	      + k6*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n6+n7,4)
	      + k6*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n6+n7,4)
	      + k6*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n6+n7,4)
	      - k12*_r(n2+n3+n5,3)*_r(n1+n4+n6+n7,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6+n7,4)
	      + k6*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n6+n7,4)
	      + k6*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n6+n7,4)
	      + k6*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n6+n7,4)
	      - k12*_r(n1+n3+n5,3)*_r(n2+n4+n6+n7,4)
	      + k24*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n6+n7,5)
	      - k24*_r(n3+n5,2)*_r(n1+n2+n4+n6+n7,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6+n7,4)
	      + k6*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n6+n7,4)
	      + k6*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n6+n7,4)
	      + k6*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n6+n7,4)
	      - k12*_r(n1+n2+n5,3)*_r(n3+n4+n6+n7,4)
	      + k24*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n6+n7,5)
	      - k24*_r(n2+n5,2)*_r(n1+n3+n4+n6+n7,5)
	      + k24*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n6+n7,5)
	      - k24*_r(n1+n5,2)*_r(n2+n3+n4+n6+n7,5)
	      - k120*_r(n5,1)*_r(n1+n2+n3+n4+n6+n7,6)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6+n7,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6+n7,3)
	      + k4*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n6+n7,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6+n7,3)
	      + k2*_r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n6+n7,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6+n7,3)
	      + k2*_r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n6+n7,3)
	      + k4*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n6+n7,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6+n7,3)
	      + k2*_r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n6+n7,3)
	      + k4*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n6+n7,3)
	      + k4*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n6+n7,3)
	      - k12*_r(n1+n2+n3+n4,4)*_r(n5+n6+n7,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6+n7,4)
	      + k6*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n6+n7,4)
	      + k6*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n6+n7,4)
	      + k6*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n6+n7,4)
	      - k12*_r(n2+n3+n4,3)*_r(n1+n5+n6+n7,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6+n7,4)
	      + k6*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n6+n7,4)
	      + k6*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n6+n7,4)
	      + k6*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n6+n7,4)
	      - k12*_r(n1+n3+n4,3)*_r(n2+n5+n6+n7,4)
	      + k24*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n6+n7,5)
	      - k24*_r(n3+n4,2)*_r(n1+n2+n5+n6+n7,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6+n7,4)
	      + k6*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n6+n7,4)
	      + k6*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n6+n7,4)
	      + k6*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n6+n7,4)
	      - k12*_r(n1+n2+n4,3)*_r(n3+n5+n6+n7,4)
	      + k24*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n6+n7,5)
	      - k24*_r(n2+n4,2)*_r(n1+n3+n5+n6+n7,5)
	      + k24*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n6+n7,5)
	      - k24*_r(n1+n4,2)*_r(n2+n3+n5+n6+n7,5)
	      - k120*_r(n4,1)*_r(n1+n2+n3+n5+n6+n7,6)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6+n7,4)
	      + k6*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n6+n7,4)
	      + k6*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n6+n7,4)
	      + k6*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n6+n7,4)
	      - k12*_r(n1+n2+n3,3)*_r(n4+n5+n6+n7,4)
	      + k24*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n6+n7,5)
	      - k24*_r(n2+n3,2)*_r(n1+n4+n5+n6+n7,5)
	      + k24*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n6+n7,5)
	      - k24*_r(n1+n3,2)*_r(n2+n4+n5+n6+n7,5)
	      - k120*_r(n3,1)*_r(n1+n2+n4+n5+n6+n7,6)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n6+n7,5)
	      - k24*_r(n1+n2,2)*_r(n3+n4+n5+n6+n7,5)
	      - k120*_r(n2,1)*_r(n1+n3+n4+n5+n6+n7,6)
	      - k120*_r(n1,1)*_r(n2+n3+n4+n5+n6+n7,6)
	      + k720*_r(n1+n2+n3+n4+n5+n6+n7,7));
    }
    /** 
     * First part of the 8-particle correlator 
     * 
     * @param n1 Harmonic
     * @param n2 Harmonic 
     * @param n3 Harmonic 
     * @param n4 Harmonic 
     * @param n5 Harmonic 
     * @param n6 Harmonic 
     * @param n7 Harmonic 
     * @param n8 Harmonic 
     * 
     * @return First part 
     */
    Complex c8P1(const HarmonicElement n1,
		 const HarmonicElement n2,
		 const HarmonicElement n3,
		 const HarmonicElement n4,
		 const HarmonicElement n5,
		 const HarmonicElement n6,
		 const HarmonicElement n7,
		 const HarmonicElement n8) const
    {
      const Real k2   = 2;
      const Real k4   = 4;
      const Real k6   = 6;
      const Real k12  = 12;
      const Real k24  = 24;
      // const Real k36  = 36;
      // const Real k48  = 48;
      const Real k120 = 120;
      const Real k720 = 720;
      return (_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      + k24*_r(n1+n2+n3+n4+n5,5)*_r(n6,1)*_r(n7,1)*_r(n8,1)
	      - _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n2+n3+n4+n5,4)*_r(n1+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n3+n4+n5,4)*_r(n2+n6,2)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n3+n4+n5,3)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n2+n4+n5,4)*_r(n3+n6,2)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n2+n4+n5,3)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n1+n4+n5,3)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n4+n5,2)*_r(n1+n2+n3+n6,4)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n2+n3+n5,4)*_r(n4+n6,2)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n2+n3+n5,3)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n1+n3+n5,3)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n3+n5,2)*_r(n1+n2+n4+n6,4)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n1+n2+n5,3)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n2+n5,2)*_r(n1+n3+n4+n6,4)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n5,2)*_r(n2+n3+n4+n6,4)*_r(n7,1)*_r(n8,1)
	      + k24*_r(n5,1)*_r(n1+n2+n3+n4+n6,5)*_r(n7,1)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n5+n6,2)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n2+n3+n4,3)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n1+n3+n4,3)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n3+n4,2)*_r(n1+n2+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n1+n2+n4,3)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n2+n4,2)*_r(n1+n3+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n4,2)*_r(n2+n3+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k24*_r(n4,1)*_r(n1+n2+n3+n5+n6,5)*_r(n7,1)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      + k4*_r(n1+n2+n3,3)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n2+n3,2)*_r(n1+n4+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n3,2)*_r(n2+n4+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k24*_r(n3,1)*_r(n1+n2+n4+n5+n6,5)*_r(n7,1)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k6*_r(n1+n2,2)*_r(n3+n4+n5+n6,4)*_r(n7,1)*_r(n8,1)
	      + k24*_r(n2,1)*_r(n1+n3+n4+n5+n6,5)*_r(n7,1)*_r(n8,1)
	      + k24*_r(n1,1)*_r(n2+n3+n4+n5+n6,5)*_r(n7,1)*_r(n8,1)
	      - k120*_r(n1+n2+n3+n4+n5+n6,6)*_r(n7,1)*_r(n8,1)
	      - _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + k6*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n1+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n1+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n1+n7,2)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n1+n7,2)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n1+n7,2)*_r(n8,1)
	      - k24*_r(n2+n3+n4+n5+n6,5)*_r(n1+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + k6*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n2+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n2+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n2+n7,2)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n2+n7,2)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n3+n4+n5+n6,4)*_r(n2+n7,2)*_r(n8,1)
	      - k24*_r(n1+n3+n4+n5+n6,5)*_r(n2+n7,2)*_r(n8,1)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k2*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k4*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k2*_r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k2*_r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k4*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n2+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n2+n7,3)*_r(n8,1)
	      - k12*_r(n3+n4+n5+n6,4)*_r(n1+n2+n7,3)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + k6*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n3+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n3+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n4+n5+n6,3)*_r(n3+n7,2)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n4+n5+n6,4)*_r(n3+n7,2)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n4+n5+n6,4)*_r(n3+n7,2)*_r(n8,1)
	      - k24*_r(n1+n2+n4+n5+n6,5)*_r(n3+n7,2)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k2*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k4*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k2*_r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k2*_r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k4*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n4+n5+n6,3)*_r(n1+n3+n7,3)*_r(n8,1)
	      - k12*_r(n2+n4+n5+n6,4)*_r(n1+n3+n7,3)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k2*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k4*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k2*_r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k2*_r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k4*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n3+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n4+n5+n6,3)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k12*_r(n1+n4+n5+n6,4)*_r(n2+n3+n7,3)*_r(n8,1)
	      - k6*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n7,4)*_r(n8,1)
	      + k6*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n3+n7,4)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n3+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n3+n7,4)*_r(n8,1)
	      - k12*_r(n4+n5+n6,3)*_r(n1+n2+n3+n7,4)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + k6*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n4+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n3,3)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n1+n5+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n2+n5+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n2+n5+n6,4)*_r(n4+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n3+n5+n6,3)*_r(n4+n7,2)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n3+n5+n6,4)*_r(n4+n7,2)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n3+n5+n6,4)*_r(n4+n7,2)*_r(n8,1)
	      - k24*_r(n1+n2+n3+n5+n6,5)*_r(n4+n7,2)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k2*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k4*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k2*_r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k2*_r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k4*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n5+n6,2)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n2+n5+n6,3)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n3+n5+n6,3)*_r(n1+n4+n7,3)*_r(n8,1)
	      - k12*_r(n2+n3+n5+n6,4)*_r(n1+n4+n7,3)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k4*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k2*_r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k2*_r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k4*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5+n6,2)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n5+n6,2)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n1+n5+n6,3)*_r(n2+n4+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n3+n5+n6,3)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k12*_r(n1+n3+n5+n6,4)*_r(n2+n4+n7,3)*_r(n8,1)
	      - k6*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n4+n7,4)*_r(n8,1)
	      + k6*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n4+n7,4)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n4+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n5+n6,2)*_r(n1+n2+n4+n7,4)*_r(n8,1)
	      - k12*_r(n3+n5+n6,3)*_r(n1+n2+n4+n7,4)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k4*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k2*_r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k2*_r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k4*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5+n6,2)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n5+n6,2)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n1+n5+n6,3)*_r(n3+n4+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n2+n5+n6,3)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k12*_r(n1+n2+n5+n6,4)*_r(n3+n4+n7,3)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n4+n7,4)*_r(n8,1)
	      + k6*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n4+n7,4)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n4+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n5+n6,2)*_r(n1+n3+n4+n7,4)*_r(n8,1)
	      - k12*_r(n2+n5+n6,3)*_r(n1+n3+n4+n7,4)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n4+n7,4)*_r(n8,1)
	      + k6*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n4+n7,4)*_r(n8,1)
	      + k6*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n4+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n5+n6,2)*_r(n2+n3+n4+n7,4)*_r(n8,1)
	      - k12*_r(n1+n5+n6,3)*_r(n2+n3+n4+n7,4)*_r(n8,1)
	      + k24*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n4+n7,5)*_r(n8,1)
	      - k24*_r(n5+n6,2)*_r(n1+n2+n3+n4+n7,5)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n6,1)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n6,4)*_r(n5+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n6,4)*_r(n5+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n6,3)*_r(n5+n7,2)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n6,4)*_r(n5+n7,2)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n6,4)*_r(n5+n7,2)*_r(n8,1)
	      - k24*_r(n1+n2+n3+n4+n6,5)*_r(n5+n7,2)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k4*_r(n2+n3+n4,3)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n2+n6,2)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n3+n6,2)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n2+n3+n6,3)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n4+n6,2)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n2+n4+n6,3)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n3+n4+n6,3)*_r(n1+n5+n7,3)*_r(n8,1)
	      - k12*_r(n2+n3+n4+n6,4)*_r(n1+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k4*_r(n1+n3+n4,3)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n1+n6,2)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n6,2)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n3+n6,2)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n1+n3+n6,3)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n6,2)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n4+n6,2)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n1+n4+n6,3)*_r(n2+n5+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n3+n4+n6,3)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k12*_r(n1+n3+n4+n6,4)*_r(n2+n5+n7,3)*_r(n8,1)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n2+n5+n7,4)*_r(n8,1)
	      + k6*_r(n3+n4,2)*_r(n6,1)*_r(n1+n2+n5+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n3+n6,2)*_r(n1+n2+n5+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n4+n6,2)*_r(n1+n2+n5+n7,4)*_r(n8,1)
	      - k12*_r(n3+n4+n6,3)*_r(n1+n2+n5+n7,4)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k4*_r(n1+n2+n4,3)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n6,2)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n1+n6,2)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n6,2)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n2+n6,2)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n1+n2+n6,3)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n6,2)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n4+n6,2)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n1+n4+n6,3)*_r(n3+n5+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n2+n4+n6,3)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k12*_r(n1+n2+n4+n6,4)*_r(n3+n5+n7,3)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n1+n3+n5+n7,4)*_r(n8,1)
	      + k6*_r(n2+n4,2)*_r(n6,1)*_r(n1+n3+n5+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n2+n6,2)*_r(n1+n3+n5+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n4+n6,2)*_r(n1+n3+n5+n7,4)*_r(n8,1)
	      - k12*_r(n2+n4+n6,3)*_r(n1+n3+n5+n7,4)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n2+n3+n5+n7,4)*_r(n8,1)
	      + k6*_r(n1+n4,2)*_r(n6,1)*_r(n2+n3+n5+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n1+n6,2)*_r(n2+n3+n5+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n4+n6,2)*_r(n2+n3+n5+n7,4)*_r(n8,1)
	      - k12*_r(n1+n4+n6,3)*_r(n2+n3+n5+n7,4)*_r(n8,1)
	      + k24*_r(n4,1)*_r(n6,1)*_r(n1+n2+n3+n5+n7,5)*_r(n8,1)
	      - k24*_r(n4+n6,2)*_r(n1+n2+n3+n5+n7,5)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k4*_r(n1+n2+n3,3)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n6,2)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n1+n6,2)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n6,2)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n2+n6,2)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n1+n2+n6,3)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n6,2)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n3+n6,2)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n1+n3+n6,3)*_r(n4+n5+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n2+n3+n6,3)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k12*_r(n1+n2+n3+n6,4)*_r(n4+n5+n7,3)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n1+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n2+n3,2)*_r(n6,1)*_r(n1+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n2+n6,2)*_r(n1+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n3+n6,2)*_r(n1+n4+n5+n7,4)*_r(n8,1)
	      - k12*_r(n2+n3+n6,3)*_r(n1+n4+n5+n7,4)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n2+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n1+n3,2)*_r(n6,1)*_r(n2+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n6,2)*_r(n2+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n3+n6,2)*_r(n2+n4+n5+n7,4)*_r(n8,1)
	      - k12*_r(n1+n3+n6,3)*_r(n2+n4+n5+n7,4)*_r(n8,1)
	      + k24*_r(n3,1)*_r(n6,1)*_r(n1+n2+n4+n5+n7,5)*_r(n8,1)
	      - k24*_r(n3+n6,2)*_r(n1+n2+n4+n5+n7,5)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n3+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n1+n2,2)*_r(n6,1)*_r(n3+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n6,2)*_r(n3+n4+n5+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n6,2)*_r(n3+n4+n5+n7,4)*_r(n8,1)
	      - k12*_r(n1+n2+n6,3)*_r(n3+n4+n5+n7,4)*_r(n8,1)
	      + k24*_r(n2,1)*_r(n6,1)*_r(n1+n3+n4+n5+n7,5)*_r(n8,1)
	      - k24*_r(n2+n6,2)*_r(n1+n3+n4+n5+n7,5)*_r(n8,1)
	      + k24*_r(n1,1)*_r(n6,1)*_r(n2+n3+n4+n5+n7,5)*_r(n8,1)
	      - k24*_r(n1+n6,2)*_r(n2+n3+n4+n5+n7,5)*_r(n8,1)
	      - k120*_r(n6,1)*_r(n1+n2+n3+n4+n5+n7,6)*_r(n8,1)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n6+n7,2)*_r(n8,1)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n6+n7,2)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n6+n7,2)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n6+n7,2)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n6+n7,2)*_r(n8,1)
	      - k24*_r(n1+n2+n3+n4+n5,5)*_r(n6+n7,2)*_r(n8,1)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k4*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n6+n7,3)*_r(n8,1)
	      - k12*_r(n2+n3+n4+n5,4)*_r(n1+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k2*_r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k12*_r(n1+n3+n4+n5,4)*_r(n2+n6+n7,3)*_r(n8,1)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6+n7,4)*_r(n8,1)
	      + k6*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n6+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n6+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n6+n7,4)*_r(n8,1)
	      - k12*_r(n3+n4+n5,3)*_r(n1+n2+n6+n7,4)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k2*_r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k4*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k12*_r(n1+n2+n4+n5,4)*_r(n3+n6+n7,3)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n6+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n6+n7,4)*_r(n8,1)
	      - k12*_r(n2+n4+n5,3)*_r(n1+n3+n6+n7,4)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n6+n7,4)*_r(n8,1)
	      + k6*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n6+n7,4)*_r(n8,1)
	      - k12*_r(n1+n4+n5,3)*_r(n2+n3+n6+n7,4)*_r(n8,1)
	      + k24*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n6+n7,5)*_r(n8,1)
	      - k24*_r(n4+n5,2)*_r(n1+n2+n3+n6+n7,5)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k12*_r(n1+n2+n3+n5,4)*_r(n4+n6+n7,3)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n6+n7,4)*_r(n8,1)
	      - k12*_r(n2+n3+n5,3)*_r(n1+n4+n6+n7,4)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n6+n7,4)*_r(n8,1)
	      - k12*_r(n1+n3+n5,3)*_r(n2+n4+n6+n7,4)*_r(n8,1)
	      + k24*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n6+n7,5)*_r(n8,1)
	      - k24*_r(n3+n5,2)*_r(n1+n2+n4+n6+n7,5)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n6+n7,4)*_r(n8,1)
	      - k12*_r(n1+n2+n5,3)*_r(n3+n4+n6+n7,4)*_r(n8,1)
	      + k24*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n6+n7,5)*_r(n8,1)
	      - k24*_r(n2+n5,2)*_r(n1+n3+n4+n6+n7,5)*_r(n8,1)
	      + k24*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n6+n7,5)*_r(n8,1)
	      - k24*_r(n1+n5,2)*_r(n2+n3+n4+n6+n7,5)*_r(n8,1)
	      - k120*_r(n5,1)*_r(n1+n2+n3+n4+n6+n7,6)*_r(n8,1)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k2*_r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k4*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k2*_r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k4*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n6+n7,3)*_r(n8,1)
	      + k4*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k12*_r(n1+n2+n3+n4,4)*_r(n5+n6+n7,3)*_r(n8,1)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n6+n7,4)*_r(n8,1)
	      - k12*_r(n2+n3+n4,3)*_r(n1+n5+n6+n7,4)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n6+n7,4)*_r(n8,1)
	      - k12*_r(n1+n3+n4,3)*_r(n2+n5+n6+n7,4)*_r(n8,1)
	      + k24*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n6+n7,5)*_r(n8,1)
	      - k24*_r(n3+n4,2)*_r(n1+n2+n5+n6+n7,5)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n6+n7,4)*_r(n8,1)
	      - k12*_r(n1+n2+n4,3)*_r(n3+n5+n6+n7,4)*_r(n8,1)
	      + k24*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n6+n7,5)*_r(n8,1)
	      - k24*_r(n2+n4,2)*_r(n1+n3+n5+n6+n7,5)*_r(n8,1)
	      + k24*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n6+n7,5)*_r(n8,1)
	      - k24*_r(n1+n4,2)*_r(n2+n3+n5+n6+n7,5)*_r(n8,1)
	      - k120*_r(n4,1)*_r(n1+n2+n3+n5+n6+n7,6)*_r(n8,1)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n6+n7,4)*_r(n8,1)
	      + k6*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n6+n7,4)*_r(n8,1)
	      - k12*_r(n1+n2+n3,3)*_r(n4+n5+n6+n7,4)*_r(n8,1)
	      + k24*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n6+n7,5)*_r(n8,1)
	      - k24*_r(n2+n3,2)*_r(n1+n4+n5+n6+n7,5)*_r(n8,1)
	      + k24*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n6+n7,5)*_r(n8,1)
	      - k24*_r(n1+n3,2)*_r(n2+n4+n5+n6+n7,5)*_r(n8,1)
	      - k120*_r(n3,1)*_r(n1+n2+n4+n5+n6+n7,6)*_r(n8,1)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n6+n7,5)*_r(n8,1)
	      - k24*_r(n1+n2,2)*_r(n3+n4+n5+n6+n7,5)*_r(n8,1)
	      - k120*_r(n2,1)*_r(n1+n3+n4+n5+n6+n7,6)*_r(n8,1)
	      - k120*_r(n1,1)*_r(n2+n3+n4+n5+n6+n7,6)*_r(n8,1)
	      + k720*_r(n1+n2+n3+n4+n5+n6+n7,7)*_r(n8,1)
	      - _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + k6*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k6*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k6*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n7,1)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k2*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n1+n8,2)
	      + k6*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n7,1)*_r(n1+n8,2)
	      + k6*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n7,1)*_r(n1+n8,2)
	      - k24*_r(n2+n3+n4+n5+n6,5)*_r(n7,1)*_r(n1+n8,2)
	      + _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n8,2)
	      - _r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n8,2)
	      - _r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n8,2)
	      + k2*_r(n3+n4+n5,3)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n8,2)
	      - _r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n1+n8,2)
	      + _r(n4+n5,2)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n1+n8,2)
	      + _r(n3+n5,2)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n1+n8,2)
	      + k2*_r(n5,1)*_r(n3+n4+n6,3)*_r(n2+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n1+n8,2)
	      + _r(n3+n4,2)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n3+n5+n6,3)*_r(n2+n7,2)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n4+n5+n6,3)*_r(n2+n7,2)*_r(n1+n8,2)
	      - k6*_r(n3+n4+n5+n6,4)*_r(n2+n7,2)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n8,2)
	      - _r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n8,2)
	      - _r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2+n4+n5,3)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n8,2)
	      - _r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n1+n8,2)
	      + _r(n4+n5,2)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n1+n8,2)
	      + _r(n2+n5,2)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n1+n8,2)
	      + k2*_r(n5,1)*_r(n2+n4+n6,3)*_r(n3+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n1+n8,2)
	      + _r(n2+n4,2)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n2+n5+n6,3)*_r(n3+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n4+n5+n6,3)*_r(n3+n7,2)*_r(n1+n8,2)
	      - k6*_r(n2+n4+n5+n6,4)*_r(n3+n7,2)*_r(n1+n8,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n1+n8,2)
	      + k2*_r(n4+n5,2)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n1+n8,2)
	      + k2*_r(n5,1)*_r(n4+n6,2)*_r(n2+n3+n7,3)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n5+n6,2)*_r(n2+n3+n7,3)*_r(n1+n8,2)
	      - k4*_r(n4+n5+n6,3)*_r(n2+n3+n7,3)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n8,2)
	      - _r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2+n3+n5,3)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n1+n8,2)
	      + _r(n3+n5,2)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n1+n8,2)
	      + _r(n2+n5,2)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n1+n8,2)
	      + k2*_r(n5,1)*_r(n2+n3+n6,3)*_r(n4+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n1+n8,2)
	      + _r(n2+n3,2)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n2+n5+n6,3)*_r(n4+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n3+n5+n6,3)*_r(n4+n7,2)*_r(n1+n8,2)
	      - k6*_r(n2+n3+n5+n6,4)*_r(n4+n7,2)*_r(n1+n8,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3+n5,2)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n1+n8,2)
	      + k2*_r(n5,1)*_r(n3+n6,2)*_r(n2+n4+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n5+n6,2)*_r(n2+n4+n7,3)*_r(n1+n8,2)
	      - k4*_r(n3+n5+n6,3)*_r(n2+n4+n7,3)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2+n5,2)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n1+n8,2)
	      + k2*_r(n5,1)*_r(n2+n6,2)*_r(n3+n4+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n5+n6,2)*_r(n3+n4+n7,3)*_r(n1+n8,2)
	      - k4*_r(n2+n5+n6,3)*_r(n3+n4+n7,3)*_r(n1+n8,2)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n2+n3+n4+n7,4)*_r(n1+n8,2)
	      - k6*_r(n5+n6,2)*_r(n2+n3+n4+n7,4)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n1+n8,2)
	      + _r(n3+n4,2)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n1+n8,2)
	      + _r(n2+n4,2)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n1+n8,2));
    }
    /** 
     * Second part of the 8-particle correlator 
     * 
     * @param n1 Harmonic
     * @param n2 Harmonic 
     * @param n3 Harmonic 
     * @param n4 Harmonic 
     * @param n5 Harmonic 
     * @param n6 Harmonic 
     * @param n7 Harmonic 
     * @param n8 Harmonic 
     * 
     * @return Second part 
     */
    Complex c8P2(const HarmonicElement n1,
		 const HarmonicElement n2,
		 const HarmonicElement n3,
		 const HarmonicElement n4,
		 const HarmonicElement n5,
		 const HarmonicElement n6,
		 const HarmonicElement n7,
		 const HarmonicElement n8) const
    {
      const Real k2   = 2;
      const Real k4   = 4;
      const Real k6   = 6;
      const Real k12  = 12;
      const Real k24  = 24;
      const Real k36  = 36;
      const Real k48  = 48;
      const Real k120 = 120;
      // const Real k720 = 720;
      return (k2*_r(n4,1)*_r(n2+n3+n6,3)*_r(n5+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n1+n8,2)
	      + _r(n2+n3,2)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4+n6,3)*_r(n5+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4+n6,3)*_r(n5+n7,2)*_r(n1+n8,2)
	      - k6*_r(n2+n3+n4+n6,4)*_r(n5+n7,2)*_r(n1+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3+n4,2)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n3+n6,2)*_r(n2+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n4+n6,2)*_r(n2+n5+n7,3)*_r(n1+n8,2)
	      - k4*_r(n3+n4+n6,3)*_r(n2+n5+n7,3)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2+n4,2)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n2+n6,2)*_r(n3+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n4+n6,2)*_r(n3+n5+n7,3)*_r(n1+n8,2)
	      - k4*_r(n2+n4+n6,3)*_r(n3+n5+n7,3)*_r(n1+n8,2)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n2+n3+n5+n7,4)*_r(n1+n8,2)
	      - k6*_r(n4+n6,2)*_r(n2+n3+n5+n7,4)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2+n3,2)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n2+n6,2)*_r(n4+n5+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n3+n6,2)*_r(n4+n5+n7,3)*_r(n1+n8,2)
	      - k4*_r(n2+n3+n6,3)*_r(n4+n5+n7,3)*_r(n1+n8,2)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n2+n4+n5+n7,4)*_r(n1+n8,2)
	      - k6*_r(n3+n6,2)*_r(n2+n4+n5+n7,4)*_r(n1+n8,2)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n3+n4+n5+n7,4)*_r(n1+n8,2)
	      - k6*_r(n2+n6,2)*_r(n3+n4+n5+n7,4)*_r(n1+n8,2)
	      - k24*_r(n6,1)*_r(n2+n3+n4+n5+n7,5)*_r(n1+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n1+n8,2)
	      + _r(n3+n4,2)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n1+n8,2)
	      + _r(n2+n4,2)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6+n7,2)*_r(n1+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n1+n8,2)
	      + _r(n2+n3,2)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6+n7,2)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6+n7,2)*_r(n1+n8,2)
	      - k6*_r(n2+n3+n4+n5,4)*_r(n6+n7,2)*_r(n1+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6+n7,3)*_r(n1+n8,2)
	      - k4*_r(n3+n4+n5,3)*_r(n2+n6+n7,3)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6+n7,3)*_r(n1+n8,2)
	      - k4*_r(n2+n4+n5,3)*_r(n3+n6+n7,3)*_r(n1+n8,2)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6+n7,4)*_r(n1+n8,2)
	      - k6*_r(n4+n5,2)*_r(n2+n3+n6+n7,4)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6+n7,3)*_r(n1+n8,2)
	      - k4*_r(n2+n3+n5,3)*_r(n4+n6+n7,3)*_r(n1+n8,2)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6+n7,4)*_r(n1+n8,2)
	      - k6*_r(n3+n5,2)*_r(n2+n4+n6+n7,4)*_r(n1+n8,2)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6+n7,4)*_r(n1+n8,2)
	      - k6*_r(n2+n5,2)*_r(n3+n4+n6+n7,4)*_r(n1+n8,2)
	      - k24*_r(n5,1)*_r(n2+n3+n4+n6+n7,5)*_r(n1+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6+n7,3)*_r(n1+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6+n7,3)*_r(n1+n8,2)
	      - k4*_r(n2+n3+n4,3)*_r(n5+n6+n7,3)*_r(n1+n8,2)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6+n7,4)*_r(n1+n8,2)
	      - k6*_r(n3+n4,2)*_r(n2+n5+n6+n7,4)*_r(n1+n8,2)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6+n7,4)*_r(n1+n8,2)
	      - k6*_r(n2+n4,2)*_r(n3+n5+n6+n7,4)*_r(n1+n8,2)
	      - k24*_r(n4,1)*_r(n2+n3+n5+n6+n7,5)*_r(n1+n8,2)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6+n7,4)*_r(n1+n8,2)
	      - k6*_r(n2+n3,2)*_r(n4+n5+n6+n7,4)*_r(n1+n8,2)
	      - k24*_r(n3,1)*_r(n2+n4+n5+n6+n7,5)*_r(n1+n8,2)
	      - k24*_r(n2,1)*_r(n3+n4+n5+n6+n7,5)*_r(n1+n8,2)
	      + k120*_r(n2+n3+n4+n5+n6+n7,6)*_r(n1+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + k6*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k6*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k6*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n7,1)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k2*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n2+n8,2)
	      + k6*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n7,1)*_r(n2+n8,2)
	      + k6*_r(n1,1)*_r(n3+n4+n5+n6,4)*_r(n7,1)*_r(n2+n8,2)
	      - k24*_r(n1+n3+n4+n5+n6,5)*_r(n7,1)*_r(n2+n8,2)
	      + _r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n8,2)
	      - _r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n8,2)
	      - _r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n8,2)
	      + k2*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n8,2)
	      - _r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n2+n8,2)
	      + _r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n2+n8,2)
	      + _r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n2+n8,2)
	      + k2*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n2+n8,2)
	      + _r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n7,2)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n7,2)*_r(n2+n8,2)
	      - k6*_r(n3+n4+n5+n6,4)*_r(n1+n7,2)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n8,2)
	      - _r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n8,2)
	      - _r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1+n4+n5,3)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n8,2)
	      - _r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n2+n8,2)
	      + _r(n4+n5,2)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n2+n8,2)
	      + _r(n1+n5,2)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n2+n8,2)
	      + k2*_r(n5,1)*_r(n1+n4+n6,3)*_r(n3+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n2+n8,2)
	      + _r(n1+n4,2)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n1+n5+n6,3)*_r(n3+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n4+n5+n6,3)*_r(n3+n7,2)*_r(n2+n8,2)
	      - k6*_r(n1+n4+n5+n6,4)*_r(n3+n7,2)*_r(n2+n8,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n2+n8,2)
	      + k2*_r(n4+n5,2)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n2+n8,2)
	      + k2*_r(n5,1)*_r(n4+n6,2)*_r(n1+n3+n7,3)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n5+n6,2)*_r(n1+n3+n7,3)*_r(n2+n8,2)
	      - k4*_r(n4+n5+n6,3)*_r(n1+n3+n7,3)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n8,2)
	      - _r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1+n3+n5,3)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n2+n8,2)
	      + _r(n3+n5,2)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n2+n8,2)
	      + _r(n1+n5,2)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n2+n8,2)
	      + k2*_r(n5,1)*_r(n1+n3+n6,3)*_r(n4+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n2+n8,2)
	      + _r(n1+n3,2)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n1+n5+n6,3)*_r(n4+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n3+n5+n6,3)*_r(n4+n7,2)*_r(n2+n8,2)
	      - k6*_r(n1+n3+n5+n6,4)*_r(n4+n7,2)*_r(n2+n8,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3+n5,2)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n2+n8,2)
	      + k2*_r(n5,1)*_r(n3+n6,2)*_r(n1+n4+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n5+n6,2)*_r(n1+n4+n7,3)*_r(n2+n8,2)
	      - k4*_r(n3+n5+n6,3)*_r(n1+n4+n7,3)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1+n5,2)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n2+n8,2)
	      + k2*_r(n5,1)*_r(n1+n6,2)*_r(n3+n4+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n5+n6,2)*_r(n3+n4+n7,3)*_r(n2+n8,2)
	      - k4*_r(n1+n5+n6,3)*_r(n3+n4+n7,3)*_r(n2+n8,2)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n1+n3+n4+n7,4)*_r(n2+n8,2)
	      - k6*_r(n5+n6,2)*_r(n1+n3+n4+n7,4)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n2+n8,2)
	      + _r(n3+n4,2)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n2+n8,2)
	      + _r(n1+n4,2)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n1+n3+n6,3)*_r(n5+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n2+n8,2)
	      + _r(n1+n3,2)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4+n6,3)*_r(n5+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4+n6,3)*_r(n5+n7,2)*_r(n2+n8,2)
	      - k6*_r(n1+n3+n4+n6,4)*_r(n5+n7,2)*_r(n2+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3+n4,2)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n3+n6,2)*_r(n1+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n4+n6,2)*_r(n1+n5+n7,3)*_r(n2+n8,2)
	      - k4*_r(n3+n4+n6,3)*_r(n1+n5+n7,3)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1+n4,2)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n1+n6,2)*_r(n3+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n4+n6,2)*_r(n3+n5+n7,3)*_r(n2+n8,2)
	      - k4*_r(n1+n4+n6,3)*_r(n3+n5+n7,3)*_r(n2+n8,2)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n1+n3+n5+n7,4)*_r(n2+n8,2)
	      - k6*_r(n4+n6,2)*_r(n1+n3+n5+n7,4)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1+n3,2)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n1+n6,2)*_r(n4+n5+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n3+n6,2)*_r(n4+n5+n7,3)*_r(n2+n8,2)
	      - k4*_r(n1+n3+n6,3)*_r(n4+n5+n7,3)*_r(n2+n8,2)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n1+n4+n5+n7,4)*_r(n2+n8,2)
	      - k6*_r(n3+n6,2)*_r(n1+n4+n5+n7,4)*_r(n2+n8,2)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n3+n4+n5+n7,4)*_r(n2+n8,2)
	      - k6*_r(n1+n6,2)*_r(n3+n4+n5+n7,4)*_r(n2+n8,2)
	      - k24*_r(n6,1)*_r(n1+n3+n4+n5+n7,5)*_r(n2+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n2+n8,2)
	      + _r(n3+n4,2)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n2+n8,2)
	      + _r(n1+n4,2)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6+n7,2)*_r(n2+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n2+n8,2)
	      + _r(n1+n3,2)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6+n7,2)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n6+n7,2)*_r(n2+n8,2)
	      - k6*_r(n1+n3+n4+n5,4)*_r(n6+n7,2)*_r(n2+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6+n7,3)*_r(n2+n8,2)
	      - k4*_r(n3+n4+n5,3)*_r(n1+n6+n7,3)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n3+n6+n7,3)*_r(n2+n8,2)
	      - k4*_r(n1+n4+n5,3)*_r(n3+n6+n7,3)*_r(n2+n8,2)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6+n7,4)*_r(n2+n8,2)
	      - k6*_r(n4+n5,2)*_r(n1+n3+n6+n7,4)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n4+n6+n7,3)*_r(n2+n8,2)
	      - k4*_r(n1+n3+n5,3)*_r(n4+n6+n7,3)*_r(n2+n8,2)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6+n7,4)*_r(n2+n8,2)
	      - k6*_r(n3+n5,2)*_r(n1+n4+n6+n7,4)*_r(n2+n8,2)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n3+n4+n6+n7,4)*_r(n2+n8,2)
	      - k6*_r(n1+n5,2)*_r(n3+n4+n6+n7,4)*_r(n2+n8,2)
	      - k24*_r(n5,1)*_r(n1+n3+n4+n6+n7,5)*_r(n2+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6+n7,3)*_r(n2+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n5+n6+n7,3)*_r(n2+n8,2)
	      - k4*_r(n1+n3+n4,3)*_r(n5+n6+n7,3)*_r(n2+n8,2)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6+n7,4)*_r(n2+n8,2)
	      - k6*_r(n3+n4,2)*_r(n1+n5+n6+n7,4)*_r(n2+n8,2)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n3+n5+n6+n7,4)*_r(n2+n8,2)
	      - k6*_r(n1+n4,2)*_r(n3+n5+n6+n7,4)*_r(n2+n8,2)
	      - k24*_r(n4,1)*_r(n1+n3+n5+n6+n7,5)*_r(n2+n8,2)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n4+n5+n6+n7,4)*_r(n2+n8,2)
	      - k6*_r(n1+n3,2)*_r(n4+n5+n6+n7,4)*_r(n2+n8,2)
	      - k24*_r(n3,1)*_r(n1+n4+n5+n6+n7,5)*_r(n2+n8,2)
	      - k24*_r(n1,1)*_r(n3+n4+n5+n6+n7,5)*_r(n2+n8,2)
	      + k120*_r(n1+n3+n4+n5+n6+n7,6)*_r(n2+n8,2)
	      + k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k4*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k2*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k2*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k4*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k2*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k4*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n1+n2+n8,3)
	      + k4*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k12*_r(n3+n4+n5+n6,4)*_r(n7,1)*_r(n1+n2+n8,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n1+n2+n8,3)
	      - k4*_r(n4+n5+n6,3)*_r(n3+n7,2)*_r(n1+n2+n8,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n1+n2+n8,3)
	      - k4*_r(n3+n5+n6,3)*_r(n4+n7,2)*_r(n1+n2+n8,3)
	      + k4*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n1+n2+n8,3)
	      - k4*_r(n5+n6,2)*_r(n3+n4+n7,3)*_r(n1+n2+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n1+n2+n8,3)
	      - k4*_r(n3+n4+n6,3)*_r(n5+n7,2)*_r(n1+n2+n8,3)
	      + k4*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n1+n2+n8,3)
	      - k4*_r(n4+n6,2)*_r(n3+n5+n7,3)*_r(n1+n2+n8,3)
	      + k4*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n1+n2+n8,3)
	      - k4*_r(n3+n6,2)*_r(n4+n5+n7,3)*_r(n1+n2+n8,3)
	      - k12*_r(n6,1)*_r(n3+n4+n5+n7,4)*_r(n1+n2+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n1+n2+n8,3)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n1+n2+n8,3)
	      - k4*_r(n3+n4+n5,3)*_r(n6+n7,2)*_r(n1+n2+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n1+n2+n8,3)
	      - k4*_r(n4+n5,2)*_r(n3+n6+n7,3)*_r(n1+n2+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n1+n2+n8,3)
	      - k4*_r(n3+n5,2)*_r(n4+n6+n7,3)*_r(n1+n2+n8,3)
	      - k12*_r(n5,1)*_r(n3+n4+n6+n7,4)*_r(n1+n2+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n1+n2+n8,3)
	      - k4*_r(n3+n4,2)*_r(n5+n6+n7,3)*_r(n1+n2+n8,3)
	      - k12*_r(n4,1)*_r(n3+n5+n6+n7,4)*_r(n1+n2+n8,3)
	      - k12*_r(n3,1)*_r(n4+n5+n6+n7,4)*_r(n1+n2+n8,3)
	      + k48*_r(n3+n4+n5+n6+n7,5)*_r(n1+n2+n8,3)
	      - _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1+n2,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + k6*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1+n2,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k6*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k6*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n7,1)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k2*_r(n1+n2,2)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n3+n8,2)
	      + k6*_r(n2,1)*_r(n1+n4+n5+n6,4)*_r(n7,1)*_r(n3+n8,2)
	      + k6*_r(n1,1)*_r(n2+n4+n5+n6,4)*_r(n7,1)*_r(n3+n8,2)
	      - k24*_r(n1+n2+n4+n5+n6,5)*_r(n7,1)*_r(n3+n8,2)
	      + _r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n8,2)
	      - _r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n8,2)
	      - _r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n8,2)
	      + k2*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n8,2)
	      - _r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n3+n8,2)
	      + _r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n3+n8,2)
	      + _r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n3+n8,2)
	      + k2*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n3+n8,2)
	      + _r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n7,2)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n4+n5+n6,3)*_r(n1+n7,2)*_r(n3+n8,2)
	      - k6*_r(n2+n4+n5+n6,4)*_r(n1+n7,2)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n8,2)
	      - _r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n8,2)
	      - _r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n8,2)
	      - _r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n3+n8,2)
	      + _r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n3+n8,2)
	      + _r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n3+n8,2)
	      + k2*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n3+n8,2)
	      + _r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n4+n5+n6,3)*_r(n2+n7,2)*_r(n3+n8,2)
	      - k6*_r(n1+n4+n5+n6,4)*_r(n2+n7,2)*_r(n3+n8,2)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n3+n8,2)
	      + k2*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n3+n8,2)
	      + k2*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n7,3)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n7,3)*_r(n3+n8,2)
	      - k4*_r(n4+n5+n6,3)*_r(n1+n2+n7,3)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n8,2)
	      - _r(n1+n2,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1+n2+n5,3)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n3+n8,2)
	      + _r(n2+n5,2)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n3+n8,2)
	      + _r(n1+n5,2)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n3+n8,2)
	      + k2*_r(n5,1)*_r(n1+n2+n6,3)*_r(n4+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n3+n8,2)
	      + _r(n1+n2,2)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n1+n5+n6,3)*_r(n4+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n2+n5+n6,3)*_r(n4+n7,2)*_r(n3+n8,2)
	      - k6*_r(n1+n2+n5+n6,4)*_r(n4+n7,2)*_r(n3+n8,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2+n5,2)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n3+n8,2)
	      + k2*_r(n5,1)*_r(n2+n6,2)*_r(n1+n4+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n5+n6,2)*_r(n1+n4+n7,3)*_r(n3+n8,2)
	      - k4*_r(n2+n5+n6,3)*_r(n1+n4+n7,3)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1+n5,2)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n3+n8,2)
	      + k2*_r(n5,1)*_r(n1+n6,2)*_r(n2+n4+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n5+n6,2)*_r(n2+n4+n7,3)*_r(n3+n8,2)
	      - k4*_r(n1+n5+n6,3)*_r(n2+n4+n7,3)*_r(n3+n8,2)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n1+n2+n4+n7,4)*_r(n3+n8,2)
	      - k6*_r(n5+n6,2)*_r(n1+n2+n4+n7,4)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n3+n8,2)
	      + _r(n2+n4,2)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n3+n8,2)
	      + _r(n1+n4,2)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n1+n2+n6,3)*_r(n5+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n3+n8,2)
	      + _r(n1+n2,2)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4+n6,3)*_r(n5+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4+n6,3)*_r(n5+n7,2)*_r(n3+n8,2)
	      - k6*_r(n1+n2+n4+n6,4)*_r(n5+n7,2)*_r(n3+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2+n4,2)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n2+n6,2)*_r(n1+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n4+n6,2)*_r(n1+n5+n7,3)*_r(n3+n8,2)
	      - k4*_r(n2+n4+n6,3)*_r(n1+n5+n7,3)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1+n4,2)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n1+n6,2)*_r(n2+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n4+n6,2)*_r(n2+n5+n7,3)*_r(n3+n8,2)
	      - k4*_r(n1+n4+n6,3)*_r(n2+n5+n7,3)*_r(n3+n8,2)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n1+n2+n5+n7,4)*_r(n3+n8,2)
	      - k6*_r(n4+n6,2)*_r(n1+n2+n5+n7,4)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1+n2,2)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n1+n6,2)*_r(n4+n5+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n2+n6,2)*_r(n4+n5+n7,3)*_r(n3+n8,2)
	      - k4*_r(n1+n2+n6,3)*_r(n4+n5+n7,3)*_r(n3+n8,2)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n1+n4+n5+n7,4)*_r(n3+n8,2)
	      - k6*_r(n2+n6,2)*_r(n1+n4+n5+n7,4)*_r(n3+n8,2)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n2+n4+n5+n7,4)*_r(n3+n8,2)
	      - k6*_r(n1+n6,2)*_r(n2+n4+n5+n7,4)*_r(n3+n8,2)
	      - k24*_r(n6,1)*_r(n1+n2+n4+n5+n7,5)*_r(n3+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n3+n8,2)
	      + _r(n2+n4,2)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n3+n8,2)
	      + _r(n1+n4,2)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6+n7,2)*_r(n3+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n3+n8,2)
	      + _r(n1+n2,2)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n6+n7,2)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n6+n7,2)*_r(n3+n8,2)
	      - k6*_r(n1+n2+n4+n5,4)*_r(n6+n7,2)*_r(n3+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n6+n7,3)*_r(n3+n8,2)
	      - k4*_r(n2+n4+n5,3)*_r(n1+n6+n7,3)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n6+n7,3)*_r(n3+n8,2)
	      - k4*_r(n1+n4+n5,3)*_r(n2+n6+n7,3)*_r(n3+n8,2)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6+n7,4)*_r(n3+n8,2)
	      - k6*_r(n4+n5,2)*_r(n1+n2+n6+n7,4)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n4+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n4+n6+n7,3)*_r(n3+n8,2)
	      - k4*_r(n1+n2+n5,3)*_r(n4+n6+n7,3)*_r(n3+n8,2)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n1+n4+n6+n7,4)*_r(n3+n8,2)
	      - k6*_r(n2+n5,2)*_r(n1+n4+n6+n7,4)*_r(n3+n8,2)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n2+n4+n6+n7,4)*_r(n3+n8,2)
	      - k6*_r(n1+n5,2)*_r(n2+n4+n6+n7,4)*_r(n3+n8,2)
	      - k24*_r(n5,1)*_r(n1+n2+n4+n6+n7,5)*_r(n3+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n5+n6+n7,3)*_r(n3+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n5+n6+n7,3)*_r(n3+n8,2)
	      - k4*_r(n1+n2+n4,3)*_r(n5+n6+n7,3)*_r(n3+n8,2)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n5+n6+n7,4)*_r(n3+n8,2)
	      - k6*_r(n2+n4,2)*_r(n1+n5+n6+n7,4)*_r(n3+n8,2)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n5+n6+n7,4)*_r(n3+n8,2)
	      - k6*_r(n1+n4,2)*_r(n2+n5+n6+n7,4)*_r(n3+n8,2)
	      - k24*_r(n4,1)*_r(n1+n2+n5+n6+n7,5)*_r(n3+n8,2)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n4+n5+n6+n7,4)*_r(n3+n8,2)
	      - k6*_r(n1+n2,2)*_r(n4+n5+n6+n7,4)*_r(n3+n8,2)
	      - k24*_r(n2,1)*_r(n1+n4+n5+n6+n7,5)*_r(n3+n8,2)
	      - k24*_r(n1,1)*_r(n2+n4+n5+n6+n7,5)*_r(n3+n8,2)
	      + k120*_r(n1+n2+n4+n5+n6+n7,6)*_r(n3+n8,2)
	      + k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k4*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k2*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k2*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k4*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k2*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k4*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n1+n3+n8,3)
	      + k4*_r(n2,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k12*_r(n2+n4+n5+n6,4)*_r(n7,1)*_r(n1+n3+n8,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n1+n3+n8,3)
	      - k4*_r(n4+n5+n6,3)*_r(n2+n7,2)*_r(n1+n3+n8,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n2,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n1+n3+n8,3)
	      - k4*_r(n2+n5+n6,3)*_r(n4+n7,2)*_r(n1+n3+n8,3)
	      + k4*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n1+n3+n8,3)
	      - k4*_r(n5+n6,2)*_r(n2+n4+n7,3)*_r(n1+n3+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n2,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n1+n3+n8,3)
	      - k4*_r(n2+n4+n6,3)*_r(n5+n7,2)*_r(n1+n3+n8,3)
	      + k4*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n1+n3+n8,3)
	      - k4*_r(n4+n6,2)*_r(n2+n5+n7,3)*_r(n1+n3+n8,3)
	      + k4*_r(n2,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n1+n3+n8,3)
	      - k4*_r(n2+n6,2)*_r(n4+n5+n7,3)*_r(n1+n3+n8,3)
	      - k12*_r(n6,1)*_r(n2+n4+n5+n7,4)*_r(n1+n3+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n1+n3+n8,3)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n1+n3+n8,3)
	      - k4*_r(n2+n4+n5,3)*_r(n6+n7,2)*_r(n1+n3+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n1+n3+n8,3)
	      - k4*_r(n4+n5,2)*_r(n2+n6+n7,3)*_r(n1+n3+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n1+n3+n8,3)
	      - k4*_r(n2+n5,2)*_r(n4+n6+n7,3)*_r(n1+n3+n8,3)
	      - k12*_r(n5,1)*_r(n2+n4+n6+n7,4)*_r(n1+n3+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n1+n3+n8,3)
	      - k4*_r(n2+n4,2)*_r(n5+n6+n7,3)*_r(n1+n3+n8,3)
	      - k12*_r(n4,1)*_r(n2+n5+n6+n7,4)*_r(n1+n3+n8,3)
	      - k12*_r(n2,1)*_r(n4+n5+n6+n7,4)*_r(n1+n3+n8,3)
	      + k48*_r(n2+n4+n5+n6+n7,5)*_r(n1+n3+n8,3)
	      + k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n1,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k4*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k2*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k2*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k4*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k2*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k4*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n2+n3+n8,3)
	      + k4*_r(n1,1)*_r(n4+n5+n6,3)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k12*_r(n1+n4+n5+n6,4)*_r(n7,1)*_r(n2+n3+n8,3)
	      - k2*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n2+n3+n8,3)
	      - k4*_r(n4+n5+n6,3)*_r(n1+n7,2)*_r(n2+n3+n8,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n1,1)*_r(n5+n6,2)*_r(n4+n7,2)*_r(n2+n3+n8,3)
	      - k4*_r(n1+n5+n6,3)*_r(n4+n7,2)*_r(n2+n3+n8,3)
	      + k4*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n2+n3+n8,3)
	      - k4*_r(n5+n6,2)*_r(n1+n4+n7,3)*_r(n2+n3+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n1,1)*_r(n4+n6,2)*_r(n5+n7,2)*_r(n2+n3+n8,3)
	      - k4*_r(n1+n4+n6,3)*_r(n5+n7,2)*_r(n2+n3+n8,3)
	      + k4*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n2+n3+n8,3)
	      - k4*_r(n4+n6,2)*_r(n1+n5+n7,3)*_r(n2+n3+n8,3)
	      + k4*_r(n1,1)*_r(n6,1)*_r(n4+n5+n7,3)*_r(n2+n3+n8,3)
	      - k4*_r(n1+n6,2)*_r(n4+n5+n7,3)*_r(n2+n3+n8,3)
	      - k12*_r(n6,1)*_r(n1+n4+n5+n7,4)*_r(n2+n3+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n2+n3+n8,3)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n6+n7,2)*_r(n2+n3+n8,3)
	      - k4*_r(n1+n4+n5,3)*_r(n6+n7,2)*_r(n2+n3+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n2+n3+n8,3)
	      - k4*_r(n4+n5,2)*_r(n1+n6+n7,3)*_r(n2+n3+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n4+n6+n7,3)*_r(n2+n3+n8,3)
	      - k4*_r(n1+n5,2)*_r(n4+n6+n7,3)*_r(n2+n3+n8,3)
	      - k12*_r(n5,1)*_r(n1+n4+n6+n7,4)*_r(n2+n3+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n5+n6+n7,3)*_r(n2+n3+n8,3)
	      - k4*_r(n1+n4,2)*_r(n5+n6+n7,3)*_r(n2+n3+n8,3)
	      - k12*_r(n4,1)*_r(n1+n5+n6+n7,4)*_r(n2+n3+n8,3)
	      - k12*_r(n1,1)*_r(n4+n5+n6+n7,4)*_r(n2+n3+n8,3)
	      + k48*_r(n1+n4+n5+n6+n7,5)*_r(n2+n3+n8,3)
	      - k6*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n3+n8,4)
	      + k6*_r(n4+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n3+n8,4)
	      + k6*_r(n5,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n2+n3+n8,4)
	      + k6*_r(n4,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n2+n3+n8,4)
	      - k12*_r(n4+n5+n6,3)*_r(n7,1)*_r(n1+n2+n3+n8,4)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n2+n3+n8,4)
	      - k6*_r(n5+n6,2)*_r(n4+n7,2)*_r(n1+n2+n3+n8,4)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n2+n3+n8,4)
	      - k6*_r(n4+n6,2)*_r(n5+n7,2)*_r(n1+n2+n3+n8,4)
	      - k12*_r(n6,1)*_r(n4+n5+n7,3)*_r(n1+n2+n3+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n2+n3+n8,4)
	      - k6*_r(n4+n5,2)*_r(n6+n7,2)*_r(n1+n2+n3+n8,4)
	      - k12*_r(n5,1)*_r(n4+n6+n7,3)*_r(n1+n2+n3+n8,4)
	      - k12*_r(n4,1)*_r(n5+n6+n7,3)*_r(n1+n2+n3+n8,4)
	      + k36*_r(n4+n5+n6+n7,4)*_r(n1+n2+n3+n8,4)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n2+n3,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1+n3,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1+n2,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + k6*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n2+n3,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n3,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1+n3,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n3,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1+n2,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k6*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n5+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n5+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n5+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n5+n6,2)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n2+n3,2)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n1+n3,2)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k6*_r(n3,1)*_r(n1+n2+n5+n6,4)*_r(n7,1)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n4+n8,2)
	      + k6*_r(n2,1)*_r(n1+n3+n5+n6,4)*_r(n7,1)*_r(n4+n8,2)
	      + k6*_r(n1,1)*_r(n2+n3+n5+n6,4)*_r(n7,1)*_r(n4+n8,2)
	      - k24*_r(n1+n2+n3+n5+n6,5)*_r(n7,1)*_r(n4+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n8,2)
	      - _r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n8,2)
	      - _r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n8,2)
	      + k2*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n8,2)
	      - _r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n4+n8,2)
	      + _r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n4+n8,2)
	      + _r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n4+n8,2)
	      + k2*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n4+n8,2)
	      + _r(n2+n3,2)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n2+n5+n6,3)*_r(n1+n7,2)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n3+n5+n6,3)*_r(n1+n7,2)*_r(n4+n8,2)
	      - k6*_r(n2+n3+n5+n6,4)*_r(n1+n7,2)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n8,2)
	      - _r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n8,2)
	      - _r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n8,2)
	      - _r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n4+n8,2)
	      + _r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n4+n8,2)
	      + _r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n4+n8,2)
	      + k2*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n4+n8,2)
	      + _r(n1+n3,2)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n1+n5+n6,3)*_r(n2+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n3+n5+n6,3)*_r(n2+n7,2)*_r(n4+n8,2)
	      - k6*_r(n1+n3+n5+n6,4)*_r(n2+n7,2)*_r(n4+n8,2)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n4+n8,2)
	      + k2*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n4+n8,2)
	      + k2*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n7,3)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n5+n6,2)*_r(n1+n2+n7,3)*_r(n4+n8,2)
	      - k4*_r(n3+n5+n6,3)*_r(n1+n2+n7,3)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n8,2)
	      - _r(n1+n2,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n4+n8,2)
	      + _r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n4+n8,2)
	      + _r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n4+n8,2)
	      + k2*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n4+n8,2)
	      + _r(n1+n2,2)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n1+n5+n6,3)*_r(n3+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n2+n5+n6,3)*_r(n3+n7,2)*_r(n4+n8,2)
	      - k6*_r(n1+n2+n5+n6,4)*_r(n3+n7,2)*_r(n4+n8,2)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n4+n8,2)
	      + k2*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n5+n6,2)*_r(n1+n3+n7,3)*_r(n4+n8,2)
	      - k4*_r(n2+n5+n6,3)*_r(n1+n3+n7,3)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n4+n8,2)
	      + k2*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n5+n6,2)*_r(n2+n3+n7,3)*_r(n4+n8,2)
	      - k4*_r(n1+n5+n6,3)*_r(n2+n3+n7,3)*_r(n4+n8,2)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n7,4)*_r(n4+n8,2)
	      - k6*_r(n5+n6,2)*_r(n1+n2+n3+n7,4)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n5+n7,2)*_r(n4+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n6,1)*_r(n5+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n6,1)*_r(n5+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n6,1)*_r(n5+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n6,1)*_r(n5+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n4+n8,2)
	      + _r(n2+n3,2)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n4+n8,2)
	      + _r(n1+n3,2)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n6,3)*_r(n5+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n4+n8,2)
	      + _r(n1+n2,2)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n6,3)*_r(n5+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n6,3)*_r(n5+n7,2)*_r(n4+n8,2)
	      - k6*_r(n1+n2+n3+n6,4)*_r(n5+n7,2)*_r(n4+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2+n3,2)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n2+n6,2)*_r(n1+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n3+n6,2)*_r(n1+n5+n7,3)*_r(n4+n8,2)
	      - k4*_r(n2+n3+n6,3)*_r(n1+n5+n7,3)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1+n3,2)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n1+n6,2)*_r(n2+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n3+n6,2)*_r(n2+n5+n7,3)*_r(n4+n8,2)
	      - k4*_r(n1+n3+n6,3)*_r(n2+n5+n7,3)*_r(n4+n8,2)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n1+n2+n5+n7,4)*_r(n4+n8,2)
	      - k6*_r(n3+n6,2)*_r(n1+n2+n5+n7,4)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1+n2,2)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n1+n6,2)*_r(n3+n5+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n2+n6,2)*_r(n3+n5+n7,3)*_r(n4+n8,2)
	      - k4*_r(n1+n2+n6,3)*_r(n3+n5+n7,3)*_r(n4+n8,2)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n1+n3+n5+n7,4)*_r(n4+n8,2)
	      - k6*_r(n2+n6,2)*_r(n1+n3+n5+n7,4)*_r(n4+n8,2)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n2+n3+n5+n7,4)*_r(n4+n8,2)
	      - k6*_r(n1+n6,2)*_r(n2+n3+n5+n7,4)*_r(n4+n8,2)
	      - k24*_r(n6,1)*_r(n1+n2+n3+n5+n7,5)*_r(n4+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6+n7,2)*_r(n4+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n6+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n6+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n6+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n6+n7,2)*_r(n4+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n4+n8,2)
	      + _r(n2+n3,2)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n4+n8,2)
	      + _r(n1+n3,2)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n6+n7,2)*_r(n4+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n4+n8,2)
	      + _r(n1+n2,2)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n6+n7,2)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n6+n7,2)*_r(n4+n8,2)
	      - k6*_r(n1+n2+n3+n5,4)*_r(n6+n7,2)*_r(n4+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n6+n7,3)*_r(n4+n8,2)
	      - k4*_r(n2+n3+n5,3)*_r(n1+n6+n7,3)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n6+n7,3)*_r(n4+n8,2)
	      - k4*_r(n1+n3+n5,3)*_r(n2+n6+n7,3)*_r(n4+n8,2)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n1+n2+n6+n7,4)*_r(n4+n8,2)
	      - k6*_r(n3+n5,2)*_r(n1+n2+n6+n7,4)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n6+n7,3)*_r(n4+n8,2)
	      - k4*_r(n1+n2+n5,3)*_r(n3+n6+n7,3)*_r(n4+n8,2)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n1+n3+n6+n7,4)*_r(n4+n8,2)
	      - k6*_r(n2+n5,2)*_r(n1+n3+n6+n7,4)*_r(n4+n8,2)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n2+n3+n6+n7,4)*_r(n4+n8,2)
	      - k6*_r(n1+n5,2)*_r(n2+n3+n6+n7,4)*_r(n4+n8,2)
	      - k24*_r(n5,1)*_r(n1+n2+n3+n6+n7,5)*_r(n4+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n5+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n5+n6+n7,3)*_r(n4+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n5+n6+n7,3)*_r(n4+n8,2)
	      - k4*_r(n1+n2+n3,3)*_r(n5+n6+n7,3)*_r(n4+n8,2)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n5+n6+n7,4)*_r(n4+n8,2)
	      - k6*_r(n2+n3,2)*_r(n1+n5+n6+n7,4)*_r(n4+n8,2)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n5+n6+n7,4)*_r(n4+n8,2)
	      - k6*_r(n1+n3,2)*_r(n2+n5+n6+n7,4)*_r(n4+n8,2)
	      - k24*_r(n3,1)*_r(n1+n2+n5+n6+n7,5)*_r(n4+n8,2)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n5+n6+n7,4)*_r(n4+n8,2)
	      - k6*_r(n1+n2,2)*_r(n3+n5+n6+n7,4)*_r(n4+n8,2)
	      - k24*_r(n2,1)*_r(n1+n3+n5+n6+n7,5)*_r(n4+n8,2)
	      - k24*_r(n1,1)*_r(n2+n3+n5+n6+n7,5)*_r(n4+n8,2)
	      + k120*_r(n1+n2+n3+n5+n6+n7,6)*_r(n4+n8,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k4*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k2*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k2*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k4*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k2*_r(n2+n3,2)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k4*_r(n3,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n1+n4+n8,3)
	      + k4*_r(n2,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k12*_r(n2+n3+n5+n6,4)*_r(n7,1)*_r(n1+n4+n8,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n3,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n1+n4+n8,3)
	      - k4*_r(n3+n5+n6,3)*_r(n2+n7,2)*_r(n1+n4+n8,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n2,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n1+n4+n8,3)
	      - k4*_r(n2+n5+n6,3)*_r(n3+n7,2)*_r(n1+n4+n8,3)
	      + k4*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n1+n4+n8,3)
	      - k4*_r(n5+n6,2)*_r(n2+n3+n7,3)*_r(n1+n4+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n2+n3,2)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n3,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n2,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n1+n4+n8,3)
	      - k4*_r(n2+n3+n6,3)*_r(n5+n7,2)*_r(n1+n4+n8,3)
	      + k4*_r(n3,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n1+n4+n8,3)
	      - k4*_r(n3+n6,2)*_r(n2+n5+n7,3)*_r(n1+n4+n8,3)
	      + k4*_r(n2,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n1+n4+n8,3)
	      - k4*_r(n2+n6,2)*_r(n3+n5+n7,3)*_r(n1+n4+n8,3)
	      - k12*_r(n6,1)*_r(n2+n3+n5+n7,4)*_r(n1+n4+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n1+n4+n8,3)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n1+n4+n8,3)
	      - k4*_r(n2+n3+n5,3)*_r(n6+n7,2)*_r(n1+n4+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n1+n4+n8,3)
	      - k4*_r(n3+n5,2)*_r(n2+n6+n7,3)*_r(n1+n4+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n1+n4+n8,3)
	      - k4*_r(n2+n5,2)*_r(n3+n6+n7,3)*_r(n1+n4+n8,3)
	      - k12*_r(n5,1)*_r(n2+n3+n6+n7,4)*_r(n1+n4+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n5+n6+n7,3)*_r(n1+n4+n8,3)
	      - k4*_r(n2+n3,2)*_r(n5+n6+n7,3)*_r(n1+n4+n8,3)
	      - k12*_r(n3,1)*_r(n2+n5+n6+n7,4)*_r(n1+n4+n8,3)
	      - k12*_r(n2,1)*_r(n3+n5+n6+n7,4)*_r(n1+n4+n8,3)
	      + k48*_r(n2+n3+n5+n6+n7,5)*_r(n1+n4+n8,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n1,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k4*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k2*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k2*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k4*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k2*_r(n1+n3,2)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k4*_r(n3,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n2+n4+n8,3)
	      + k4*_r(n1,1)*_r(n3+n5+n6,3)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k12*_r(n1+n3+n5+n6,4)*_r(n7,1)*_r(n2+n4+n8,3)
	      - k2*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n3,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n2+n4+n8,3)
	      - k4*_r(n3+n5+n6,3)*_r(n1+n7,2)*_r(n2+n4+n8,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n1,1)*_r(n5+n6,2)*_r(n3+n7,2)*_r(n2+n4+n8,3)
	      - k4*_r(n1+n5+n6,3)*_r(n3+n7,2)*_r(n2+n4+n8,3)
	      + k4*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n2+n4+n8,3)
	      - k4*_r(n5+n6,2)*_r(n1+n3+n7,3)*_r(n2+n4+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n1+n3,2)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n3,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n1,1)*_r(n3+n6,2)*_r(n5+n7,2)*_r(n2+n4+n8,3)
	      - k4*_r(n1+n3+n6,3)*_r(n5+n7,2)*_r(n2+n4+n8,3)
	      + k4*_r(n3,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n2+n4+n8,3)
	      - k4*_r(n3+n6,2)*_r(n1+n5+n7,3)*_r(n2+n4+n8,3)
	      + k4*_r(n1,1)*_r(n6,1)*_r(n3+n5+n7,3)*_r(n2+n4+n8,3)
	      - k4*_r(n1+n6,2)*_r(n3+n5+n7,3)*_r(n2+n4+n8,3)
	      - k12*_r(n6,1)*_r(n1+n3+n5+n7,4)*_r(n2+n4+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n2+n4+n8,3)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n6+n7,2)*_r(n2+n4+n8,3)
	      - k4*_r(n1+n3+n5,3)*_r(n6+n7,2)*_r(n2+n4+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n2+n4+n8,3)
	      - k4*_r(n3+n5,2)*_r(n1+n6+n7,3)*_r(n2+n4+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n3+n6+n7,3)*_r(n2+n4+n8,3)
	      - k4*_r(n1+n5,2)*_r(n3+n6+n7,3)*_r(n2+n4+n8,3)
	      - k12*_r(n5,1)*_r(n1+n3+n6+n7,4)*_r(n2+n4+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n5+n6+n7,3)*_r(n2+n4+n8,3)
	      - k4*_r(n1+n3,2)*_r(n5+n6+n7,3)*_r(n2+n4+n8,3)
	      - k12*_r(n3,1)*_r(n1+n5+n6+n7,4)*_r(n2+n4+n8,3)
	      - k12*_r(n1,1)*_r(n3+n5+n6+n7,4)*_r(n2+n4+n8,3)
	      + k48*_r(n1+n3+n5+n6+n7,5)*_r(n2+n4+n8,3)
	      - k6*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n4+n8,4)
	      + k6*_r(n3+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n4+n8,4)
	      + k6*_r(n5,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n2+n4+n8,4)
	      + k6*_r(n3,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n2+n4+n8,4)
	      - k12*_r(n3+n5+n6,3)*_r(n7,1)*_r(n1+n2+n4+n8,4)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n2+n4+n8,4)
	      - k6*_r(n5+n6,2)*_r(n3+n7,2)*_r(n1+n2+n4+n8,4)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n2+n4+n8,4)
	      - k6*_r(n3+n6,2)*_r(n5+n7,2)*_r(n1+n2+n4+n8,4)
	      - k12*_r(n6,1)*_r(n3+n5+n7,3)*_r(n1+n2+n4+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n2+n4+n8,4)
	      - k6*_r(n3+n5,2)*_r(n6+n7,2)*_r(n1+n2+n4+n8,4)
	      - k12*_r(n5,1)*_r(n3+n6+n7,3)*_r(n1+n2+n4+n8,4)
	      - k12*_r(n3,1)*_r(n5+n6+n7,3)*_r(n1+n2+n4+n8,4)
	      + k36*_r(n3+n5+n6+n7,4)*_r(n1+n2+n4+n8,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n1+n2,2)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n2,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n1,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k4*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k2*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k2*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k4*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k2*_r(n1+n2,2)*_r(n5+n6,2)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k4*_r(n2,1)*_r(n1+n5+n6,3)*_r(n7,1)*_r(n3+n4+n8,3)
	      + k4*_r(n1,1)*_r(n2+n5+n6,3)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k12*_r(n1+n2+n5+n6,4)*_r(n7,1)*_r(n3+n4+n8,3)
	      - k2*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n2,1)*_r(n5+n6,2)*_r(n1+n7,2)*_r(n3+n4+n8,3)
	      - k4*_r(n2+n5+n6,3)*_r(n1+n7,2)*_r(n3+n4+n8,3)
	      - k2*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n4+n8,3));
    }
    /** 
     * Third part of the 8-particle correlator 
     * 
     * @param n1 Harmonic
     * @param n2 Harmonic 
     * @param n3 Harmonic 
     * @param n4 Harmonic 
     * @param n5 Harmonic 
     * @param n6 Harmonic 
     * @param n7 Harmonic 
     * @param n8 Harmonic 
     * 
     * @return Third part 
     */
    Complex c8P3(const HarmonicElement n1,
		 const HarmonicElement n2,
		 const HarmonicElement n3,
		 const HarmonicElement n4,
		 const HarmonicElement n5,
		 const HarmonicElement n6,
		 const HarmonicElement n7,
		 const HarmonicElement n8) const
    {
      const Real k2   = 2;
      const Real k4   = 4;
      const Real k6   = 6;
      const Real k12  = 12;
      const Real k24  = 24;
      const Real k36  = 36;
      const Real k48  = 48;
      const Real k120 = 120;
      // const Real k720 = 720;
      return (k2*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n1,1)*_r(n5+n6,2)*_r(n2+n7,2)*_r(n3+n4+n8,3)
	      - k4*_r(n1+n5+n6,3)*_r(n2+n7,2)*_r(n3+n4+n8,3)
	      + k4*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n3+n4+n8,3)
	      - k4*_r(n5+n6,2)*_r(n1+n2+n7,3)*_r(n3+n4+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n1+n2,2)*_r(n6,1)*_r(n5+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n2,1)*_r(n1+n6,2)*_r(n5+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n1,1)*_r(n2+n6,2)*_r(n5+n7,2)*_r(n3+n4+n8,3)
	      - k4*_r(n1+n2+n6,3)*_r(n5+n7,2)*_r(n3+n4+n8,3)
	      + k4*_r(n2,1)*_r(n6,1)*_r(n1+n5+n7,3)*_r(n3+n4+n8,3)
	      - k4*_r(n2+n6,2)*_r(n1+n5+n7,3)*_r(n3+n4+n8,3)
	      + k4*_r(n1,1)*_r(n6,1)*_r(n2+n5+n7,3)*_r(n3+n4+n8,3)
	      - k4*_r(n1+n6,2)*_r(n2+n5+n7,3)*_r(n3+n4+n8,3)
	      - k12*_r(n6,1)*_r(n1+n2+n5+n7,4)*_r(n3+n4+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n6+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n6+n7,2)*_r(n3+n4+n8,3)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n6+n7,2)*_r(n3+n4+n8,3)
	      - k4*_r(n1+n2+n5,3)*_r(n6+n7,2)*_r(n3+n4+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n1+n6+n7,3)*_r(n3+n4+n8,3)
	      - k4*_r(n2+n5,2)*_r(n1+n6+n7,3)*_r(n3+n4+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n2+n6+n7,3)*_r(n3+n4+n8,3)
	      - k4*_r(n1+n5,2)*_r(n2+n6+n7,3)*_r(n3+n4+n8,3)
	      - k12*_r(n5,1)*_r(n1+n2+n6+n7,4)*_r(n3+n4+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n5+n6+n7,3)*_r(n3+n4+n8,3)
	      - k4*_r(n1+n2,2)*_r(n5+n6+n7,3)*_r(n3+n4+n8,3)
	      - k12*_r(n2,1)*_r(n1+n5+n6+n7,4)*_r(n3+n4+n8,3)
	      - k12*_r(n1,1)*_r(n2+n5+n6+n7,4)*_r(n3+n4+n8,3)
	      + k48*_r(n1+n2+n5+n6+n7,5)*_r(n3+n4+n8,3)
	      - k6*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n4+n8,4)
	      + k6*_r(n2+n5,2)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n4+n8,4)
	      + k6*_r(n5,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n3+n4+n8,4)
	      + k6*_r(n2,1)*_r(n5+n6,2)*_r(n7,1)*_r(n1+n3+n4+n8,4)
	      - k12*_r(n2+n5+n6,3)*_r(n7,1)*_r(n1+n3+n4+n8,4)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n3+n4+n8,4)
	      - k6*_r(n5+n6,2)*_r(n2+n7,2)*_r(n1+n3+n4+n8,4)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n5+n7,2)*_r(n1+n3+n4+n8,4)
	      - k6*_r(n2+n6,2)*_r(n5+n7,2)*_r(n1+n3+n4+n8,4)
	      - k12*_r(n6,1)*_r(n2+n5+n7,3)*_r(n1+n3+n4+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n6+n7,2)*_r(n1+n3+n4+n8,4)
	      - k6*_r(n2+n5,2)*_r(n6+n7,2)*_r(n1+n3+n4+n8,4)
	      - k12*_r(n5,1)*_r(n2+n6+n7,3)*_r(n1+n3+n4+n8,4)
	      - k12*_r(n2,1)*_r(n5+n6+n7,3)*_r(n1+n3+n4+n8,4)
	      + k36*_r(n2+n5+n6+n7,4)*_r(n1+n3+n4+n8,4)
	      - k6*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n4+n8,4)
	      + k6*_r(n1+n5,2)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n4+n8,4)
	      + k6*_r(n5,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n3+n4+n8,4)
	      + k6*_r(n1,1)*_r(n5+n6,2)*_r(n7,1)*_r(n2+n3+n4+n8,4)
	      - k12*_r(n1+n5+n6,3)*_r(n7,1)*_r(n2+n3+n4+n8,4)
	      + k6*_r(n5,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n3+n4+n8,4)
	      - k6*_r(n5+n6,2)*_r(n1+n7,2)*_r(n2+n3+n4+n8,4)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n5+n7,2)*_r(n2+n3+n4+n8,4)
	      - k6*_r(n1+n6,2)*_r(n5+n7,2)*_r(n2+n3+n4+n8,4)
	      - k12*_r(n6,1)*_r(n1+n5+n7,3)*_r(n2+n3+n4+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n6+n7,2)*_r(n2+n3+n4+n8,4)
	      - k6*_r(n1+n5,2)*_r(n6+n7,2)*_r(n2+n3+n4+n8,4)
	      - k12*_r(n5,1)*_r(n1+n6+n7,3)*_r(n2+n3+n4+n8,4)
	      - k12*_r(n1,1)*_r(n5+n6+n7,3)*_r(n2+n3+n4+n8,4)
	      + k36*_r(n1+n5+n6+n7,4)*_r(n2+n3+n4+n8,4)
	      + k24*_r(n5,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n3+n4+n8,5)
	      - k24*_r(n5+n6,2)*_r(n7,1)*_r(n1+n2+n3+n4+n8,5)
	      - k24*_r(n6,1)*_r(n5+n7,2)*_r(n1+n2+n3+n4+n8,5)
	      - k24*_r(n5,1)*_r(n6+n7,2)*_r(n1+n2+n3+n4+n8,5)
	      + k48*_r(n5+n6+n7,3)*_r(n1+n2+n3+n4+n8,5)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n2+n3,2)*_r(n1+n4,2)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1+n3,2)*_r(n2+n4,2)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1+n2,2)*_r(n3+n4,2)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n6,1)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n1+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n1+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n1+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n2+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n2+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n2+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n3+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n3+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n3+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n6,4)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n6,2)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n6,4)*_r(n7,1)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n5+n8,2)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n6,4)*_r(n7,1)*_r(n5+n8,2)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n6,4)*_r(n7,1)*_r(n5+n8,2)
	      - k24*_r(n1+n2+n3+n4+n6,5)*_r(n7,1)*_r(n5+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n7,2)*_r(n5+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n1+n7,2)*_r(n5+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n1+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n1+n7,2)*_r(n5+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n6,1)*_r(n1+n7,2)*_r(n5+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n5+n8,2)
	      + _r(n3+n4,2)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n5+n8,2)
	      + _r(n2+n4,2)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n5+n8,2)
	      + k2*_r(n4,1)*_r(n2+n3+n6,3)*_r(n1+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n5+n8,2)
	      + _r(n2+n3,2)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4+n6,3)*_r(n1+n7,2)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4+n6,3)*_r(n1+n7,2)*_r(n5+n8,2)
	      - k6*_r(n2+n3+n4+n6,4)*_r(n1+n7,2)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n2+n7,2)*_r(n5+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n2+n7,2)*_r(n5+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n2+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n6,1)*_r(n2+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n6,1)*_r(n2+n7,2)*_r(n5+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n5+n8,2)
	      + _r(n3+n4,2)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n5+n8,2)
	      + _r(n1+n4,2)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n5+n8,2)
	      + k2*_r(n4,1)*_r(n1+n3+n6,3)*_r(n2+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n5+n8,2)
	      + _r(n1+n3,2)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4+n6,3)*_r(n2+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4+n6,3)*_r(n2+n7,2)*_r(n5+n8,2)
	      - k6*_r(n1+n3+n4+n6,4)*_r(n2+n7,2)*_r(n5+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n5+n8,2)
	      + k2*_r(n3+n4,2)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n5+n8,2)
	      + k2*_r(n4,1)*_r(n3+n6,2)*_r(n1+n2+n7,3)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n4+n6,2)*_r(n1+n2+n7,3)*_r(n5+n8,2)
	      - k4*_r(n3+n4+n6,3)*_r(n1+n2+n7,3)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n3+n7,2)*_r(n5+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n6,1)*_r(n3+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n6,1)*_r(n3+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n6,1)*_r(n3+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n6,1)*_r(n3+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n5+n8,2)
	      + _r(n2+n4,2)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n5+n8,2)
	      + _r(n1+n4,2)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n5+n8,2)
	      + k2*_r(n4,1)*_r(n1+n2+n6,3)*_r(n3+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n5+n8,2)
	      + _r(n1+n2,2)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4+n6,3)*_r(n3+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4+n6,3)*_r(n3+n7,2)*_r(n5+n8,2)
	      - k6*_r(n1+n2+n4+n6,4)*_r(n3+n7,2)*_r(n5+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2+n4,2)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n5+n8,2)
	      + k2*_r(n4,1)*_r(n2+n6,2)*_r(n1+n3+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n4+n6,2)*_r(n1+n3+n7,3)*_r(n5+n8,2)
	      - k4*_r(n2+n4+n6,3)*_r(n1+n3+n7,3)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1+n4,2)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n5+n8,2)
	      + k2*_r(n4,1)*_r(n1+n6,2)*_r(n2+n3+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n4+n6,2)*_r(n2+n3+n7,3)*_r(n5+n8,2)
	      - k4*_r(n1+n4+n6,3)*_r(n2+n3+n7,3)*_r(n5+n8,2)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n1+n2+n3+n7,4)*_r(n5+n8,2)
	      - k6*_r(n4+n6,2)*_r(n1+n2+n3+n7,4)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n4+n7,2)*_r(n5+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n6,1)*_r(n4+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n6,1)*_r(n4+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n6,1)*_r(n4+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n6,1)*_r(n4+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n5+n8,2)
	      + _r(n2+n3,2)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n5+n8,2)
	      + _r(n1+n3,2)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n6,3)*_r(n4+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n5+n8,2)
	      + _r(n1+n2,2)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n6,3)*_r(n4+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n6,3)*_r(n4+n7,2)*_r(n5+n8,2)
	      - k6*_r(n1+n2+n3+n6,4)*_r(n4+n7,2)*_r(n5+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2+n3,2)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n2+n6,2)*_r(n1+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n3+n6,2)*_r(n1+n4+n7,3)*_r(n5+n8,2)
	      - k4*_r(n2+n3+n6,3)*_r(n1+n4+n7,3)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1+n3,2)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n1+n6,2)*_r(n2+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n3+n6,2)*_r(n2+n4+n7,3)*_r(n5+n8,2)
	      - k4*_r(n1+n3+n6,3)*_r(n2+n4+n7,3)*_r(n5+n8,2)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n1+n2+n4+n7,4)*_r(n5+n8,2)
	      - k6*_r(n3+n6,2)*_r(n1+n2+n4+n7,4)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1+n2,2)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n1+n6,2)*_r(n3+n4+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n2+n6,2)*_r(n3+n4+n7,3)*_r(n5+n8,2)
	      - k4*_r(n1+n2+n6,3)*_r(n3+n4+n7,3)*_r(n5+n8,2)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n1+n3+n4+n7,4)*_r(n5+n8,2)
	      - k6*_r(n2+n6,2)*_r(n1+n3+n4+n7,4)*_r(n5+n8,2)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n2+n3+n4+n7,4)*_r(n5+n8,2)
	      - k6*_r(n1+n6,2)*_r(n2+n3+n4+n7,4)*_r(n5+n8,2)
	      - k24*_r(n6,1)*_r(n1+n2+n3+n4+n7,5)*_r(n5+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6+n7,2)*_r(n5+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n6+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n6+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n6+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n6+n7,2)*_r(n5+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n6+n7,2)*_r(n5+n8,2)
	      + _r(n2+n3,2)*_r(n1+n4,2)*_r(n6+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n6+n7,2)*_r(n5+n8,2)
	      + _r(n1+n3,2)*_r(n2+n4,2)*_r(n6+n7,2)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n6+n7,2)*_r(n5+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n6+n7,2)*_r(n5+n8,2)
	      + _r(n1+n2,2)*_r(n3+n4,2)*_r(n6+n7,2)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n6+n7,2)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n6+n7,2)*_r(n5+n8,2)
	      - k6*_r(n1+n2+n3+n4,4)*_r(n6+n7,2)*_r(n5+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n6+n7,3)*_r(n5+n8,2)
	      - k4*_r(n2+n3+n4,3)*_r(n1+n6+n7,3)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n6+n7,3)*_r(n5+n8,2)
	      - k4*_r(n1+n3+n4,3)*_r(n2+n6+n7,3)*_r(n5+n8,2)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n2+n6+n7,4)*_r(n5+n8,2)
	      - k6*_r(n3+n4,2)*_r(n1+n2+n6+n7,4)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n6+n7,3)*_r(n5+n8,2)
	      - k4*_r(n1+n2+n4,3)*_r(n3+n6+n7,3)*_r(n5+n8,2)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n3+n6+n7,4)*_r(n5+n8,2)
	      - k6*_r(n2+n4,2)*_r(n1+n3+n6+n7,4)*_r(n5+n8,2)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n3+n6+n7,4)*_r(n5+n8,2)
	      - k6*_r(n1+n4,2)*_r(n2+n3+n6+n7,4)*_r(n5+n8,2)
	      - k24*_r(n4,1)*_r(n1+n2+n3+n6+n7,5)*_r(n5+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n6+n7,3)*_r(n5+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n6+n7,3)*_r(n5+n8,2)
	      - k4*_r(n1+n2+n3,3)*_r(n4+n6+n7,3)*_r(n5+n8,2)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n4+n6+n7,4)*_r(n5+n8,2)
	      - k6*_r(n2+n3,2)*_r(n1+n4+n6+n7,4)*_r(n5+n8,2)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n4+n6+n7,4)*_r(n5+n8,2)
	      - k6*_r(n1+n3,2)*_r(n2+n4+n6+n7,4)*_r(n5+n8,2)
	      - k24*_r(n3,1)*_r(n1+n2+n4+n6+n7,5)*_r(n5+n8,2)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n4+n6+n7,4)*_r(n5+n8,2)
	      - k6*_r(n1+n2,2)*_r(n3+n4+n6+n7,4)*_r(n5+n8,2)
	      - k24*_r(n2,1)*_r(n1+n3+n4+n6+n7,5)*_r(n5+n8,2)
	      - k24*_r(n1,1)*_r(n2+n3+n4+n6+n7,5)*_r(n5+n8,2)
	      + k120*_r(n1+n2+n3+n4+n6+n7,6)*_r(n5+n8,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k4*_r(n2+n3+n4,3)*_r(n6,1)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k2*_r(n3+n4,2)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k2*_r(n2+n4,2)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k4*_r(n4,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k4*_r(n3,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n1+n5+n8,3)
	      + k4*_r(n2,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k12*_r(n2+n3+n4+n6,4)*_r(n7,1)*_r(n1+n5+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n3+n4,2)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n4,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n3,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n1+n5+n8,3)
	      - k4*_r(n3+n4+n6,3)*_r(n2+n7,2)*_r(n1+n5+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n2+n4,2)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n4,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n2,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n1+n5+n8,3)
	      - k4*_r(n2+n4+n6,3)*_r(n3+n7,2)*_r(n1+n5+n8,3)
	      + k4*_r(n4,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n1+n5+n8,3)
	      - k4*_r(n4+n6,2)*_r(n2+n3+n7,3)*_r(n1+n5+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n2+n3,2)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n3,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n2,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n1+n5+n8,3)
	      - k4*_r(n2+n3+n6,3)*_r(n4+n7,2)*_r(n1+n5+n8,3)
	      + k4*_r(n3,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n1+n5+n8,3)
	      - k4*_r(n3+n6,2)*_r(n2+n4+n7,3)*_r(n1+n5+n8,3)
	      + k4*_r(n2,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n1+n5+n8,3)
	      - k4*_r(n2+n6,2)*_r(n3+n4+n7,3)*_r(n1+n5+n8,3)
	      - k12*_r(n6,1)*_r(n2+n3+n4+n7,4)*_r(n1+n5+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n6+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n6+n7,2)*_r(n1+n5+n8,3)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n6+n7,2)*_r(n1+n5+n8,3)
	      - k4*_r(n2+n3+n4,3)*_r(n6+n7,2)*_r(n1+n5+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n2+n6+n7,3)*_r(n1+n5+n8,3)
	      - k4*_r(n3+n4,2)*_r(n2+n6+n7,3)*_r(n1+n5+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n3+n6+n7,3)*_r(n1+n5+n8,3)
	      - k4*_r(n2+n4,2)*_r(n3+n6+n7,3)*_r(n1+n5+n8,3)
	      - k12*_r(n4,1)*_r(n2+n3+n6+n7,4)*_r(n1+n5+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n4+n6+n7,3)*_r(n1+n5+n8,3)
	      - k4*_r(n2+n3,2)*_r(n4+n6+n7,3)*_r(n1+n5+n8,3)
	      - k12*_r(n3,1)*_r(n2+n4+n6+n7,4)*_r(n1+n5+n8,3)
	      - k12*_r(n2,1)*_r(n3+n4+n6+n7,4)*_r(n1+n5+n8,3)
	      + k48*_r(n2+n3+n4+n6+n7,5)*_r(n1+n5+n8,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n6,1)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k4*_r(n1+n3+n4,3)*_r(n6,1)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k2*_r(n3+n4,2)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k2*_r(n1+n4,2)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k4*_r(n4,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k4*_r(n3,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n2+n5+n8,3)
	      + k4*_r(n1,1)*_r(n3+n4+n6,3)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k12*_r(n1+n3+n4+n6,4)*_r(n7,1)*_r(n2+n5+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n3+n4,2)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n4,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n3,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n2+n5+n8,3)
	      - k4*_r(n3+n4+n6,3)*_r(n1+n7,2)*_r(n2+n5+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n1+n4,2)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n4,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n1,1)*_r(n4+n6,2)*_r(n3+n7,2)*_r(n2+n5+n8,3)
	      - k4*_r(n1+n4+n6,3)*_r(n3+n7,2)*_r(n2+n5+n8,3)
	      + k4*_r(n4,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n2+n5+n8,3)
	      - k4*_r(n4+n6,2)*_r(n1+n3+n7,3)*_r(n2+n5+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n1+n3,2)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n3,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n1,1)*_r(n3+n6,2)*_r(n4+n7,2)*_r(n2+n5+n8,3)
	      - k4*_r(n1+n3+n6,3)*_r(n4+n7,2)*_r(n2+n5+n8,3)
	      + k4*_r(n3,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n2+n5+n8,3)
	      - k4*_r(n3+n6,2)*_r(n1+n4+n7,3)*_r(n2+n5+n8,3)
	      + k4*_r(n1,1)*_r(n6,1)*_r(n3+n4+n7,3)*_r(n2+n5+n8,3)
	      - k4*_r(n1+n6,2)*_r(n3+n4+n7,3)*_r(n2+n5+n8,3)
	      - k12*_r(n6,1)*_r(n1+n3+n4+n7,4)*_r(n2+n5+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n6+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n6+n7,2)*_r(n2+n5+n8,3)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n6+n7,2)*_r(n2+n5+n8,3)
	      - k4*_r(n1+n3+n4,3)*_r(n6+n7,2)*_r(n2+n5+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n1+n6+n7,3)*_r(n2+n5+n8,3)
	      - k4*_r(n3+n4,2)*_r(n1+n6+n7,3)*_r(n2+n5+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n3+n6+n7,3)*_r(n2+n5+n8,3)
	      - k4*_r(n1+n4,2)*_r(n3+n6+n7,3)*_r(n2+n5+n8,3)
	      - k12*_r(n4,1)*_r(n1+n3+n6+n7,4)*_r(n2+n5+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n4+n6+n7,3)*_r(n2+n5+n8,3)
	      - k4*_r(n1+n3,2)*_r(n4+n6+n7,3)*_r(n2+n5+n8,3)
	      - k12*_r(n3,1)*_r(n1+n4+n6+n7,4)*_r(n2+n5+n8,3)
	      - k12*_r(n1,1)*_r(n3+n4+n6+n7,4)*_r(n2+n5+n8,3)
	      + k48*_r(n1+n3+n4+n6+n7,5)*_r(n2+n5+n8,3)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n5+n8,4)
	      + k6*_r(n3+n4,2)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n5+n8,4)
	      + k6*_r(n4,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n2+n5+n8,4)
	      + k6*_r(n3,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n2+n5+n8,4)
	      - k12*_r(n3+n4+n6,3)*_r(n7,1)*_r(n1+n2+n5+n8,4)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n2+n5+n8,4)
	      - k6*_r(n4+n6,2)*_r(n3+n7,2)*_r(n1+n2+n5+n8,4)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n2+n5+n8,4)
	      - k6*_r(n3+n6,2)*_r(n4+n7,2)*_r(n1+n2+n5+n8,4)
	      - k12*_r(n6,1)*_r(n3+n4+n7,3)*_r(n1+n2+n5+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n6+n7,2)*_r(n1+n2+n5+n8,4)
	      - k6*_r(n3+n4,2)*_r(n6+n7,2)*_r(n1+n2+n5+n8,4)
	      - k12*_r(n4,1)*_r(n3+n6+n7,3)*_r(n1+n2+n5+n8,4)
	      - k12*_r(n3,1)*_r(n4+n6+n7,3)*_r(n1+n2+n5+n8,4)
	      + k36*_r(n3+n4+n6+n7,4)*_r(n1+n2+n5+n8,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n6,1)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n6,1)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k4*_r(n1+n2+n4,3)*_r(n6,1)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k2*_r(n2+n4,2)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k2*_r(n1+n4,2)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k4*_r(n4,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4+n6,2)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k4*_r(n2,1)*_r(n1+n4+n6,3)*_r(n7,1)*_r(n3+n5+n8,3)
	      + k4*_r(n1,1)*_r(n2+n4+n6,3)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k12*_r(n1+n2+n4+n6,4)*_r(n7,1)*_r(n3+n5+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n2+n4,2)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n4,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n2,1)*_r(n4+n6,2)*_r(n1+n7,2)*_r(n3+n5+n8,3)
	      - k4*_r(n2+n4+n6,3)*_r(n1+n7,2)*_r(n3+n5+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n1+n4,2)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n4,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n1,1)*_r(n4+n6,2)*_r(n2+n7,2)*_r(n3+n5+n8,3)
	      - k4*_r(n1+n4+n6,3)*_r(n2+n7,2)*_r(n3+n5+n8,3)
	      + k4*_r(n4,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n3+n5+n8,3)
	      - k4*_r(n4+n6,2)*_r(n1+n2+n7,3)*_r(n3+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n1+n2,2)*_r(n6,1)*_r(n4+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n2,1)*_r(n1+n6,2)*_r(n4+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n1,1)*_r(n2+n6,2)*_r(n4+n7,2)*_r(n3+n5+n8,3)
	      - k4*_r(n1+n2+n6,3)*_r(n4+n7,2)*_r(n3+n5+n8,3)
	      + k4*_r(n2,1)*_r(n6,1)*_r(n1+n4+n7,3)*_r(n3+n5+n8,3)
	      - k4*_r(n2+n6,2)*_r(n1+n4+n7,3)*_r(n3+n5+n8,3)
	      + k4*_r(n1,1)*_r(n6,1)*_r(n2+n4+n7,3)*_r(n3+n5+n8,3)
	      - k4*_r(n1+n6,2)*_r(n2+n4+n7,3)*_r(n3+n5+n8,3)
	      - k12*_r(n6,1)*_r(n1+n2+n4+n7,4)*_r(n3+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n6+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n6+n7,2)*_r(n3+n5+n8,3)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n6+n7,2)*_r(n3+n5+n8,3)
	      - k4*_r(n1+n2+n4,3)*_r(n6+n7,2)*_r(n3+n5+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n1+n6+n7,3)*_r(n3+n5+n8,3)
	      - k4*_r(n2+n4,2)*_r(n1+n6+n7,3)*_r(n3+n5+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n2+n6+n7,3)*_r(n3+n5+n8,3)
	      - k4*_r(n1+n4,2)*_r(n2+n6+n7,3)*_r(n3+n5+n8,3)
	      - k12*_r(n4,1)*_r(n1+n2+n6+n7,4)*_r(n3+n5+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n4+n6+n7,3)*_r(n3+n5+n8,3)
	      - k4*_r(n1+n2,2)*_r(n4+n6+n7,3)*_r(n3+n5+n8,3)
	      - k12*_r(n2,1)*_r(n1+n4+n6+n7,4)*_r(n3+n5+n8,3)
	      - k12*_r(n1,1)*_r(n2+n4+n6+n7,4)*_r(n3+n5+n8,3)
	      + k48*_r(n1+n2+n4+n6+n7,5)*_r(n3+n5+n8,3)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n5+n8,4)
	      + k6*_r(n2+n4,2)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n5+n8,4)
	      + k6*_r(n4,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n3+n5+n8,4)
	      + k6*_r(n2,1)*_r(n4+n6,2)*_r(n7,1)*_r(n1+n3+n5+n8,4)
	      - k12*_r(n2+n4+n6,3)*_r(n7,1)*_r(n1+n3+n5+n8,4)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n3+n5+n8,4)
	      - k6*_r(n4+n6,2)*_r(n2+n7,2)*_r(n1+n3+n5+n8,4)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n4+n7,2)*_r(n1+n3+n5+n8,4)
	      - k6*_r(n2+n6,2)*_r(n4+n7,2)*_r(n1+n3+n5+n8,4)
	      - k12*_r(n6,1)*_r(n2+n4+n7,3)*_r(n1+n3+n5+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n6+n7,2)*_r(n1+n3+n5+n8,4)
	      - k6*_r(n2+n4,2)*_r(n6+n7,2)*_r(n1+n3+n5+n8,4)
	      - k12*_r(n4,1)*_r(n2+n6+n7,3)*_r(n1+n3+n5+n8,4)
	      - k12*_r(n2,1)*_r(n4+n6+n7,3)*_r(n1+n3+n5+n8,4)
	      + k36*_r(n2+n4+n6+n7,4)*_r(n1+n3+n5+n8,4)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n5+n8,4)
	      + k6*_r(n1+n4,2)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n5+n8,4)
	      + k6*_r(n4,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n3+n5+n8,4)
	      + k6*_r(n1,1)*_r(n4+n6,2)*_r(n7,1)*_r(n2+n3+n5+n8,4)
	      - k12*_r(n1+n4+n6,3)*_r(n7,1)*_r(n2+n3+n5+n8,4)
	      + k6*_r(n4,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n3+n5+n8,4)
	      - k6*_r(n4+n6,2)*_r(n1+n7,2)*_r(n2+n3+n5+n8,4)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n4+n7,2)*_r(n2+n3+n5+n8,4)
	      - k6*_r(n1+n6,2)*_r(n4+n7,2)*_r(n2+n3+n5+n8,4)
	      - k12*_r(n6,1)*_r(n1+n4+n7,3)*_r(n2+n3+n5+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n6+n7,2)*_r(n2+n3+n5+n8,4)
	      - k6*_r(n1+n4,2)*_r(n6+n7,2)*_r(n2+n3+n5+n8,4)
	      - k12*_r(n4,1)*_r(n1+n6+n7,3)*_r(n2+n3+n5+n8,4)
	      - k12*_r(n1,1)*_r(n4+n6+n7,3)*_r(n2+n3+n5+n8,4)
	      + k36*_r(n1+n4+n6+n7,4)*_r(n2+n3+n5+n8,4)
	      + k24*_r(n4,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n3+n5+n8,5)
	      - k24*_r(n4+n6,2)*_r(n7,1)*_r(n1+n2+n3+n5+n8,5)
	      - k24*_r(n6,1)*_r(n4+n7,2)*_r(n1+n2+n3+n5+n8,5)
	      - k24*_r(n4,1)*_r(n6+n7,2)*_r(n1+n2+n3+n5+n8,5)
	      + k48*_r(n4+n6+n7,3)*_r(n1+n2+n3+n5+n8,5)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n6,1)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n6,1)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n6,1)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k4*_r(n1+n2+n3,3)*_r(n6,1)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k2*_r(n2+n3,2)*_r(n1+n6,2)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k2*_r(n1+n3,2)*_r(n2+n6,2)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k4*_r(n3,1)*_r(n1+n2+n6,3)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3+n6,2)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k4*_r(n2,1)*_r(n1+n3+n6,3)*_r(n7,1)*_r(n4+n5+n8,3)
	      + k4*_r(n1,1)*_r(n2+n3+n6,3)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k12*_r(n1+n2+n3+n6,4)*_r(n7,1)*_r(n4+n5+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n2+n3,2)*_r(n6,1)*_r(n1+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n3,1)*_r(n2+n6,2)*_r(n1+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n2,1)*_r(n3+n6,2)*_r(n1+n7,2)*_r(n4+n5+n8,3)
	      - k4*_r(n2+n3+n6,3)*_r(n1+n7,2)*_r(n4+n5+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n1+n3,2)*_r(n6,1)*_r(n2+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n3,1)*_r(n1+n6,2)*_r(n2+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n1,1)*_r(n3+n6,2)*_r(n2+n7,2)*_r(n4+n5+n8,3)
	      - k4*_r(n1+n3+n6,3)*_r(n2+n7,2)*_r(n4+n5+n8,3)
	      + k4*_r(n3,1)*_r(n6,1)*_r(n1+n2+n7,3)*_r(n4+n5+n8,3)
	      - k4*_r(n3+n6,2)*_r(n1+n2+n7,3)*_r(n4+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n1+n2,2)*_r(n6,1)*_r(n3+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n2,1)*_r(n1+n6,2)*_r(n3+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n1,1)*_r(n2+n6,2)*_r(n3+n7,2)*_r(n4+n5+n8,3)
	      - k4*_r(n1+n2+n6,3)*_r(n3+n7,2)*_r(n4+n5+n8,3)
	      + k4*_r(n2,1)*_r(n6,1)*_r(n1+n3+n7,3)*_r(n4+n5+n8,3)
	      - k4*_r(n2+n6,2)*_r(n1+n3+n7,3)*_r(n4+n5+n8,3)
	      + k4*_r(n1,1)*_r(n6,1)*_r(n2+n3+n7,3)*_r(n4+n5+n8,3)
	      - k4*_r(n1+n6,2)*_r(n2+n3+n7,3)*_r(n4+n5+n8,3)
	      - k12*_r(n6,1)*_r(n1+n2+n3+n7,4)*_r(n4+n5+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n6+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n6+n7,2)*_r(n4+n5+n8,3)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n6+n7,2)*_r(n4+n5+n8,3)
	      - k4*_r(n1+n2+n3,3)*_r(n6+n7,2)*_r(n4+n5+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n1+n6+n7,3)*_r(n4+n5+n8,3)
	      - k4*_r(n2+n3,2)*_r(n1+n6+n7,3)*_r(n4+n5+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n2+n6+n7,3)*_r(n4+n5+n8,3)
	      - k4*_r(n1+n3,2)*_r(n2+n6+n7,3)*_r(n4+n5+n8,3)
	      - k12*_r(n3,1)*_r(n1+n2+n6+n7,4)*_r(n4+n5+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n3+n6+n7,3)*_r(n4+n5+n8,3)
	      - k4*_r(n1+n2,2)*_r(n3+n6+n7,3)*_r(n4+n5+n8,3)
	      - k12*_r(n2,1)*_r(n1+n3+n6+n7,4)*_r(n4+n5+n8,3)
	      - k12*_r(n1,1)*_r(n2+n3+n6+n7,4)*_r(n4+n5+n8,3)
	      + k48*_r(n1+n2+n3+n6+n7,5)*_r(n4+n5+n8,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n5+n8,4)
	      + k6*_r(n2+n3,2)*_r(n6,1)*_r(n7,1)*_r(n1+n4+n5+n8,4)
	      + k6*_r(n3,1)*_r(n2+n6,2)*_r(n7,1)*_r(n1+n4+n5+n8,4)
	      + k6*_r(n2,1)*_r(n3+n6,2)*_r(n7,1)*_r(n1+n4+n5+n8,4)
	      - k12*_r(n2+n3+n6,3)*_r(n7,1)*_r(n1+n4+n5+n8,4)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n2+n7,2)*_r(n1+n4+n5+n8,4)
	      - k6*_r(n3+n6,2)*_r(n2+n7,2)*_r(n1+n4+n5+n8,4)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n3+n7,2)*_r(n1+n4+n5+n8,4)
	      - k6*_r(n2+n6,2)*_r(n3+n7,2)*_r(n1+n4+n5+n8,4)
	      - k12*_r(n6,1)*_r(n2+n3+n7,3)*_r(n1+n4+n5+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n6+n7,2)*_r(n1+n4+n5+n8,4)
	      - k6*_r(n2+n3,2)*_r(n6+n7,2)*_r(n1+n4+n5+n8,4)
	      - k12*_r(n3,1)*_r(n2+n6+n7,3)*_r(n1+n4+n5+n8,4)
	      - k12*_r(n2,1)*_r(n3+n6+n7,3)*_r(n1+n4+n5+n8,4)
	      + k36*_r(n2+n3+n6+n7,4)*_r(n1+n4+n5+n8,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n5+n8,4)
	      + k6*_r(n1+n3,2)*_r(n6,1)*_r(n7,1)*_r(n2+n4+n5+n8,4)
	      + k6*_r(n3,1)*_r(n1+n6,2)*_r(n7,1)*_r(n2+n4+n5+n8,4)
	      + k6*_r(n1,1)*_r(n3+n6,2)*_r(n7,1)*_r(n2+n4+n5+n8,4)
	      - k12*_r(n1+n3+n6,3)*_r(n7,1)*_r(n2+n4+n5+n8,4)
	      + k6*_r(n3,1)*_r(n6,1)*_r(n1+n7,2)*_r(n2+n4+n5+n8,4)
	      - k6*_r(n3+n6,2)*_r(n1+n7,2)*_r(n2+n4+n5+n8,4)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n3+n7,2)*_r(n2+n4+n5+n8,4)
	      - k6*_r(n1+n6,2)*_r(n3+n7,2)*_r(n2+n4+n5+n8,4)
	      - k12*_r(n6,1)*_r(n1+n3+n7,3)*_r(n2+n4+n5+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n6+n7,2)*_r(n2+n4+n5+n8,4)
	      - k6*_r(n1+n3,2)*_r(n6+n7,2)*_r(n2+n4+n5+n8,4)
	      - k12*_r(n3,1)*_r(n1+n6+n7,3)*_r(n2+n4+n5+n8,4)
	      - k12*_r(n1,1)*_r(n3+n6+n7,3)*_r(n2+n4+n5+n8,4)
	      + k36*_r(n1+n3+n6+n7,4)*_r(n2+n4+n5+n8,4)
	      + k24*_r(n3,1)*_r(n6,1)*_r(n7,1)*_r(n1+n2+n4+n5+n8,5)
	      - k24*_r(n3+n6,2)*_r(n7,1)*_r(n1+n2+n4+n5+n8,5)
	      - k24*_r(n6,1)*_r(n3+n7,2)*_r(n1+n2+n4+n5+n8,5)
	      - k24*_r(n3,1)*_r(n6+n7,2)*_r(n1+n2+n4+n5+n8,5)
	      + k48*_r(n3+n6+n7,3)*_r(n1+n2+n4+n5+n8,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n5+n8,4)
	      + k6*_r(n1+n2,2)*_r(n6,1)*_r(n7,1)*_r(n3+n4+n5+n8,4)
	      + k6*_r(n2,1)*_r(n1+n6,2)*_r(n7,1)*_r(n3+n4+n5+n8,4)
	      + k6*_r(n1,1)*_r(n2+n6,2)*_r(n7,1)*_r(n3+n4+n5+n8,4)
	      - k12*_r(n1+n2+n6,3)*_r(n7,1)*_r(n3+n4+n5+n8,4)
	      + k6*_r(n2,1)*_r(n6,1)*_r(n1+n7,2)*_r(n3+n4+n5+n8,4)
	      - k6*_r(n2+n6,2)*_r(n1+n7,2)*_r(n3+n4+n5+n8,4)
	      + k6*_r(n1,1)*_r(n6,1)*_r(n2+n7,2)*_r(n3+n4+n5+n8,4)
	      - k6*_r(n1+n6,2)*_r(n2+n7,2)*_r(n3+n4+n5+n8,4)
	      - k12*_r(n6,1)*_r(n1+n2+n7,3)*_r(n3+n4+n5+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n6+n7,2)*_r(n3+n4+n5+n8,4)
	      - k6*_r(n1+n2,2)*_r(n6+n7,2)*_r(n3+n4+n5+n8,4)
	      - k12*_r(n2,1)*_r(n1+n6+n7,3)*_r(n3+n4+n5+n8,4)
	      - k12*_r(n1,1)*_r(n2+n6+n7,3)*_r(n3+n4+n5+n8,4)
	      + k36*_r(n1+n2+n6+n7,4)*_r(n3+n4+n5+n8,4)
	      + k24*_r(n2,1)*_r(n6,1)*_r(n7,1)*_r(n1+n3+n4+n5+n8,5)
	      - k24*_r(n2+n6,2)*_r(n7,1)*_r(n1+n3+n4+n5+n8,5)
	      - k24*_r(n6,1)*_r(n2+n7,2)*_r(n1+n3+n4+n5+n8,5)
	      - k24*_r(n2,1)*_r(n6+n7,2)*_r(n1+n3+n4+n5+n8,5)
	      + k48*_r(n2+n6+n7,3)*_r(n1+n3+n4+n5+n8,5)
	      + k24*_r(n1,1)*_r(n6,1)*_r(n7,1)*_r(n2+n3+n4+n5+n8,5)
	      - k24*_r(n1+n6,2)*_r(n7,1)*_r(n2+n3+n4+n5+n8,5)
	      - k24*_r(n6,1)*_r(n1+n7,2)*_r(n2+n3+n4+n5+n8,5)
	      - k24*_r(n1,1)*_r(n6+n7,2)*_r(n2+n3+n4+n5+n8,5)
	      + k48*_r(n1+n6+n7,3)*_r(n2+n3+n4+n5+n8,5)
	      - k120*_r(n6,1)*_r(n7,1)*_r(n1+n2+n3+n4+n5+n8,6)
	      + k120*_r(n6+n7,2)*_r(n1+n2+n3+n4+n5+n8,6)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n7,1)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n7,1)*_r(n6+n8,2)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n7,1)*_r(n6+n8,2)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n7,1)*_r(n6+n8,2)
	      - k24*_r(n1+n2+n3+n4+n5,5)*_r(n7,1)*_r(n6+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n7,2)*_r(n6+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n7,2)*_r(n6+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n7,2)*_r(n6+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n7,2)*_r(n6+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n7,2)*_r(n6+n8,2)
	      + _r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n7,2)*_r(n6+n8,2)
	      + _r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n7,2)*_r(n6+n8,2)
	      + k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n7,2)*_r(n6+n8,2)
	      + _r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n7,2)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n7,2)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n7,2)*_r(n6+n8,2)
	      - k6*_r(n2+n3+n4+n5,4)*_r(n1+n7,2)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n7,2)*_r(n6+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n7,2)*_r(n6+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n7,2)*_r(n6+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n7,2)*_r(n6+n8,2)
	      + _r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n7,2)*_r(n6+n8,2)
	      + _r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n7,2)*_r(n6+n8,2)
	      + k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n7,2)*_r(n6+n8,2)
	      + _r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n7,2)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n7,2)*_r(n6+n8,2)
	      - k6*_r(n1+n3+n4+n5,4)*_r(n2+n7,2)*_r(n6+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n7,3)*_r(n6+n8,2)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n7,3)*_r(n6+n8,2)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n7,3)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n7,3)*_r(n6+n8,2)
	      - k4*_r(n3+n4+n5,3)*_r(n1+n2+n7,3)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n7,2)*_r(n6+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n7,2)*_r(n6+n8,2)
	      + _r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n7,2)*_r(n6+n8,2)
	      + _r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n7,2)*_r(n6+n8,2)
	      + k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n7,2)*_r(n6+n8,2)
	      + _r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n7,2)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n7,2)*_r(n6+n8,2)
	      - k6*_r(n1+n2+n4+n5,4)*_r(n3+n7,2)*_r(n6+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n7,3)*_r(n6+n8,2)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n7,3)*_r(n6+n8,2)
	      - k4*_r(n2+n4+n5,3)*_r(n1+n3+n7,3)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n7,3)*_r(n6+n8,2)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n7,3)*_r(n6+n8,2)
	      - k4*_r(n1+n4+n5,3)*_r(n2+n3+n7,3)*_r(n6+n8,2)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n7,4)*_r(n6+n8,2)
	      - k6*_r(n4+n5,2)*_r(n1+n2+n3+n7,4)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n7,2)*_r(n6+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n7,2)*_r(n6+n8,2)
	      + _r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n7,2)*_r(n6+n8,2)
	      + _r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n7,2)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n7,2)*_r(n6+n8,2)
	      + _r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n7,2)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n7,2)*_r(n6+n8,2)
	      - k6*_r(n1+n2+n3+n5,4)*_r(n4+n7,2)*_r(n6+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n7,3)*_r(n6+n8,2)
	      - k4*_r(n2+n3+n5,3)*_r(n1+n4+n7,3)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n7,3)*_r(n6+n8,2)
	      - k4*_r(n1+n3+n5,3)*_r(n2+n4+n7,3)*_r(n6+n8,2)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n7,4)*_r(n6+n8,2)
	      - k6*_r(n3+n5,2)*_r(n1+n2+n4+n7,4)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n7,3)*_r(n6+n8,2)
	      - k4*_r(n1+n2+n5,3)*_r(n3+n4+n7,3)*_r(n6+n8,2)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n7,4)*_r(n6+n8,2)
	      - k6*_r(n2+n5,2)*_r(n1+n3+n4+n7,4)*_r(n6+n8,2)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n7,4)*_r(n6+n8,2)
	      - k6*_r(n1+n5,2)*_r(n2+n3+n4+n7,4)*_r(n6+n8,2)
	      - k24*_r(n5,1)*_r(n1+n2+n3+n4+n7,5)*_r(n6+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n7,2)*_r(n6+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n7,2)*_r(n6+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n7,2)*_r(n6+n8,2)
	      + _r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n7,2)*_r(n6+n8,2)
	      + _r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n7,2)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n7,2)*_r(n6+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n7,2)*_r(n6+n8,2)
	      + _r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n7,2)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n7,2)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n7,2)*_r(n6+n8,2)
	      - k6*_r(n1+n2+n3+n4,4)*_r(n5+n7,2)*_r(n6+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n7,3)*_r(n6+n8,2)
	      - k4*_r(n2+n3+n4,3)*_r(n1+n5+n7,3)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n7,3)*_r(n6+n8,2)
	      - k4*_r(n1+n3+n4,3)*_r(n2+n5+n7,3)*_r(n6+n8,2)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n7,4)*_r(n6+n8,2)
	      - k6*_r(n3+n4,2)*_r(n1+n2+n5+n7,4)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n7,3)*_r(n6+n8,2)
	      - k4*_r(n1+n2+n4,3)*_r(n3+n5+n7,3)*_r(n6+n8,2)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n7,4)*_r(n6+n8,2)
	      - k6*_r(n2+n4,2)*_r(n1+n3+n5+n7,4)*_r(n6+n8,2)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n7,4)*_r(n6+n8,2)
	      - k6*_r(n1+n4,2)*_r(n2+n3+n5+n7,4)*_r(n6+n8,2)
	      - k24*_r(n4,1)*_r(n1+n2+n3+n5+n7,5)*_r(n6+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n7,3)*_r(n6+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n7,3)*_r(n6+n8,2)
	      - k4*_r(n1+n2+n3,3)*_r(n4+n5+n7,3)*_r(n6+n8,2)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n7,4)*_r(n6+n8,2)
	      - k6*_r(n2+n3,2)*_r(n1+n4+n5+n7,4)*_r(n6+n8,2)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n7,4)*_r(n6+n8,2)
	      - k6*_r(n1+n3,2)*_r(n2+n4+n5+n7,4)*_r(n6+n8,2)
	      - k24*_r(n3,1)*_r(n1+n2+n4+n5+n7,5)*_r(n6+n8,2)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n7,4)*_r(n6+n8,2)
	      - k6*_r(n1+n2,2)*_r(n3+n4+n5+n7,4)*_r(n6+n8,2)
	      - k24*_r(n2,1)*_r(n1+n3+n4+n5+n7,5)*_r(n6+n8,2)
	      - k24*_r(n1,1)*_r(n2+n3+n4+n5+n7,5)*_r(n6+n8,2)
	      + k120*_r(n1+n2+n3+n4+n5+n7,6)*_r(n6+n8,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k4*_r(n2+n3+n4,3)*_r(n5,1)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k2*_r(n3+n4,2)*_r(n2+n5,2)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k2*_r(n2+n4,2)*_r(n3+n5,2)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k4*_r(n4,1)*_r(n2+n3+n5,3)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4+n5,2)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k4*_r(n3,1)*_r(n2+n4+n5,3)*_r(n7,1)*_r(n1+n6+n8,3)
	      + k4*_r(n2,1)*_r(n3+n4+n5,3)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k12*_r(n2+n3+n4+n5,4)*_r(n7,1)*_r(n1+n6+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n2+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n2+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n2+n7,2)*_r(n1+n6+n8,3)
	      - k4*_r(n3+n4+n5,3)*_r(n2+n7,2)*_r(n1+n6+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n3+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n3+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n3+n7,2)*_r(n1+n6+n8,3)
	      - k4*_r(n2+n4+n5,3)*_r(n3+n7,2)*_r(n1+n6+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n2+n3+n7,3)*_r(n1+n6+n8,3)
	      - k4*_r(n4+n5,2)*_r(n2+n3+n7,3)*_r(n1+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n4+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n4+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n4+n7,2)*_r(n1+n6+n8,3)
	      - k4*_r(n2+n3+n5,3)*_r(n4+n7,2)*_r(n1+n6+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n2+n4+n7,3)*_r(n1+n6+n8,3)
	      - k4*_r(n3+n5,2)*_r(n2+n4+n7,3)*_r(n1+n6+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n3+n4+n7,3)*_r(n1+n6+n8,3)
	      - k4*_r(n2+n5,2)*_r(n3+n4+n7,3)*_r(n1+n6+n8,3)
	      - k12*_r(n5,1)*_r(n2+n3+n4+n7,4)*_r(n1+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n5+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n5+n7,2)*_r(n1+n6+n8,3)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n5+n7,2)*_r(n1+n6+n8,3)
	      - k4*_r(n2+n3+n4,3)*_r(n5+n7,2)*_r(n1+n6+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n2+n5+n7,3)*_r(n1+n6+n8,3)
	      - k4*_r(n3+n4,2)*_r(n2+n5+n7,3)*_r(n1+n6+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n3+n5+n7,3)*_r(n1+n6+n8,3)
	      - k4*_r(n2+n4,2)*_r(n3+n5+n7,3)*_r(n1+n6+n8,3)
	      - k12*_r(n4,1)*_r(n2+n3+n5+n7,4)*_r(n1+n6+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n4+n5+n7,3)*_r(n1+n6+n8,3)
	      - k4*_r(n2+n3,2)*_r(n4+n5+n7,3)*_r(n1+n6+n8,3)
	      - k12*_r(n3,1)*_r(n2+n4+n5+n7,4)*_r(n1+n6+n8,3)
	      - k12*_r(n2,1)*_r(n3+n4+n5+n7,4)*_r(n1+n6+n8,3)
	      + k48*_r(n2+n3+n4+n5+n7,5)*_r(n1+n6+n8,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k4*_r(n1+n3+n4,3)*_r(n5,1)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k2*_r(n3+n4,2)*_r(n1+n5,2)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k2*_r(n1+n4,2)*_r(n3+n5,2)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k4*_r(n4,1)*_r(n1+n3+n5,3)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4+n5,2)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k4*_r(n3,1)*_r(n1+n4+n5,3)*_r(n7,1)*_r(n2+n6+n8,3)
	      + k4*_r(n1,1)*_r(n3+n4+n5,3)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k12*_r(n1+n3+n4+n5,4)*_r(n7,1)*_r(n2+n6+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n7,2)*_r(n2+n6+n8,3)
	      - k4*_r(n3+n4+n5,3)*_r(n1+n7,2)*_r(n2+n6+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n3+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n3+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n3+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n3+n7,2)*_r(n2+n6+n8,3)
	      - k4*_r(n1+n4+n5,3)*_r(n3+n7,2)*_r(n2+n6+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n1+n3+n7,3)*_r(n2+n6+n8,3)
	      - k4*_r(n4+n5,2)*_r(n1+n3+n7,3)*_r(n2+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n4+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n4+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n4+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n4+n7,2)*_r(n2+n6+n8,3)
	      - k4*_r(n1+n3+n5,3)*_r(n4+n7,2)*_r(n2+n6+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n1+n4+n7,3)*_r(n2+n6+n8,3)
	      - k4*_r(n3+n5,2)*_r(n1+n4+n7,3)*_r(n2+n6+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n3+n4+n7,3)*_r(n2+n6+n8,3)
	      - k4*_r(n1+n5,2)*_r(n3+n4+n7,3)*_r(n2+n6+n8,3)
	      - k12*_r(n5,1)*_r(n1+n3+n4+n7,4)*_r(n2+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n5+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n5+n7,2)*_r(n2+n6+n8,3)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n5+n7,2)*_r(n2+n6+n8,3)
	      - k4*_r(n1+n3+n4,3)*_r(n5+n7,2)*_r(n2+n6+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n1+n5+n7,3)*_r(n2+n6+n8,3)
	      - k4*_r(n3+n4,2)*_r(n1+n5+n7,3)*_r(n2+n6+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n3+n5+n7,3)*_r(n2+n6+n8,3)
	      - k4*_r(n1+n4,2)*_r(n3+n5+n7,3)*_r(n2+n6+n8,3)
	      - k12*_r(n4,1)*_r(n1+n3+n5+n7,4)*_r(n2+n6+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n4+n5+n7,3)*_r(n2+n6+n8,3)
	      - k4*_r(n1+n3,2)*_r(n4+n5+n7,3)*_r(n2+n6+n8,3)
	      - k12*_r(n3,1)*_r(n1+n4+n5+n7,4)*_r(n2+n6+n8,3)
	      - k12*_r(n1,1)*_r(n3+n4+n5+n7,4)*_r(n2+n6+n8,3)
	      + k48*_r(n1+n3+n4+n5+n7,5)*_r(n2+n6+n8,3)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n1+n2+n6+n8,4)
	      + k6*_r(n3+n4,2)*_r(n5,1)*_r(n7,1)*_r(n1+n2+n6+n8,4)
	      + k6*_r(n4,1)*_r(n3+n5,2)*_r(n7,1)*_r(n1+n2+n6+n8,4)
	      + k6*_r(n3,1)*_r(n4+n5,2)*_r(n7,1)*_r(n1+n2+n6+n8,4)
	      - k12*_r(n3+n4+n5,3)*_r(n7,1)*_r(n1+n2+n6+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n3+n7,2)*_r(n1+n2+n6+n8,4)
	      - k6*_r(n4+n5,2)*_r(n3+n7,2)*_r(n1+n2+n6+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n4+n7,2)*_r(n1+n2+n6+n8,4)
	      - k6*_r(n3+n5,2)*_r(n4+n7,2)*_r(n1+n2+n6+n8,4)
	      - k12*_r(n5,1)*_r(n3+n4+n7,3)*_r(n1+n2+n6+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n5+n7,2)*_r(n1+n2+n6+n8,4)
	      - k6*_r(n3+n4,2)*_r(n5+n7,2)*_r(n1+n2+n6+n8,4)
	      - k12*_r(n4,1)*_r(n3+n5+n7,3)*_r(n1+n2+n6+n8,4)
	      - k12*_r(n3,1)*_r(n4+n5+n7,3)*_r(n1+n2+n6+n8,4)
	      + k36*_r(n3+n4+n5+n7,4)*_r(n1+n2+n6+n8,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k4*_r(n1+n2+n4,3)*_r(n5,1)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k2*_r(n2+n4,2)*_r(n1+n5,2)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k2*_r(n1+n4,2)*_r(n2+n5,2)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k4*_r(n4,1)*_r(n1+n2+n5,3)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4+n5,2)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k4*_r(n2,1)*_r(n1+n4+n5,3)*_r(n7,1)*_r(n3+n6+n8,3)
	      + k4*_r(n1,1)*_r(n2+n4+n5,3)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k12*_r(n1+n2+n4+n5,4)*_r(n7,1)*_r(n3+n6+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n7,2)*_r(n3+n6+n8,3)
	      - k4*_r(n2+n4+n5,3)*_r(n1+n7,2)*_r(n3+n6+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n7,2)*_r(n3+n6+n8,3)
	      - k4*_r(n1+n4+n5,3)*_r(n2+n7,2)*_r(n3+n6+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n1+n2+n7,3)*_r(n3+n6+n8,3)
	      - k4*_r(n4+n5,2)*_r(n1+n2+n7,3)*_r(n3+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n4+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n4+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n4+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n4+n7,2)*_r(n3+n6+n8,3)
	      - k4*_r(n1+n2+n5,3)*_r(n4+n7,2)*_r(n3+n6+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n1+n4+n7,3)*_r(n3+n6+n8,3)
	      - k4*_r(n2+n5,2)*_r(n1+n4+n7,3)*_r(n3+n6+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n2+n4+n7,3)*_r(n3+n6+n8,3)
	      - k4*_r(n1+n5,2)*_r(n2+n4+n7,3)*_r(n3+n6+n8,3)
	      - k12*_r(n5,1)*_r(n1+n2+n4+n7,4)*_r(n3+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n5+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n5+n7,2)*_r(n3+n6+n8,3)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n5+n7,2)*_r(n3+n6+n8,3)
	      - k4*_r(n1+n2+n4,3)*_r(n5+n7,2)*_r(n3+n6+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n1+n5+n7,3)*_r(n3+n6+n8,3)
	      - k4*_r(n2+n4,2)*_r(n1+n5+n7,3)*_r(n3+n6+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n2+n5+n7,3)*_r(n3+n6+n8,3)
	      - k4*_r(n1+n4,2)*_r(n2+n5+n7,3)*_r(n3+n6+n8,3)
	      - k12*_r(n4,1)*_r(n1+n2+n5+n7,4)*_r(n3+n6+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n4+n5+n7,3)*_r(n3+n6+n8,3)
	      - k4*_r(n1+n2,2)*_r(n4+n5+n7,3)*_r(n3+n6+n8,3)
	      - k12*_r(n2,1)*_r(n1+n4+n5+n7,4)*_r(n3+n6+n8,3)
	      - k12*_r(n1,1)*_r(n2+n4+n5+n7,4)*_r(n3+n6+n8,3)
	      + k48*_r(n1+n2+n4+n5+n7,5)*_r(n3+n6+n8,3)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n1+n3+n6+n8,4)
	      + k6*_r(n2+n4,2)*_r(n5,1)*_r(n7,1)*_r(n1+n3+n6+n8,4)
	      + k6*_r(n4,1)*_r(n2+n5,2)*_r(n7,1)*_r(n1+n3+n6+n8,4)
	      + k6*_r(n2,1)*_r(n4+n5,2)*_r(n7,1)*_r(n1+n3+n6+n8,4)
	      - k12*_r(n2+n4+n5,3)*_r(n7,1)*_r(n1+n3+n6+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n2+n7,2)*_r(n1+n3+n6+n8,4)
	      - k6*_r(n4+n5,2)*_r(n2+n7,2)*_r(n1+n3+n6+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n4+n7,2)*_r(n1+n3+n6+n8,4)
	      - k6*_r(n2+n5,2)*_r(n4+n7,2)*_r(n1+n3+n6+n8,4)
	      - k12*_r(n5,1)*_r(n2+n4+n7,3)*_r(n1+n3+n6+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n5+n7,2)*_r(n1+n3+n6+n8,4)
	      - k6*_r(n2+n4,2)*_r(n5+n7,2)*_r(n1+n3+n6+n8,4)
	      - k12*_r(n4,1)*_r(n2+n5+n7,3)*_r(n1+n3+n6+n8,4)
	      - k12*_r(n2,1)*_r(n4+n5+n7,3)*_r(n1+n3+n6+n8,4)
	      + k36*_r(n2+n4+n5+n7,4)*_r(n1+n3+n6+n8,4)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n2+n3+n6+n8,4)
	      + k6*_r(n1+n4,2)*_r(n5,1)*_r(n7,1)*_r(n2+n3+n6+n8,4)
	      + k6*_r(n4,1)*_r(n1+n5,2)*_r(n7,1)*_r(n2+n3+n6+n8,4)
	      + k6*_r(n1,1)*_r(n4+n5,2)*_r(n7,1)*_r(n2+n3+n6+n8,4)
	      - k12*_r(n1+n4+n5,3)*_r(n7,1)*_r(n2+n3+n6+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n1+n7,2)*_r(n2+n3+n6+n8,4)
	      - k6*_r(n4+n5,2)*_r(n1+n7,2)*_r(n2+n3+n6+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n4+n7,2)*_r(n2+n3+n6+n8,4)
	      - k6*_r(n1+n5,2)*_r(n4+n7,2)*_r(n2+n3+n6+n8,4)
	      - k12*_r(n5,1)*_r(n1+n4+n7,3)*_r(n2+n3+n6+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n5+n7,2)*_r(n2+n3+n6+n8,4)
	      - k6*_r(n1+n4,2)*_r(n5+n7,2)*_r(n2+n3+n6+n8,4)
	      - k12*_r(n4,1)*_r(n1+n5+n7,3)*_r(n2+n3+n6+n8,4)
	      - k12*_r(n1,1)*_r(n4+n5+n7,3)*_r(n2+n3+n6+n8,4)
	      + k36*_r(n1+n4+n5+n7,4)*_r(n2+n3+n6+n8,4)
	      + k24*_r(n4,1)*_r(n5,1)*_r(n7,1)*_r(n1+n2+n3+n6+n8,5)
	      - k24*_r(n4+n5,2)*_r(n7,1)*_r(n1+n2+n3+n6+n8,5)
	      - k24*_r(n5,1)*_r(n4+n7,2)*_r(n1+n2+n3+n6+n8,5)
	      - k24*_r(n4,1)*_r(n5+n7,2)*_r(n1+n2+n3+n6+n8,5)
	      + k48*_r(n4+n5+n7,3)*_r(n1+n2+n3+n6+n8,5)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n7,1)*_r(n4+n6+n8,3));
    }
    /** 
     * Fourth part of the 8-particle correlator 
     * 
     * @param n1 Harmonic
     * @param n2 Harmonic 
     * @param n3 Harmonic 
     * @param n4 Harmonic 
     * @param n5 Harmonic 
     * @param n6 Harmonic 
     * @param n7 Harmonic 
     * @param n8 Harmonic 
     * 
     * @return Fourth part 
     */
    Complex c8P4(const HarmonicElement n1,
		 const HarmonicElement n2,
		 const HarmonicElement n3,
		 const HarmonicElement n4,
		 const HarmonicElement n5,
		 const HarmonicElement n6,
		 const HarmonicElement n7,
		 const HarmonicElement n8) const
    {
      const Real k2   = 2;
      const Real k4   = 4;
      const Real k6   = 6;
      const Real k12  = 12;
      const Real k24  = 24;
      const Real k36  = 36;
      const Real k48  = 48;
      const Real k120 = 120;
      const Real k720 = 720;
      const Real k5040 = 5040;
      return (k2*_r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k4*_r(n1+n2+n3,3)*_r(n5,1)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n1+n5,2)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n2+n5,2)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k4*_r(n3,1)*_r(n1+n2+n5,3)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3+n5,2)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k4*_r(n2,1)*_r(n1+n3+n5,3)*_r(n7,1)*_r(n4+n6+n8,3)
	      + k4*_r(n1,1)*_r(n2+n3+n5,3)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k12*_r(n1+n2+n3+n5,4)*_r(n7,1)*_r(n4+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n7,2)*_r(n4+n6+n8,3)
	      - k4*_r(n2+n3+n5,3)*_r(n1+n7,2)*_r(n4+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n7,2)*_r(n4+n6+n8,3)
	      - k4*_r(n1+n3+n5,3)*_r(n2+n7,2)*_r(n4+n6+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n1+n2+n7,3)*_r(n4+n6+n8,3)
	      - k4*_r(n3+n5,2)*_r(n1+n2+n7,3)*_r(n4+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n7,2)*_r(n4+n6+n8,3)
	      - k4*_r(n1+n2+n5,3)*_r(n3+n7,2)*_r(n4+n6+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n1+n3+n7,3)*_r(n4+n6+n8,3)
	      - k4*_r(n2+n5,2)*_r(n1+n3+n7,3)*_r(n4+n6+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n2+n3+n7,3)*_r(n4+n6+n8,3)
	      - k4*_r(n1+n5,2)*_r(n2+n3+n7,3)*_r(n4+n6+n8,3)
	      - k12*_r(n5,1)*_r(n1+n2+n3+n7,4)*_r(n4+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n5+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n5+n7,2)*_r(n4+n6+n8,3)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n5+n7,2)*_r(n4+n6+n8,3)
	      - k4*_r(n1+n2+n3,3)*_r(n5+n7,2)*_r(n4+n6+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n1+n5+n7,3)*_r(n4+n6+n8,3)
	      - k4*_r(n2+n3,2)*_r(n1+n5+n7,3)*_r(n4+n6+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n2+n5+n7,3)*_r(n4+n6+n8,3)
	      - k4*_r(n1+n3,2)*_r(n2+n5+n7,3)*_r(n4+n6+n8,3)
	      - k12*_r(n3,1)*_r(n1+n2+n5+n7,4)*_r(n4+n6+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n3+n5+n7,3)*_r(n4+n6+n8,3)
	      - k4*_r(n1+n2,2)*_r(n3+n5+n7,3)*_r(n4+n6+n8,3)
	      - k12*_r(n2,1)*_r(n1+n3+n5+n7,4)*_r(n4+n6+n8,3)
	      - k12*_r(n1,1)*_r(n2+n3+n5+n7,4)*_r(n4+n6+n8,3)
	      + k48*_r(n1+n2+n3+n5+n7,5)*_r(n4+n6+n8,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n7,1)*_r(n1+n4+n6+n8,4)
	      + k6*_r(n2+n3,2)*_r(n5,1)*_r(n7,1)*_r(n1+n4+n6+n8,4)
	      + k6*_r(n3,1)*_r(n2+n5,2)*_r(n7,1)*_r(n1+n4+n6+n8,4)
	      + k6*_r(n2,1)*_r(n3+n5,2)*_r(n7,1)*_r(n1+n4+n6+n8,4)
	      - k12*_r(n2+n3+n5,3)*_r(n7,1)*_r(n1+n4+n6+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n2+n7,2)*_r(n1+n4+n6+n8,4)
	      - k6*_r(n3+n5,2)*_r(n2+n7,2)*_r(n1+n4+n6+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n3+n7,2)*_r(n1+n4+n6+n8,4)
	      - k6*_r(n2+n5,2)*_r(n3+n7,2)*_r(n1+n4+n6+n8,4)
	      - k12*_r(n5,1)*_r(n2+n3+n7,3)*_r(n1+n4+n6+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n5+n7,2)*_r(n1+n4+n6+n8,4)
	      - k6*_r(n2+n3,2)*_r(n5+n7,2)*_r(n1+n4+n6+n8,4)
	      - k12*_r(n3,1)*_r(n2+n5+n7,3)*_r(n1+n4+n6+n8,4)
	      - k12*_r(n2,1)*_r(n3+n5+n7,3)*_r(n1+n4+n6+n8,4)
	      + k36*_r(n2+n3+n5+n7,4)*_r(n1+n4+n6+n8,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n7,1)*_r(n2+n4+n6+n8,4)
	      + k6*_r(n1+n3,2)*_r(n5,1)*_r(n7,1)*_r(n2+n4+n6+n8,4)
	      + k6*_r(n3,1)*_r(n1+n5,2)*_r(n7,1)*_r(n2+n4+n6+n8,4)
	      + k6*_r(n1,1)*_r(n3+n5,2)*_r(n7,1)*_r(n2+n4+n6+n8,4)
	      - k12*_r(n1+n3+n5,3)*_r(n7,1)*_r(n2+n4+n6+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n1+n7,2)*_r(n2+n4+n6+n8,4)
	      - k6*_r(n3+n5,2)*_r(n1+n7,2)*_r(n2+n4+n6+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n3+n7,2)*_r(n2+n4+n6+n8,4)
	      - k6*_r(n1+n5,2)*_r(n3+n7,2)*_r(n2+n4+n6+n8,4)
	      - k12*_r(n5,1)*_r(n1+n3+n7,3)*_r(n2+n4+n6+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n5+n7,2)*_r(n2+n4+n6+n8,4)
	      - k6*_r(n1+n3,2)*_r(n5+n7,2)*_r(n2+n4+n6+n8,4)
	      - k12*_r(n3,1)*_r(n1+n5+n7,3)*_r(n2+n4+n6+n8,4)
	      - k12*_r(n1,1)*_r(n3+n5+n7,3)*_r(n2+n4+n6+n8,4)
	      + k36*_r(n1+n3+n5+n7,4)*_r(n2+n4+n6+n8,4)
	      + k24*_r(n3,1)*_r(n5,1)*_r(n7,1)*_r(n1+n2+n4+n6+n8,5)
	      - k24*_r(n3+n5,2)*_r(n7,1)*_r(n1+n2+n4+n6+n8,5)
	      - k24*_r(n5,1)*_r(n3+n7,2)*_r(n1+n2+n4+n6+n8,5)
	      - k24*_r(n3,1)*_r(n5+n7,2)*_r(n1+n2+n4+n6+n8,5)
	      + k48*_r(n3+n5+n7,3)*_r(n1+n2+n4+n6+n8,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n7,1)*_r(n3+n4+n6+n8,4)
	      + k6*_r(n1+n2,2)*_r(n5,1)*_r(n7,1)*_r(n3+n4+n6+n8,4)
	      + k6*_r(n2,1)*_r(n1+n5,2)*_r(n7,1)*_r(n3+n4+n6+n8,4)
	      + k6*_r(n1,1)*_r(n2+n5,2)*_r(n7,1)*_r(n3+n4+n6+n8,4)
	      - k12*_r(n1+n2+n5,3)*_r(n7,1)*_r(n3+n4+n6+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n1+n7,2)*_r(n3+n4+n6+n8,4)
	      - k6*_r(n2+n5,2)*_r(n1+n7,2)*_r(n3+n4+n6+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n2+n7,2)*_r(n3+n4+n6+n8,4)
	      - k6*_r(n1+n5,2)*_r(n2+n7,2)*_r(n3+n4+n6+n8,4)
	      - k12*_r(n5,1)*_r(n1+n2+n7,3)*_r(n3+n4+n6+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n5+n7,2)*_r(n3+n4+n6+n8,4)
	      - k6*_r(n1+n2,2)*_r(n5+n7,2)*_r(n3+n4+n6+n8,4)
	      - k12*_r(n2,1)*_r(n1+n5+n7,3)*_r(n3+n4+n6+n8,4)
	      - k12*_r(n1,1)*_r(n2+n5+n7,3)*_r(n3+n4+n6+n8,4)
	      + k36*_r(n1+n2+n5+n7,4)*_r(n3+n4+n6+n8,4)
	      + k24*_r(n2,1)*_r(n5,1)*_r(n7,1)*_r(n1+n3+n4+n6+n8,5)
	      - k24*_r(n2+n5,2)*_r(n7,1)*_r(n1+n3+n4+n6+n8,5)
	      - k24*_r(n5,1)*_r(n2+n7,2)*_r(n1+n3+n4+n6+n8,5)
	      - k24*_r(n2,1)*_r(n5+n7,2)*_r(n1+n3+n4+n6+n8,5)
	      + k48*_r(n2+n5+n7,3)*_r(n1+n3+n4+n6+n8,5)
	      + k24*_r(n1,1)*_r(n5,1)*_r(n7,1)*_r(n2+n3+n4+n6+n8,5)
	      - k24*_r(n1+n5,2)*_r(n7,1)*_r(n2+n3+n4+n6+n8,5)
	      - k24*_r(n5,1)*_r(n1+n7,2)*_r(n2+n3+n4+n6+n8,5)
	      - k24*_r(n1,1)*_r(n5+n7,2)*_r(n2+n3+n4+n6+n8,5)
	      + k48*_r(n1+n5+n7,3)*_r(n2+n3+n4+n6+n8,5)
	      - k120*_r(n5,1)*_r(n7,1)*_r(n1+n2+n3+n4+n6+n8,6)
	      + k120*_r(n5+n7,2)*_r(n1+n2+n3+n4+n6+n8,6)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k4*_r(n1+n2+n3,3)*_r(n4,1)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n1+n4,2)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n2+n4,2)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k4*_r(n3,1)*_r(n1+n2+n4,3)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3+n4,2)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k4*_r(n2,1)*_r(n1+n3+n4,3)*_r(n7,1)*_r(n5+n6+n8,3)
	      + k4*_r(n1,1)*_r(n2+n3+n4,3)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k12*_r(n1+n2+n3+n4,4)*_r(n7,1)*_r(n5+n6+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n7,2)*_r(n5+n6+n8,3)
	      - k4*_r(n2+n3+n4,3)*_r(n1+n7,2)*_r(n5+n6+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n7,2)*_r(n5+n6+n8,3)
	      - k4*_r(n1+n3+n4,3)*_r(n2+n7,2)*_r(n5+n6+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n1+n2+n7,3)*_r(n5+n6+n8,3)
	      - k4*_r(n3+n4,2)*_r(n1+n2+n7,3)*_r(n5+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n7,2)*_r(n5+n6+n8,3)
	      - k4*_r(n1+n2+n4,3)*_r(n3+n7,2)*_r(n5+n6+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n1+n3+n7,3)*_r(n5+n6+n8,3)
	      - k4*_r(n2+n4,2)*_r(n1+n3+n7,3)*_r(n5+n6+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n2+n3+n7,3)*_r(n5+n6+n8,3)
	      - k4*_r(n1+n4,2)*_r(n2+n3+n7,3)*_r(n5+n6+n8,3)
	      - k12*_r(n4,1)*_r(n1+n2+n3+n7,4)*_r(n5+n6+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n7,2)*_r(n5+n6+n8,3)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n7,2)*_r(n5+n6+n8,3)
	      - k4*_r(n1+n2+n3,3)*_r(n4+n7,2)*_r(n5+n6+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n1+n4+n7,3)*_r(n5+n6+n8,3)
	      - k4*_r(n2+n3,2)*_r(n1+n4+n7,3)*_r(n5+n6+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n2+n4+n7,3)*_r(n5+n6+n8,3)
	      - k4*_r(n1+n3,2)*_r(n2+n4+n7,3)*_r(n5+n6+n8,3)
	      - k12*_r(n3,1)*_r(n1+n2+n4+n7,4)*_r(n5+n6+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n3+n4+n7,3)*_r(n5+n6+n8,3)
	      - k4*_r(n1+n2,2)*_r(n3+n4+n7,3)*_r(n5+n6+n8,3)
	      - k12*_r(n2,1)*_r(n1+n3+n4+n7,4)*_r(n5+n6+n8,3)
	      - k12*_r(n1,1)*_r(n2+n3+n4+n7,4)*_r(n5+n6+n8,3)
	      + k48*_r(n1+n2+n3+n4+n7,5)*_r(n5+n6+n8,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n7,1)*_r(n1+n5+n6+n8,4)
	      + k6*_r(n2+n3,2)*_r(n4,1)*_r(n7,1)*_r(n1+n5+n6+n8,4)
	      + k6*_r(n3,1)*_r(n2+n4,2)*_r(n7,1)*_r(n1+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n3+n4,2)*_r(n7,1)*_r(n1+n5+n6+n8,4)
	      - k12*_r(n2+n3+n4,3)*_r(n7,1)*_r(n1+n5+n6+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n2+n7,2)*_r(n1+n5+n6+n8,4)
	      - k6*_r(n3+n4,2)*_r(n2+n7,2)*_r(n1+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n3+n7,2)*_r(n1+n5+n6+n8,4)
	      - k6*_r(n2+n4,2)*_r(n3+n7,2)*_r(n1+n5+n6+n8,4)
	      - k12*_r(n4,1)*_r(n2+n3+n7,3)*_r(n1+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n4+n7,2)*_r(n1+n5+n6+n8,4)
	      - k6*_r(n2+n3,2)*_r(n4+n7,2)*_r(n1+n5+n6+n8,4)
	      - k12*_r(n3,1)*_r(n2+n4+n7,3)*_r(n1+n5+n6+n8,4)
	      - k12*_r(n2,1)*_r(n3+n4+n7,3)*_r(n1+n5+n6+n8,4)
	      + k36*_r(n2+n3+n4+n7,4)*_r(n1+n5+n6+n8,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n7,1)*_r(n2+n5+n6+n8,4)
	      + k6*_r(n1+n3,2)*_r(n4,1)*_r(n7,1)*_r(n2+n5+n6+n8,4)
	      + k6*_r(n3,1)*_r(n1+n4,2)*_r(n7,1)*_r(n2+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n3+n4,2)*_r(n7,1)*_r(n2+n5+n6+n8,4)
	      - k12*_r(n1+n3+n4,3)*_r(n7,1)*_r(n2+n5+n6+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n7,2)*_r(n2+n5+n6+n8,4)
	      - k6*_r(n3+n4,2)*_r(n1+n7,2)*_r(n2+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n3+n7,2)*_r(n2+n5+n6+n8,4)
	      - k6*_r(n1+n4,2)*_r(n3+n7,2)*_r(n2+n5+n6+n8,4)
	      - k12*_r(n4,1)*_r(n1+n3+n7,3)*_r(n2+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n4+n7,2)*_r(n2+n5+n6+n8,4)
	      - k6*_r(n1+n3,2)*_r(n4+n7,2)*_r(n2+n5+n6+n8,4)
	      - k12*_r(n3,1)*_r(n1+n4+n7,3)*_r(n2+n5+n6+n8,4)
	      - k12*_r(n1,1)*_r(n3+n4+n7,3)*_r(n2+n5+n6+n8,4)
	      + k36*_r(n1+n3+n4+n7,4)*_r(n2+n5+n6+n8,4)
	      + k24*_r(n3,1)*_r(n4,1)*_r(n7,1)*_r(n1+n2+n5+n6+n8,5)
	      - k24*_r(n3+n4,2)*_r(n7,1)*_r(n1+n2+n5+n6+n8,5)
	      - k24*_r(n4,1)*_r(n3+n7,2)*_r(n1+n2+n5+n6+n8,5)
	      - k24*_r(n3,1)*_r(n4+n7,2)*_r(n1+n2+n5+n6+n8,5)
	      + k48*_r(n3+n4+n7,3)*_r(n1+n2+n5+n6+n8,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n7,1)*_r(n3+n5+n6+n8,4)
	      + k6*_r(n1+n2,2)*_r(n4,1)*_r(n7,1)*_r(n3+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n1+n4,2)*_r(n7,1)*_r(n3+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n2+n4,2)*_r(n7,1)*_r(n3+n5+n6+n8,4)
	      - k12*_r(n1+n2+n4,3)*_r(n7,1)*_r(n3+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n7,2)*_r(n3+n5+n6+n8,4)
	      - k6*_r(n2+n4,2)*_r(n1+n7,2)*_r(n3+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n7,2)*_r(n3+n5+n6+n8,4)
	      - k6*_r(n1+n4,2)*_r(n2+n7,2)*_r(n3+n5+n6+n8,4)
	      - k12*_r(n4,1)*_r(n1+n2+n7,3)*_r(n3+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n4+n7,2)*_r(n3+n5+n6+n8,4)
	      - k6*_r(n1+n2,2)*_r(n4+n7,2)*_r(n3+n5+n6+n8,4)
	      - k12*_r(n2,1)*_r(n1+n4+n7,3)*_r(n3+n5+n6+n8,4)
	      - k12*_r(n1,1)*_r(n2+n4+n7,3)*_r(n3+n5+n6+n8,4)
	      + k36*_r(n1+n2+n4+n7,4)*_r(n3+n5+n6+n8,4)
	      + k24*_r(n2,1)*_r(n4,1)*_r(n7,1)*_r(n1+n3+n5+n6+n8,5)
	      - k24*_r(n2+n4,2)*_r(n7,1)*_r(n1+n3+n5+n6+n8,5)
	      - k24*_r(n4,1)*_r(n2+n7,2)*_r(n1+n3+n5+n6+n8,5)
	      - k24*_r(n2,1)*_r(n4+n7,2)*_r(n1+n3+n5+n6+n8,5)
	      + k48*_r(n2+n4+n7,3)*_r(n1+n3+n5+n6+n8,5)
	      + k24*_r(n1,1)*_r(n4,1)*_r(n7,1)*_r(n2+n3+n5+n6+n8,5)
	      - k24*_r(n1+n4,2)*_r(n7,1)*_r(n2+n3+n5+n6+n8,5)
	      - k24*_r(n4,1)*_r(n1+n7,2)*_r(n2+n3+n5+n6+n8,5)
	      - k24*_r(n1,1)*_r(n4+n7,2)*_r(n2+n3+n5+n6+n8,5)
	      + k48*_r(n1+n4+n7,3)*_r(n2+n3+n5+n6+n8,5)
	      - k120*_r(n4,1)*_r(n7,1)*_r(n1+n2+n3+n5+n6+n8,6)
	      + k120*_r(n4+n7,2)*_r(n1+n2+n3+n5+n6+n8,6)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n7,1)*_r(n4+n5+n6+n8,4)
	      + k6*_r(n1+n2,2)*_r(n3,1)*_r(n7,1)*_r(n4+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n1+n3,2)*_r(n7,1)*_r(n4+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n2+n3,2)*_r(n7,1)*_r(n4+n5+n6+n8,4)
	      - k12*_r(n1+n2+n3,3)*_r(n7,1)*_r(n4+n5+n6+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n7,2)*_r(n4+n5+n6+n8,4)
	      - k6*_r(n2+n3,2)*_r(n1+n7,2)*_r(n4+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n7,2)*_r(n4+n5+n6+n8,4)
	      - k6*_r(n1+n3,2)*_r(n2+n7,2)*_r(n4+n5+n6+n8,4)
	      - k12*_r(n3,1)*_r(n1+n2+n7,3)*_r(n4+n5+n6+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n7,2)*_r(n4+n5+n6+n8,4)
	      - k6*_r(n1+n2,2)*_r(n3+n7,2)*_r(n4+n5+n6+n8,4)
	      - k12*_r(n2,1)*_r(n1+n3+n7,3)*_r(n4+n5+n6+n8,4)
	      - k12*_r(n1,1)*_r(n2+n3+n7,3)*_r(n4+n5+n6+n8,4)
	      + k36*_r(n1+n2+n3+n7,4)*_r(n4+n5+n6+n8,4)
	      + k24*_r(n2,1)*_r(n3,1)*_r(n7,1)*_r(n1+n4+n5+n6+n8,5)
	      - k24*_r(n2+n3,2)*_r(n7,1)*_r(n1+n4+n5+n6+n8,5)
	      - k24*_r(n3,1)*_r(n2+n7,2)*_r(n1+n4+n5+n6+n8,5)
	      - k24*_r(n2,1)*_r(n3+n7,2)*_r(n1+n4+n5+n6+n8,5)
	      + k48*_r(n2+n3+n7,3)*_r(n1+n4+n5+n6+n8,5)
	      + k24*_r(n1,1)*_r(n3,1)*_r(n7,1)*_r(n2+n4+n5+n6+n8,5)
	      - k24*_r(n1+n3,2)*_r(n7,1)*_r(n2+n4+n5+n6+n8,5)
	      - k24*_r(n3,1)*_r(n1+n7,2)*_r(n2+n4+n5+n6+n8,5)
	      - k24*_r(n1,1)*_r(n3+n7,2)*_r(n2+n4+n5+n6+n8,5)
	      + k48*_r(n1+n3+n7,3)*_r(n2+n4+n5+n6+n8,5)
	      - k120*_r(n3,1)*_r(n7,1)*_r(n1+n2+n4+n5+n6+n8,6)
	      + k120*_r(n3+n7,2)*_r(n1+n2+n4+n5+n6+n8,6)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n7,1)*_r(n3+n4+n5+n6+n8,5)
	      - k24*_r(n1+n2,2)*_r(n7,1)*_r(n3+n4+n5+n6+n8,5)
	      - k24*_r(n2,1)*_r(n1+n7,2)*_r(n3+n4+n5+n6+n8,5)
	      - k24*_r(n1,1)*_r(n2+n7,2)*_r(n3+n4+n5+n6+n8,5)
	      + k48*_r(n1+n2+n7,3)*_r(n3+n4+n5+n6+n8,5)
	      - k120*_r(n2,1)*_r(n7,1)*_r(n1+n3+n4+n5+n6+n8,6)
	      + k120*_r(n2+n7,2)*_r(n1+n3+n4+n5+n6+n8,6)
	      - k120*_r(n1,1)*_r(n7,1)*_r(n2+n3+n4+n5+n6+n8,6)
	      + k120*_r(n1+n7,2)*_r(n2+n3+n4+n5+n6+n8,6)
	      + k720*_r(n7,1)*_r(n1+n2+n3+n4+n5+n6+n8,7)
	      - _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + k6*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k6*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k6*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n7+n8,2)
	      + k6*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n7+n8,2)
	      + k6*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n7+n8,2)
	      - k24*_r(n1+n2+n3+n4+n5,5)*_r(n6,1)*_r(n7+n8,2)
	      + _r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7+n8,2)
	      - _r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n7+n8,2)
	      - _r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n7+n8,2)
	      + k2*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n6,2)*_r(n7+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7+n8,2)
	      + _r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7+n8,2)
	      + _r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n7+n8,2)
	      + k2*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7+n8,2)
	      + _r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n7+n8,2)
	      - k6*_r(n2+n3+n4+n5,4)*_r(n1+n6,2)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7+n8,2)
	      - _r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n7+n8,2)
	      - _r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n6,2)*_r(n7+n8,2)
	      - _r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7+n8,2)
	      + _r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7+n8,2)
	      + _r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n7+n8,2)
	      + k2*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7+n8,2)
	      + _r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n7+n8,2)
	      - k6*_r(n1+n3+n4+n5,4)*_r(n2+n6,2)*_r(n7+n8,2)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7+n8,2)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n7+n8,2)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n7+n8,2)
	      - k4*_r(n3+n4+n5,3)*_r(n1+n2+n6,3)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7+n8,2)
	      - _r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7+n8,2)
	      + _r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7+n8,2)
	      + _r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n7+n8,2)
	      + k2*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7+n8,2)
	      + _r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n7+n8,2)
	      - k6*_r(n1+n2+n4+n5,4)*_r(n3+n6,2)*_r(n7+n8,2)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n7+n8,2)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n7+n8,2)
	      - k4*_r(n2+n4+n5,3)*_r(n1+n3+n6,3)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n7+n8,2)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n7+n8,2)
	      - k4*_r(n1+n4+n5,3)*_r(n2+n3+n6,3)*_r(n7+n8,2)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n7+n8,2)
	      - k6*_r(n4+n5,2)*_r(n1+n2+n3+n6,4)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7+n8,2)
	      + _r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7+n8,2)
	      + _r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7+n8,2)
	      + _r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n7+n8,2)
	      - k6*_r(n1+n2+n3+n5,4)*_r(n4+n6,2)*_r(n7+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n7+n8,2)
	      - k4*_r(n2+n3+n5,3)*_r(n1+n4+n6,3)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n7+n8,2)
	      - k4*_r(n1+n3+n5,3)*_r(n2+n4+n6,3)*_r(n7+n8,2)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n7+n8,2)
	      - k6*_r(n3+n5,2)*_r(n1+n2+n4+n6,4)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n7+n8,2)
	      - k4*_r(n1+n2+n5,3)*_r(n3+n4+n6,3)*_r(n7+n8,2)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n7+n8,2)
	      - k6*_r(n2+n5,2)*_r(n1+n3+n4+n6,4)*_r(n7+n8,2)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n7+n8,2)
	      - k6*_r(n1+n5,2)*_r(n2+n3+n4+n6,4)*_r(n7+n8,2)
	      - k24*_r(n5,1)*_r(n1+n2+n3+n4+n6,5)*_r(n7+n8,2)
	      + _r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7+n8,2)
	      - _r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n6,2)*_r(n7+n8,2)
	      - _r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7+n8,2)
	      + _r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7+n8,2)
	      + _r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n7+n8,2)
	      - _r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7+n8,2)
	      + _r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n7+n8,2)
	      - k6*_r(n1+n2+n3+n4,4)*_r(n5+n6,2)*_r(n7+n8,2)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n7+n8,2)
	      - k4*_r(n2+n3+n4,3)*_r(n1+n5+n6,3)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n7+n8,2)
	      - k4*_r(n1+n3+n4,3)*_r(n2+n5+n6,3)*_r(n7+n8,2)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n7+n8,2)
	      - k6*_r(n3+n4,2)*_r(n1+n2+n5+n6,4)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n7+n8,2)
	      - k4*_r(n1+n2+n4,3)*_r(n3+n5+n6,3)*_r(n7+n8,2)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n7+n8,2)
	      - k6*_r(n2+n4,2)*_r(n1+n3+n5+n6,4)*_r(n7+n8,2)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n7+n8,2)
	      - k6*_r(n1+n4,2)*_r(n2+n3+n5+n6,4)*_r(n7+n8,2)
	      - k24*_r(n4,1)*_r(n1+n2+n3+n5+n6,5)*_r(n7+n8,2)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n7+n8,2)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n7+n8,2)
	      - k4*_r(n1+n2+n3,3)*_r(n4+n5+n6,3)*_r(n7+n8,2)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n7+n8,2)
	      - k6*_r(n2+n3,2)*_r(n1+n4+n5+n6,4)*_r(n7+n8,2)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n7+n8,2)
	      - k6*_r(n1+n3,2)*_r(n2+n4+n5+n6,4)*_r(n7+n8,2)
	      - k24*_r(n3,1)*_r(n1+n2+n4+n5+n6,5)*_r(n7+n8,2)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n7+n8,2)
	      - k6*_r(n1+n2,2)*_r(n3+n4+n5+n6,4)*_r(n7+n8,2)
	      - k24*_r(n2,1)*_r(n1+n3+n4+n5+n6,5)*_r(n7+n8,2)
	      - k24*_r(n1,1)*_r(n2+n3+n4+n5+n6,5)*_r(n7+n8,2)
	      + k120*_r(n1+n2+n3+n4+n5+n6,6)*_r(n7+n8,2)
	      + k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k4*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k2*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k2*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k4*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k4*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n7+n8,3)
	      + k4*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k12*_r(n2+n3+n4+n5,4)*_r(n6,1)*_r(n1+n7+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n7+n8,3)
	      - k4*_r(n3+n4+n5,3)*_r(n2+n6,2)*_r(n1+n7+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n7+n8,3)
	      - k4*_r(n2+n4+n5,3)*_r(n3+n6,2)*_r(n1+n7+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n7+n8,3)
	      - k4*_r(n4+n5,2)*_r(n2+n3+n6,3)*_r(n1+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n7+n8,3)
	      - k4*_r(n2+n3+n5,3)*_r(n4+n6,2)*_r(n1+n7+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n7+n8,3)
	      - k4*_r(n3+n5,2)*_r(n2+n4+n6,3)*_r(n1+n7+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n7+n8,3)
	      - k4*_r(n2+n5,2)*_r(n3+n4+n6,3)*_r(n1+n7+n8,3)
	      - k12*_r(n5,1)*_r(n2+n3+n4+n6,4)*_r(n1+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n7+n8,3)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n7+n8,3)
	      - k4*_r(n2+n3+n4,3)*_r(n5+n6,2)*_r(n1+n7+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n7+n8,3)
	      - k4*_r(n3+n4,2)*_r(n2+n5+n6,3)*_r(n1+n7+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n7+n8,3)
	      - k4*_r(n2+n4,2)*_r(n3+n5+n6,3)*_r(n1+n7+n8,3)
	      - k12*_r(n4,1)*_r(n2+n3+n5+n6,4)*_r(n1+n7+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n7+n8,3)
	      - k4*_r(n2+n3,2)*_r(n4+n5+n6,3)*_r(n1+n7+n8,3)
	      - k12*_r(n3,1)*_r(n2+n4+n5+n6,4)*_r(n1+n7+n8,3)
	      - k12*_r(n2,1)*_r(n3+n4+n5+n6,4)*_r(n1+n7+n8,3)
	      + k48*_r(n2+n3+n4+n5+n6,5)*_r(n1+n7+n8,3)
	      + k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k4*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k2*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k2*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k4*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k4*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n7+n8,3)
	      + k4*_r(n1,1)*_r(n3+n4+n5,3)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k12*_r(n1+n3+n4+n5,4)*_r(n6,1)*_r(n2+n7+n8,3)
	      - k2*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n7+n8,3)
	      - k4*_r(n3+n4+n5,3)*_r(n1+n6,2)*_r(n2+n7+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n3+n6,2)*_r(n2+n7+n8,3)
	      - k4*_r(n1+n4+n5,3)*_r(n3+n6,2)*_r(n2+n7+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n7+n8,3)
	      - k4*_r(n4+n5,2)*_r(n1+n3+n6,3)*_r(n2+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n4+n6,2)*_r(n2+n7+n8,3)
	      - k4*_r(n1+n3+n5,3)*_r(n4+n6,2)*_r(n2+n7+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n7+n8,3)
	      - k4*_r(n3+n5,2)*_r(n1+n4+n6,3)*_r(n2+n7+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n3+n4+n6,3)*_r(n2+n7+n8,3)
	      - k4*_r(n1+n5,2)*_r(n3+n4+n6,3)*_r(n2+n7+n8,3)
	      - k12*_r(n5,1)*_r(n1+n3+n4+n6,4)*_r(n2+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n7+n8,3)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n5+n6,2)*_r(n2+n7+n8,3)
	      - k4*_r(n1+n3+n4,3)*_r(n5+n6,2)*_r(n2+n7+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n7+n8,3)
	      - k4*_r(n3+n4,2)*_r(n1+n5+n6,3)*_r(n2+n7+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n3+n5+n6,3)*_r(n2+n7+n8,3)
	      - k4*_r(n1+n4,2)*_r(n3+n5+n6,3)*_r(n2+n7+n8,3)
	      - k12*_r(n4,1)*_r(n1+n3+n5+n6,4)*_r(n2+n7+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n4+n5+n6,3)*_r(n2+n7+n8,3)
	      - k4*_r(n1+n3,2)*_r(n4+n5+n6,3)*_r(n2+n7+n8,3)
	      - k12*_r(n3,1)*_r(n1+n4+n5+n6,4)*_r(n2+n7+n8,3)
	      - k12*_r(n1,1)*_r(n3+n4+n5+n6,4)*_r(n2+n7+n8,3)
	      + k48*_r(n1+n3+n4+n5+n6,5)*_r(n2+n7+n8,3)
	      - k6*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7+n8,4)
	      + k6*_r(n3+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n7+n8,4)
	      + k6*_r(n4,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n7+n8,4)
	      + k6*_r(n3,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n7+n8,4)
	      - k12*_r(n3+n4+n5,3)*_r(n6,1)*_r(n1+n2+n7+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n7+n8,4)
	      - k6*_r(n4+n5,2)*_r(n3+n6,2)*_r(n1+n2+n7+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n7+n8,4)
	      - k6*_r(n3+n5,2)*_r(n4+n6,2)*_r(n1+n2+n7+n8,4)
	      - k12*_r(n5,1)*_r(n3+n4+n6,3)*_r(n1+n2+n7+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n7+n8,4)
	      - k6*_r(n3+n4,2)*_r(n5+n6,2)*_r(n1+n2+n7+n8,4)
	      - k12*_r(n4,1)*_r(n3+n5+n6,3)*_r(n1+n2+n7+n8,4)
	      - k12*_r(n3,1)*_r(n4+n5+n6,3)*_r(n1+n2+n7+n8,4)
	      + k36*_r(n3+n4+n5+n6,4)*_r(n1+n2+n7+n8,4)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k4*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k2*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k2*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k4*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4+n5,2)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k4*_r(n2,1)*_r(n1+n4+n5,3)*_r(n6,1)*_r(n3+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2+n4+n5,3)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k12*_r(n1+n2+n4+n5,4)*_r(n6,1)*_r(n3+n7+n8,3)
	      - k2*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n2,1)*_r(n4+n5,2)*_r(n1+n6,2)*_r(n3+n7+n8,3)
	      - k4*_r(n2+n4+n5,3)*_r(n1+n6,2)*_r(n3+n7+n8,3)
	      - k2*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n1,1)*_r(n4+n5,2)*_r(n2+n6,2)*_r(n3+n7+n8,3)
	      - k4*_r(n1+n4+n5,3)*_r(n2+n6,2)*_r(n3+n7+n8,3)
	      + k4*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n7+n8,3)
	      - k4*_r(n4+n5,2)*_r(n1+n2+n6,3)*_r(n3+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n4+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n4+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n4+n6,2)*_r(n3+n7+n8,3)
	      - k4*_r(n1+n2+n5,3)*_r(n4+n6,2)*_r(n3+n7+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n1+n4+n6,3)*_r(n3+n7+n8,3)
	      - k4*_r(n2+n5,2)*_r(n1+n4+n6,3)*_r(n3+n7+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n2+n4+n6,3)*_r(n3+n7+n8,3)
	      - k4*_r(n1+n5,2)*_r(n2+n4+n6,3)*_r(n3+n7+n8,3)
	      - k12*_r(n5,1)*_r(n1+n2+n4+n6,4)*_r(n3+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n5+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n5+n6,2)*_r(n3+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n5+n6,2)*_r(n3+n7+n8,3)
	      - k4*_r(n1+n2+n4,3)*_r(n5+n6,2)*_r(n3+n7+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n1+n5+n6,3)*_r(n3+n7+n8,3)
	      - k4*_r(n2+n4,2)*_r(n1+n5+n6,3)*_r(n3+n7+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n2+n5+n6,3)*_r(n3+n7+n8,3)
	      - k4*_r(n1+n4,2)*_r(n2+n5+n6,3)*_r(n3+n7+n8,3)
	      - k12*_r(n4,1)*_r(n1+n2+n5+n6,4)*_r(n3+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n4+n5+n6,3)*_r(n3+n7+n8,3)
	      - k4*_r(n1+n2,2)*_r(n4+n5+n6,3)*_r(n3+n7+n8,3)
	      - k12*_r(n2,1)*_r(n1+n4+n5+n6,4)*_r(n3+n7+n8,3)
	      - k12*_r(n1,1)*_r(n2+n4+n5+n6,4)*_r(n3+n7+n8,3)
	      + k48*_r(n1+n2+n4+n5+n6,5)*_r(n3+n7+n8,3)
	      - k6*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7+n8,4)
	      + k6*_r(n2+n4,2)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n7+n8,4)
	      + k6*_r(n4,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n7+n8,4)
	      + k6*_r(n2,1)*_r(n4+n5,2)*_r(n6,1)*_r(n1+n3+n7+n8,4)
	      - k12*_r(n2+n4+n5,3)*_r(n6,1)*_r(n1+n3+n7+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n7+n8,4)
	      - k6*_r(n4+n5,2)*_r(n2+n6,2)*_r(n1+n3+n7+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n4+n6,2)*_r(n1+n3+n7+n8,4)
	      - k6*_r(n2+n5,2)*_r(n4+n6,2)*_r(n1+n3+n7+n8,4)
	      - k12*_r(n5,1)*_r(n2+n4+n6,3)*_r(n1+n3+n7+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n5+n6,2)*_r(n1+n3+n7+n8,4)
	      - k6*_r(n2+n4,2)*_r(n5+n6,2)*_r(n1+n3+n7+n8,4)
	      - k12*_r(n4,1)*_r(n2+n5+n6,3)*_r(n1+n3+n7+n8,4)
	      - k12*_r(n2,1)*_r(n4+n5+n6,3)*_r(n1+n3+n7+n8,4)
	      + k36*_r(n2+n4+n5+n6,4)*_r(n1+n3+n7+n8,4)
	      - k6*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7+n8,4)
	      + k6*_r(n1+n4,2)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n7+n8,4)
	      + k6*_r(n4,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n7+n8,4)
	      + k6*_r(n1,1)*_r(n4+n5,2)*_r(n6,1)*_r(n2+n3+n7+n8,4)
	      - k12*_r(n1+n4+n5,3)*_r(n6,1)*_r(n2+n3+n7+n8,4)
	      + k6*_r(n4,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n7+n8,4)
	      - k6*_r(n4+n5,2)*_r(n1+n6,2)*_r(n2+n3+n7+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n4+n6,2)*_r(n2+n3+n7+n8,4)
	      - k6*_r(n1+n5,2)*_r(n4+n6,2)*_r(n2+n3+n7+n8,4)
	      - k12*_r(n5,1)*_r(n1+n4+n6,3)*_r(n2+n3+n7+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n5+n6,2)*_r(n2+n3+n7+n8,4)
	      - k6*_r(n1+n4,2)*_r(n5+n6,2)*_r(n2+n3+n7+n8,4)
	      - k12*_r(n4,1)*_r(n1+n5+n6,3)*_r(n2+n3+n7+n8,4)
	      - k12*_r(n1,1)*_r(n4+n5+n6,3)*_r(n2+n3+n7+n8,4)
	      + k36*_r(n1+n4+n5+n6,4)*_r(n2+n3+n7+n8,4)
	      + k24*_r(n4,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n7+n8,5)
	      - k24*_r(n4+n5,2)*_r(n6,1)*_r(n1+n2+n3+n7+n8,5)
	      - k24*_r(n5,1)*_r(n4+n6,2)*_r(n1+n2+n3+n7+n8,5)
	      - k24*_r(n4,1)*_r(n5+n6,2)*_r(n1+n2+n3+n7+n8,5)
	      + k48*_r(n4+n5+n6,3)*_r(n1+n2+n3+n7+n8,5)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k4*_r(n1+n2+n3,3)*_r(n5,1)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n1+n5,2)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n2+n5,2)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k4*_r(n3,1)*_r(n1+n2+n5,3)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3+n5,2)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k4*_r(n2,1)*_r(n1+n3+n5,3)*_r(n6,1)*_r(n4+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2+n3+n5,3)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k12*_r(n1+n2+n3+n5,4)*_r(n6,1)*_r(n4+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n5,1)*_r(n1+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n3,1)*_r(n2+n5,2)*_r(n1+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n2,1)*_r(n3+n5,2)*_r(n1+n6,2)*_r(n4+n7+n8,3)
	      - k4*_r(n2+n3+n5,3)*_r(n1+n6,2)*_r(n4+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n5,1)*_r(n2+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n3,1)*_r(n1+n5,2)*_r(n2+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n1,1)*_r(n3+n5,2)*_r(n2+n6,2)*_r(n4+n7+n8,3)
	      - k4*_r(n1+n3+n5,3)*_r(n2+n6,2)*_r(n4+n7+n8,3)
	      + k4*_r(n3,1)*_r(n5,1)*_r(n1+n2+n6,3)*_r(n4+n7+n8,3)
	      - k4*_r(n3+n5,2)*_r(n1+n2+n6,3)*_r(n4+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n5,1)*_r(n3+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n5,2)*_r(n3+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n5,2)*_r(n3+n6,2)*_r(n4+n7+n8,3)
	      - k4*_r(n1+n2+n5,3)*_r(n3+n6,2)*_r(n4+n7+n8,3)
	      + k4*_r(n2,1)*_r(n5,1)*_r(n1+n3+n6,3)*_r(n4+n7+n8,3)
	      - k4*_r(n2+n5,2)*_r(n1+n3+n6,3)*_r(n4+n7+n8,3)
	      + k4*_r(n1,1)*_r(n5,1)*_r(n2+n3+n6,3)*_r(n4+n7+n8,3)
	      - k4*_r(n1+n5,2)*_r(n2+n3+n6,3)*_r(n4+n7+n8,3)
	      - k12*_r(n5,1)*_r(n1+n2+n3+n6,4)*_r(n4+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n5+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n5+n6,2)*_r(n4+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n5+n6,2)*_r(n4+n7+n8,3)
	      - k4*_r(n1+n2+n3,3)*_r(n5+n6,2)*_r(n4+n7+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n1+n5+n6,3)*_r(n4+n7+n8,3)
	      - k4*_r(n2+n3,2)*_r(n1+n5+n6,3)*_r(n4+n7+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n2+n5+n6,3)*_r(n4+n7+n8,3)
	      - k4*_r(n1+n3,2)*_r(n2+n5+n6,3)*_r(n4+n7+n8,3)
	      - k12*_r(n3,1)*_r(n1+n2+n5+n6,4)*_r(n4+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n3+n5+n6,3)*_r(n4+n7+n8,3)
	      - k4*_r(n1+n2,2)*_r(n3+n5+n6,3)*_r(n4+n7+n8,3)
	      - k12*_r(n2,1)*_r(n1+n3+n5+n6,4)*_r(n4+n7+n8,3)
	      - k12*_r(n1,1)*_r(n2+n3+n5+n6,4)*_r(n4+n7+n8,3)
	      + k48*_r(n1+n2+n3+n5+n6,5)*_r(n4+n7+n8,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7+n8,4)
	      + k6*_r(n2+n3,2)*_r(n5,1)*_r(n6,1)*_r(n1+n4+n7+n8,4)
	      + k6*_r(n3,1)*_r(n2+n5,2)*_r(n6,1)*_r(n1+n4+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3+n5,2)*_r(n6,1)*_r(n1+n4+n7+n8,4)
	      - k12*_r(n2+n3+n5,3)*_r(n6,1)*_r(n1+n4+n7+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n2+n6,2)*_r(n1+n4+n7+n8,4)
	      - k6*_r(n3+n5,2)*_r(n2+n6,2)*_r(n1+n4+n7+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n3+n6,2)*_r(n1+n4+n7+n8,4)
	      - k6*_r(n2+n5,2)*_r(n3+n6,2)*_r(n1+n4+n7+n8,4)
	      - k12*_r(n5,1)*_r(n2+n3+n6,3)*_r(n1+n4+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n5+n6,2)*_r(n1+n4+n7+n8,4)
	      - k6*_r(n2+n3,2)*_r(n5+n6,2)*_r(n1+n4+n7+n8,4)
	      - k12*_r(n3,1)*_r(n2+n5+n6,3)*_r(n1+n4+n7+n8,4)
	      - k12*_r(n2,1)*_r(n3+n5+n6,3)*_r(n1+n4+n7+n8,4)
	      + k36*_r(n2+n3+n5+n6,4)*_r(n1+n4+n7+n8,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7+n8,4)
	      + k6*_r(n1+n3,2)*_r(n5,1)*_r(n6,1)*_r(n2+n4+n7+n8,4)
	      + k6*_r(n3,1)*_r(n1+n5,2)*_r(n6,1)*_r(n2+n4+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3+n5,2)*_r(n6,1)*_r(n2+n4+n7+n8,4)
	      - k12*_r(n1+n3+n5,3)*_r(n6,1)*_r(n2+n4+n7+n8,4)
	      + k6*_r(n3,1)*_r(n5,1)*_r(n1+n6,2)*_r(n2+n4+n7+n8,4)
	      - k6*_r(n3+n5,2)*_r(n1+n6,2)*_r(n2+n4+n7+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n3+n6,2)*_r(n2+n4+n7+n8,4)
	      - k6*_r(n1+n5,2)*_r(n3+n6,2)*_r(n2+n4+n7+n8,4)
	      - k12*_r(n5,1)*_r(n1+n3+n6,3)*_r(n2+n4+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n5+n6,2)*_r(n2+n4+n7+n8,4)
	      - k6*_r(n1+n3,2)*_r(n5+n6,2)*_r(n2+n4+n7+n8,4)
	      - k12*_r(n3,1)*_r(n1+n5+n6,3)*_r(n2+n4+n7+n8,4)
	      - k12*_r(n1,1)*_r(n3+n5+n6,3)*_r(n2+n4+n7+n8,4)
	      + k36*_r(n1+n3+n5+n6,4)*_r(n2+n4+n7+n8,4)
	      + k24*_r(n3,1)*_r(n5,1)*_r(n6,1)*_r(n1+n2+n4+n7+n8,5)
	      - k24*_r(n3+n5,2)*_r(n6,1)*_r(n1+n2+n4+n7+n8,5)
	      - k24*_r(n5,1)*_r(n3+n6,2)*_r(n1+n2+n4+n7+n8,5)
	      - k24*_r(n3,1)*_r(n5+n6,2)*_r(n1+n2+n4+n7+n8,5)
	      + k48*_r(n3+n5+n6,3)*_r(n1+n2+n4+n7+n8,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7+n8,4)
	      + k6*_r(n1+n2,2)*_r(n5,1)*_r(n6,1)*_r(n3+n4+n7+n8,4)
	      + k6*_r(n2,1)*_r(n1+n5,2)*_r(n6,1)*_r(n3+n4+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2+n5,2)*_r(n6,1)*_r(n3+n4+n7+n8,4)
	      - k12*_r(n1+n2+n5,3)*_r(n6,1)*_r(n3+n4+n7+n8,4)
	      + k6*_r(n2,1)*_r(n5,1)*_r(n1+n6,2)*_r(n3+n4+n7+n8,4)
	      - k6*_r(n2+n5,2)*_r(n1+n6,2)*_r(n3+n4+n7+n8,4)
	      + k6*_r(n1,1)*_r(n5,1)*_r(n2+n6,2)*_r(n3+n4+n7+n8,4)
	      - k6*_r(n1+n5,2)*_r(n2+n6,2)*_r(n3+n4+n7+n8,4)
	      - k12*_r(n5,1)*_r(n1+n2+n6,3)*_r(n3+n4+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n5+n6,2)*_r(n3+n4+n7+n8,4)
	      - k6*_r(n1+n2,2)*_r(n5+n6,2)*_r(n3+n4+n7+n8,4)
	      - k12*_r(n2,1)*_r(n1+n5+n6,3)*_r(n3+n4+n7+n8,4)
	      - k12*_r(n1,1)*_r(n2+n5+n6,3)*_r(n3+n4+n7+n8,4)
	      + k36*_r(n1+n2+n5+n6,4)*_r(n3+n4+n7+n8,4)
	      + k24*_r(n2,1)*_r(n5,1)*_r(n6,1)*_r(n1+n3+n4+n7+n8,5)
	      - k24*_r(n2+n5,2)*_r(n6,1)*_r(n1+n3+n4+n7+n8,5)
	      - k24*_r(n5,1)*_r(n2+n6,2)*_r(n1+n3+n4+n7+n8,5)
	      - k24*_r(n2,1)*_r(n5+n6,2)*_r(n1+n3+n4+n7+n8,5)
	      + k48*_r(n2+n5+n6,3)*_r(n1+n3+n4+n7+n8,5)
	      + k24*_r(n1,1)*_r(n5,1)*_r(n6,1)*_r(n2+n3+n4+n7+n8,5)
	      - k24*_r(n1+n5,2)*_r(n6,1)*_r(n2+n3+n4+n7+n8,5)
	      - k24*_r(n5,1)*_r(n1+n6,2)*_r(n2+n3+n4+n7+n8,5)
	      - k24*_r(n1,1)*_r(n5+n6,2)*_r(n2+n3+n4+n7+n8,5)
	      + k48*_r(n1+n5+n6,3)*_r(n2+n3+n4+n7+n8,5)
	      - k120*_r(n5,1)*_r(n6,1)*_r(n1+n2+n3+n4+n7+n8,6)
	      + k120*_r(n5+n6,2)*_r(n1+n2+n3+n4+n7+n8,6)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k4*_r(n1+n2+n3,3)*_r(n4,1)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n1+n4,2)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n2+n4,2)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k4*_r(n3,1)*_r(n1+n2+n4,3)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3+n4,2)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k4*_r(n2,1)*_r(n1+n3+n4,3)*_r(n6,1)*_r(n5+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2+n3+n4,3)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k12*_r(n1+n2+n3+n4,4)*_r(n6,1)*_r(n5+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n6,2)*_r(n5+n7+n8,3)
	      - k4*_r(n2+n3+n4,3)*_r(n1+n6,2)*_r(n5+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n6,2)*_r(n5+n7+n8,3)
	      - k4*_r(n1+n3+n4,3)*_r(n2+n6,2)*_r(n5+n7+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n1+n2+n6,3)*_r(n5+n7+n8,3)
	      - k4*_r(n3+n4,2)*_r(n1+n2+n6,3)*_r(n5+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n6,2)*_r(n5+n7+n8,3)
	      - k4*_r(n1+n2+n4,3)*_r(n3+n6,2)*_r(n5+n7+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n1+n3+n6,3)*_r(n5+n7+n8,3)
	      - k4*_r(n2+n4,2)*_r(n1+n3+n6,3)*_r(n5+n7+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n2+n3+n6,3)*_r(n5+n7+n8,3)
	      - k4*_r(n1+n4,2)*_r(n2+n3+n6,3)*_r(n5+n7+n8,3)
	      - k12*_r(n4,1)*_r(n1+n2+n3+n6,4)*_r(n5+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n6,2)*_r(n5+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n6,2)*_r(n5+n7+n8,3)
	      - k4*_r(n1+n2+n3,3)*_r(n4+n6,2)*_r(n5+n7+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n1+n4+n6,3)*_r(n5+n7+n8,3)
	      - k4*_r(n2+n3,2)*_r(n1+n4+n6,3)*_r(n5+n7+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n2+n4+n6,3)*_r(n5+n7+n8,3)
	      - k4*_r(n1+n3,2)*_r(n2+n4+n6,3)*_r(n5+n7+n8,3)
	      - k12*_r(n3,1)*_r(n1+n2+n4+n6,4)*_r(n5+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n3+n4+n6,3)*_r(n5+n7+n8,3)
	      - k4*_r(n1+n2,2)*_r(n3+n4+n6,3)*_r(n5+n7+n8,3)
	      - k12*_r(n2,1)*_r(n1+n3+n4+n6,4)*_r(n5+n7+n8,3)
	      - k12*_r(n1,1)*_r(n2+n3+n4+n6,4)*_r(n5+n7+n8,3)
	      + k48*_r(n1+n2+n3+n4+n6,5)*_r(n5+n7+n8,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7+n8,4)
	      + k6*_r(n2+n3,2)*_r(n4,1)*_r(n6,1)*_r(n1+n5+n7+n8,4)
	      + k6*_r(n3,1)*_r(n2+n4,2)*_r(n6,1)*_r(n1+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3+n4,2)*_r(n6,1)*_r(n1+n5+n7+n8,4)
	      - k12*_r(n2+n3+n4,3)*_r(n6,1)*_r(n1+n5+n7+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n2+n6,2)*_r(n1+n5+n7+n8,4)
	      - k6*_r(n3+n4,2)*_r(n2+n6,2)*_r(n1+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n3+n6,2)*_r(n1+n5+n7+n8,4)
	      - k6*_r(n2+n4,2)*_r(n3+n6,2)*_r(n1+n5+n7+n8,4)
	      - k12*_r(n4,1)*_r(n2+n3+n6,3)*_r(n1+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n4+n6,2)*_r(n1+n5+n7+n8,4)
	      - k6*_r(n2+n3,2)*_r(n4+n6,2)*_r(n1+n5+n7+n8,4)
	      - k12*_r(n3,1)*_r(n2+n4+n6,3)*_r(n1+n5+n7+n8,4)
	      - k12*_r(n2,1)*_r(n3+n4+n6,3)*_r(n1+n5+n7+n8,4)
	      + k36*_r(n2+n3+n4+n6,4)*_r(n1+n5+n7+n8,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7+n8,4)
	      + k6*_r(n1+n3,2)*_r(n4,1)*_r(n6,1)*_r(n2+n5+n7+n8,4)
	      + k6*_r(n3,1)*_r(n1+n4,2)*_r(n6,1)*_r(n2+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3+n4,2)*_r(n6,1)*_r(n2+n5+n7+n8,4)
	      - k12*_r(n1+n3+n4,3)*_r(n6,1)*_r(n2+n5+n7+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n6,2)*_r(n2+n5+n7+n8,4)
	      - k6*_r(n3+n4,2)*_r(n1+n6,2)*_r(n2+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n3+n6,2)*_r(n2+n5+n7+n8,4)
	      - k6*_r(n1+n4,2)*_r(n3+n6,2)*_r(n2+n5+n7+n8,4)
	      - k12*_r(n4,1)*_r(n1+n3+n6,3)*_r(n2+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n4+n6,2)*_r(n2+n5+n7+n8,4)
	      - k6*_r(n1+n3,2)*_r(n4+n6,2)*_r(n2+n5+n7+n8,4)
	      - k12*_r(n3,1)*_r(n1+n4+n6,3)*_r(n2+n5+n7+n8,4)
	      - k12*_r(n1,1)*_r(n3+n4+n6,3)*_r(n2+n5+n7+n8,4)
	      + k36*_r(n1+n3+n4+n6,4)*_r(n2+n5+n7+n8,4)
	      + k24*_r(n3,1)*_r(n4,1)*_r(n6,1)*_r(n1+n2+n5+n7+n8,5)
	      - k24*_r(n3+n4,2)*_r(n6,1)*_r(n1+n2+n5+n7+n8,5)
	      - k24*_r(n4,1)*_r(n3+n6,2)*_r(n1+n2+n5+n7+n8,5)
	      - k24*_r(n3,1)*_r(n4+n6,2)*_r(n1+n2+n5+n7+n8,5)
	      + k48*_r(n3+n4+n6,3)*_r(n1+n2+n5+n7+n8,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7+n8,4)
	      + k6*_r(n1+n2,2)*_r(n4,1)*_r(n6,1)*_r(n3+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n1+n4,2)*_r(n6,1)*_r(n3+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2+n4,2)*_r(n6,1)*_r(n3+n5+n7+n8,4)
	      - k12*_r(n1+n2+n4,3)*_r(n6,1)*_r(n3+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n6,2)*_r(n3+n5+n7+n8,4)
	      - k6*_r(n2+n4,2)*_r(n1+n6,2)*_r(n3+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n6,2)*_r(n3+n5+n7+n8,4)
	      - k6*_r(n1+n4,2)*_r(n2+n6,2)*_r(n3+n5+n7+n8,4)
	      - k12*_r(n4,1)*_r(n1+n2+n6,3)*_r(n3+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n4+n6,2)*_r(n3+n5+n7+n8,4)
	      - k6*_r(n1+n2,2)*_r(n4+n6,2)*_r(n3+n5+n7+n8,4)
	      - k12*_r(n2,1)*_r(n1+n4+n6,3)*_r(n3+n5+n7+n8,4)
	      - k12*_r(n1,1)*_r(n2+n4+n6,3)*_r(n3+n5+n7+n8,4)
	      + k36*_r(n1+n2+n4+n6,4)*_r(n3+n5+n7+n8,4)
	      + k24*_r(n2,1)*_r(n4,1)*_r(n6,1)*_r(n1+n3+n5+n7+n8,5)
	      - k24*_r(n2+n4,2)*_r(n6,1)*_r(n1+n3+n5+n7+n8,5)
	      - k24*_r(n4,1)*_r(n2+n6,2)*_r(n1+n3+n5+n7+n8,5)
	      - k24*_r(n2,1)*_r(n4+n6,2)*_r(n1+n3+n5+n7+n8,5)
	      + k48*_r(n2+n4+n6,3)*_r(n1+n3+n5+n7+n8,5)
	      + k24*_r(n1,1)*_r(n4,1)*_r(n6,1)*_r(n2+n3+n5+n7+n8,5)
	      - k24*_r(n1+n4,2)*_r(n6,1)*_r(n2+n3+n5+n7+n8,5)
	      - k24*_r(n4,1)*_r(n1+n6,2)*_r(n2+n3+n5+n7+n8,5)
	      - k24*_r(n1,1)*_r(n4+n6,2)*_r(n2+n3+n5+n7+n8,5)
	      + k48*_r(n1+n4+n6,3)*_r(n2+n3+n5+n7+n8,5)
	      - k120*_r(n4,1)*_r(n6,1)*_r(n1+n2+n3+n5+n7+n8,6)
	      + k120*_r(n4+n6,2)*_r(n1+n2+n3+n5+n7+n8,6)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7+n8,4)
	      + k6*_r(n1+n2,2)*_r(n3,1)*_r(n6,1)*_r(n4+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n1+n3,2)*_r(n6,1)*_r(n4+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2+n3,2)*_r(n6,1)*_r(n4+n5+n7+n8,4)
	      - k12*_r(n1+n2+n3,3)*_r(n6,1)*_r(n4+n5+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n6,2)*_r(n4+n5+n7+n8,4)
	      - k6*_r(n2+n3,2)*_r(n1+n6,2)*_r(n4+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n6,2)*_r(n4+n5+n7+n8,4)
	      - k6*_r(n1+n3,2)*_r(n2+n6,2)*_r(n4+n5+n7+n8,4)
	      - k12*_r(n3,1)*_r(n1+n2+n6,3)*_r(n4+n5+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n6,2)*_r(n4+n5+n7+n8,4)
	      - k6*_r(n1+n2,2)*_r(n3+n6,2)*_r(n4+n5+n7+n8,4)
	      - k12*_r(n2,1)*_r(n1+n3+n6,3)*_r(n4+n5+n7+n8,4)
	      - k12*_r(n1,1)*_r(n2+n3+n6,3)*_r(n4+n5+n7+n8,4)
	      + k36*_r(n1+n2+n3+n6,4)*_r(n4+n5+n7+n8,4)
	      + k24*_r(n2,1)*_r(n3,1)*_r(n6,1)*_r(n1+n4+n5+n7+n8,5)
	      - k24*_r(n2+n3,2)*_r(n6,1)*_r(n1+n4+n5+n7+n8,5)
	      - k24*_r(n3,1)*_r(n2+n6,2)*_r(n1+n4+n5+n7+n8,5)
	      - k24*_r(n2,1)*_r(n3+n6,2)*_r(n1+n4+n5+n7+n8,5)
	      + k48*_r(n2+n3+n6,3)*_r(n1+n4+n5+n7+n8,5)
	      + k24*_r(n1,1)*_r(n3,1)*_r(n6,1)*_r(n2+n4+n5+n7+n8,5)
	      - k24*_r(n1+n3,2)*_r(n6,1)*_r(n2+n4+n5+n7+n8,5)
	      - k24*_r(n3,1)*_r(n1+n6,2)*_r(n2+n4+n5+n7+n8,5)
	      - k24*_r(n1,1)*_r(n3+n6,2)*_r(n2+n4+n5+n7+n8,5)
	      + k48*_r(n1+n3+n6,3)*_r(n2+n4+n5+n7+n8,5)
	      - k120*_r(n3,1)*_r(n6,1)*_r(n1+n2+n4+n5+n7+n8,6)
	      + k120*_r(n3+n6,2)*_r(n1+n2+n4+n5+n7+n8,6)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n6,1)*_r(n3+n4+n5+n7+n8,5)
	      - k24*_r(n1+n2,2)*_r(n6,1)*_r(n3+n4+n5+n7+n8,5)
	      - k24*_r(n2,1)*_r(n1+n6,2)*_r(n3+n4+n5+n7+n8,5)
	      - k24*_r(n1,1)*_r(n2+n6,2)*_r(n3+n4+n5+n7+n8,5)
	      + k48*_r(n1+n2+n6,3)*_r(n3+n4+n5+n7+n8,5)
	      - k120*_r(n2,1)*_r(n6,1)*_r(n1+n3+n4+n5+n7+n8,6)
	      + k120*_r(n2+n6,2)*_r(n1+n3+n4+n5+n7+n8,6)
	      - k120*_r(n1,1)*_r(n6,1)*_r(n2+n3+n4+n5+n7+n8,6)
	      + k120*_r(n1+n6,2)*_r(n2+n3+n4+n5+n7+n8,6)
	      + k720*_r(n6,1)*_r(n1+n2+n3+n4+n5+n7+n8,7)
	      + k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k4*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n1+n4,2)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n2+n4,2)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k4*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3+n4,2)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k4*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5,1)*_r(n6+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k12*_r(n1+n2+n3+n4,4)*_r(n5,1)*_r(n6+n7+n8,3)
	      - k2*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5,2)*_r(n6+n7+n8,3)
	      - k4*_r(n2+n3+n4,3)*_r(n1+n5,2)*_r(n6+n7+n8,3)
	      - k2*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5,2)*_r(n6+n7+n8,3)
	      - k4*_r(n1+n3+n4,3)*_r(n2+n5,2)*_r(n6+n7+n8,3)
	      + k4*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5,3)*_r(n6+n7+n8,3)
	      - k4*_r(n3+n4,2)*_r(n1+n2+n5,3)*_r(n6+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5,2)*_r(n6+n7+n8,3)
	      - k4*_r(n1+n2+n4,3)*_r(n3+n5,2)*_r(n6+n7+n8,3)
	      + k4*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5,3)*_r(n6+n7+n8,3)
	      - k4*_r(n2+n4,2)*_r(n1+n3+n5,3)*_r(n6+n7+n8,3)
	      + k4*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5,3)*_r(n6+n7+n8,3)
	      - k4*_r(n1+n4,2)*_r(n2+n3+n5,3)*_r(n6+n7+n8,3)
	      - k12*_r(n4,1)*_r(n1+n2+n3+n5,4)*_r(n6+n7+n8,3)
	      - k2*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5,2)*_r(n6+n7+n8,3)
	      + k2*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5,2)*_r(n6+n7+n8,3)
	      - k4*_r(n1+n2+n3,3)*_r(n4+n5,2)*_r(n6+n7+n8,3)
	      + k4*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5,3)*_r(n6+n7+n8,3)
	      - k4*_r(n2+n3,2)*_r(n1+n4+n5,3)*_r(n6+n7+n8,3)
	      + k4*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5,3)*_r(n6+n7+n8,3)
	      - k4*_r(n1+n3,2)*_r(n2+n4+n5,3)*_r(n6+n7+n8,3)
	      - k12*_r(n3,1)*_r(n1+n2+n4+n5,4)*_r(n6+n7+n8,3)
	      + k4*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5,3)*_r(n6+n7+n8,3)
	      - k4*_r(n1+n2,2)*_r(n3+n4+n5,3)*_r(n6+n7+n8,3)
	      - k12*_r(n2,1)*_r(n1+n3+n4+n5,4)*_r(n6+n7+n8,3)
	      - k12*_r(n1,1)*_r(n2+n3+n4+n5,4)*_r(n6+n7+n8,3)
	      + k48*_r(n1+n2+n3+n4+n5,5)*_r(n6+n7+n8,3)
	      - k6*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7+n8,4)
	      + k6*_r(n2+n3,2)*_r(n4,1)*_r(n5,1)*_r(n1+n6+n7+n8,4)
	      + k6*_r(n3,1)*_r(n2+n4,2)*_r(n5,1)*_r(n1+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3+n4,2)*_r(n5,1)*_r(n1+n6+n7+n8,4)
	      - k12*_r(n2+n3+n4,3)*_r(n5,1)*_r(n1+n6+n7+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n2+n5,2)*_r(n1+n6+n7+n8,4)
	      - k6*_r(n3+n4,2)*_r(n2+n5,2)*_r(n1+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n3+n5,2)*_r(n1+n6+n7+n8,4)
	      - k6*_r(n2+n4,2)*_r(n3+n5,2)*_r(n1+n6+n7+n8,4)
	      - k12*_r(n4,1)*_r(n2+n3+n5,3)*_r(n1+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n4+n5,2)*_r(n1+n6+n7+n8,4)
	      - k6*_r(n2+n3,2)*_r(n4+n5,2)*_r(n1+n6+n7+n8,4)
	      - k12*_r(n3,1)*_r(n2+n4+n5,3)*_r(n1+n6+n7+n8,4)
	      - k12*_r(n2,1)*_r(n3+n4+n5,3)*_r(n1+n6+n7+n8,4)
	      + k36*_r(n2+n3+n4+n5,4)*_r(n1+n6+n7+n8,4)
	      - k6*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7+n8,4)
	      + k6*_r(n1+n3,2)*_r(n4,1)*_r(n5,1)*_r(n2+n6+n7+n8,4)
	      + k6*_r(n3,1)*_r(n1+n4,2)*_r(n5,1)*_r(n2+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3+n4,2)*_r(n5,1)*_r(n2+n6+n7+n8,4)
	      - k12*_r(n1+n3+n4,3)*_r(n5,1)*_r(n2+n6+n7+n8,4)
	      + k6*_r(n3,1)*_r(n4,1)*_r(n1+n5,2)*_r(n2+n6+n7+n8,4)
	      - k6*_r(n3+n4,2)*_r(n1+n5,2)*_r(n2+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n3+n5,2)*_r(n2+n6+n7+n8,4)
	      - k6*_r(n1+n4,2)*_r(n3+n5,2)*_r(n2+n6+n7+n8,4)
	      - k12*_r(n4,1)*_r(n1+n3+n5,3)*_r(n2+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n4+n5,2)*_r(n2+n6+n7+n8,4)
	      - k6*_r(n1+n3,2)*_r(n4+n5,2)*_r(n2+n6+n7+n8,4)
	      - k12*_r(n3,1)*_r(n1+n4+n5,3)*_r(n2+n6+n7+n8,4)
	      - k12*_r(n1,1)*_r(n3+n4+n5,3)*_r(n2+n6+n7+n8,4)
	      + k36*_r(n1+n3+n4+n5,4)*_r(n2+n6+n7+n8,4)
	      + k24*_r(n3,1)*_r(n4,1)*_r(n5,1)*_r(n1+n2+n6+n7+n8,5)
	      - k24*_r(n3+n4,2)*_r(n5,1)*_r(n1+n2+n6+n7+n8,5)
	      - k24*_r(n4,1)*_r(n3+n5,2)*_r(n1+n2+n6+n7+n8,5)
	      - k24*_r(n3,1)*_r(n4+n5,2)*_r(n1+n2+n6+n7+n8,5)
	      + k48*_r(n3+n4+n5,3)*_r(n1+n2+n6+n7+n8,5)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7+n8,4)
	      + k6*_r(n1+n2,2)*_r(n4,1)*_r(n5,1)*_r(n3+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n1+n4,2)*_r(n5,1)*_r(n3+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2+n4,2)*_r(n5,1)*_r(n3+n6+n7+n8,4)
	      - k12*_r(n1+n2+n4,3)*_r(n5,1)*_r(n3+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n4,1)*_r(n1+n5,2)*_r(n3+n6+n7+n8,4)
	      - k6*_r(n2+n4,2)*_r(n1+n5,2)*_r(n3+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n4,1)*_r(n2+n5,2)*_r(n3+n6+n7+n8,4)
	      - k6*_r(n1+n4,2)*_r(n2+n5,2)*_r(n3+n6+n7+n8,4)
	      - k12*_r(n4,1)*_r(n1+n2+n5,3)*_r(n3+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n4+n5,2)*_r(n3+n6+n7+n8,4)
	      - k6*_r(n1+n2,2)*_r(n4+n5,2)*_r(n3+n6+n7+n8,4)
	      - k12*_r(n2,1)*_r(n1+n4+n5,3)*_r(n3+n6+n7+n8,4)
	      - k12*_r(n1,1)*_r(n2+n4+n5,3)*_r(n3+n6+n7+n8,4)
	      + k36*_r(n1+n2+n4+n5,4)*_r(n3+n6+n7+n8,4)
	      + k24*_r(n2,1)*_r(n4,1)*_r(n5,1)*_r(n1+n3+n6+n7+n8,5)
	      - k24*_r(n2+n4,2)*_r(n5,1)*_r(n1+n3+n6+n7+n8,5)
	      - k24*_r(n4,1)*_r(n2+n5,2)*_r(n1+n3+n6+n7+n8,5)
	      - k24*_r(n2,1)*_r(n4+n5,2)*_r(n1+n3+n6+n7+n8,5)
	      + k48*_r(n2+n4+n5,3)*_r(n1+n3+n6+n7+n8,5)
	      + k24*_r(n1,1)*_r(n4,1)*_r(n5,1)*_r(n2+n3+n6+n7+n8,5)
	      - k24*_r(n1+n4,2)*_r(n5,1)*_r(n2+n3+n6+n7+n8,5)
	      - k24*_r(n4,1)*_r(n1+n5,2)*_r(n2+n3+n6+n7+n8,5)
	      - k24*_r(n1,1)*_r(n4+n5,2)*_r(n2+n3+n6+n7+n8,5)
	      + k48*_r(n1+n4+n5,3)*_r(n2+n3+n6+n7+n8,5)
	      - k120*_r(n4,1)*_r(n5,1)*_r(n1+n2+n3+n6+n7+n8,6)
	      + k120*_r(n4+n5,2)*_r(n1+n2+n3+n6+n7+n8,6)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7+n8,4)
	      + k6*_r(n1+n2,2)*_r(n3,1)*_r(n5,1)*_r(n4+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n1+n3,2)*_r(n5,1)*_r(n4+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2+n3,2)*_r(n5,1)*_r(n4+n6+n7+n8,4)
	      - k12*_r(n1+n2+n3,3)*_r(n5,1)*_r(n4+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n5,2)*_r(n4+n6+n7+n8,4)
	      - k6*_r(n2+n3,2)*_r(n1+n5,2)*_r(n4+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n5,2)*_r(n4+n6+n7+n8,4)
	      - k6*_r(n1+n3,2)*_r(n2+n5,2)*_r(n4+n6+n7+n8,4)
	      - k12*_r(n3,1)*_r(n1+n2+n5,3)*_r(n4+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n5,2)*_r(n4+n6+n7+n8,4)
	      - k6*_r(n1+n2,2)*_r(n3+n5,2)*_r(n4+n6+n7+n8,4)
	      - k12*_r(n2,1)*_r(n1+n3+n5,3)*_r(n4+n6+n7+n8,4)
	      - k12*_r(n1,1)*_r(n2+n3+n5,3)*_r(n4+n6+n7+n8,4)
	      + k36*_r(n1+n2+n3+n5,4)*_r(n4+n6+n7+n8,4)
	      + k24*_r(n2,1)*_r(n3,1)*_r(n5,1)*_r(n1+n4+n6+n7+n8,5)
	      - k24*_r(n2+n3,2)*_r(n5,1)*_r(n1+n4+n6+n7+n8,5)
	      - k24*_r(n3,1)*_r(n2+n5,2)*_r(n1+n4+n6+n7+n8,5)
	      - k24*_r(n2,1)*_r(n3+n5,2)*_r(n1+n4+n6+n7+n8,5)
	      + k48*_r(n2+n3+n5,3)*_r(n1+n4+n6+n7+n8,5)
	      + k24*_r(n1,1)*_r(n3,1)*_r(n5,1)*_r(n2+n4+n6+n7+n8,5)
	      - k24*_r(n1+n3,2)*_r(n5,1)*_r(n2+n4+n6+n7+n8,5)
	      - k24*_r(n3,1)*_r(n1+n5,2)*_r(n2+n4+n6+n7+n8,5)
	      - k24*_r(n1,1)*_r(n3+n5,2)*_r(n2+n4+n6+n7+n8,5)
	      + k48*_r(n1+n3+n5,3)*_r(n2+n4+n6+n7+n8,5)
	      - k120*_r(n3,1)*_r(n5,1)*_r(n1+n2+n4+n6+n7+n8,6)
	      + k120*_r(n3+n5,2)*_r(n1+n2+n4+n6+n7+n8,6)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n5,1)*_r(n3+n4+n6+n7+n8,5)
	      - k24*_r(n1+n2,2)*_r(n5,1)*_r(n3+n4+n6+n7+n8,5)
	      - k24*_r(n2,1)*_r(n1+n5,2)*_r(n3+n4+n6+n7+n8,5)
	      - k24*_r(n1,1)*_r(n2+n5,2)*_r(n3+n4+n6+n7+n8,5)
	      + k48*_r(n1+n2+n5,3)*_r(n3+n4+n6+n7+n8,5)
	      - k120*_r(n2,1)*_r(n5,1)*_r(n1+n3+n4+n6+n7+n8,6)
	      + k120*_r(n2+n5,2)*_r(n1+n3+n4+n6+n7+n8,6)
	      - k120*_r(n1,1)*_r(n5,1)*_r(n2+n3+n4+n6+n7+n8,6)
	      + k120*_r(n1+n5,2)*_r(n2+n3+n4+n6+n7+n8,6)
	      + k720*_r(n5,1)*_r(n1+n2+n3+n4+n6+n7+n8,7)
	      - k6*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7+n8,4)
	      + k6*_r(n1+n2,2)*_r(n3,1)*_r(n4,1)*_r(n5+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n1+n3,2)*_r(n4,1)*_r(n5+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2+n3,2)*_r(n4,1)*_r(n5+n6+n7+n8,4)
	      - k12*_r(n1+n2+n3,3)*_r(n4,1)*_r(n5+n6+n7+n8,4)
	      + k6*_r(n2,1)*_r(n3,1)*_r(n1+n4,2)*_r(n5+n6+n7+n8,4)
	      - k6*_r(n2+n3,2)*_r(n1+n4,2)*_r(n5+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n3,1)*_r(n2+n4,2)*_r(n5+n6+n7+n8,4)
	      - k6*_r(n1+n3,2)*_r(n2+n4,2)*_r(n5+n6+n7+n8,4)
	      - k12*_r(n3,1)*_r(n1+n2+n4,3)*_r(n5+n6+n7+n8,4)
	      + k6*_r(n1,1)*_r(n2,1)*_r(n3+n4,2)*_r(n5+n6+n7+n8,4)
	      - k6*_r(n1+n2,2)*_r(n3+n4,2)*_r(n5+n6+n7+n8,4)
	      - k12*_r(n2,1)*_r(n1+n3+n4,3)*_r(n5+n6+n7+n8,4)
	      - k12*_r(n1,1)*_r(n2+n3+n4,3)*_r(n5+n6+n7+n8,4)
	      + k36*_r(n1+n2+n3+n4,4)*_r(n5+n6+n7+n8,4)
	      + k24*_r(n2,1)*_r(n3,1)*_r(n4,1)*_r(n1+n5+n6+n7+n8,5)
	      - k24*_r(n2+n3,2)*_r(n4,1)*_r(n1+n5+n6+n7+n8,5)
	      - k24*_r(n3,1)*_r(n2+n4,2)*_r(n1+n5+n6+n7+n8,5)
	      - k24*_r(n2,1)*_r(n3+n4,2)*_r(n1+n5+n6+n7+n8,5)
	      + k48*_r(n2+n3+n4,3)*_r(n1+n5+n6+n7+n8,5)
	      + k24*_r(n1,1)*_r(n3,1)*_r(n4,1)*_r(n2+n5+n6+n7+n8,5)
	      - k24*_r(n1+n3,2)*_r(n4,1)*_r(n2+n5+n6+n7+n8,5)
	      - k24*_r(n3,1)*_r(n1+n4,2)*_r(n2+n5+n6+n7+n8,5)
	      - k24*_r(n1,1)*_r(n3+n4,2)*_r(n2+n5+n6+n7+n8,5)
	      + k48*_r(n1+n3+n4,3)*_r(n2+n5+n6+n7+n8,5)
	      - k120*_r(n3,1)*_r(n4,1)*_r(n1+n2+n5+n6+n7+n8,6)
	      + k120*_r(n3+n4,2)*_r(n1+n2+n5+n6+n7+n8,6)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n4,1)*_r(n3+n5+n6+n7+n8,5)
	      - k24*_r(n1+n2,2)*_r(n4,1)*_r(n3+n5+n6+n7+n8,5)
	      - k24*_r(n2,1)*_r(n1+n4,2)*_r(n3+n5+n6+n7+n8,5)
	      - k24*_r(n1,1)*_r(n2+n4,2)*_r(n3+n5+n6+n7+n8,5)
	      + k48*_r(n1+n2+n4,3)*_r(n3+n5+n6+n7+n8,5)
	      - k120*_r(n2,1)*_r(n4,1)*_r(n1+n3+n5+n6+n7+n8,6)
	      + k120*_r(n2+n4,2)*_r(n1+n3+n5+n6+n7+n8,6)
	      - k120*_r(n1,1)*_r(n4,1)*_r(n2+n3+n5+n6+n7+n8,6)
	      + k120*_r(n1+n4,2)*_r(n2+n3+n5+n6+n7+n8,6)
	      + k720*_r(n4,1)*_r(n1+n2+n3+n5+n6+n7+n8,7)
	      + k24*_r(n1,1)*_r(n2,1)*_r(n3,1)*_r(n4+n5+n6+n7+n8,5)
	      - k24*_r(n1+n2,2)*_r(n3,1)*_r(n4+n5+n6+n7+n8,5)
	      - k24*_r(n2,1)*_r(n1+n3,2)*_r(n4+n5+n6+n7+n8,5)
	      - k24*_r(n1,1)*_r(n2+n3,2)*_r(n4+n5+n6+n7+n8,5)
	      + k48*_r(n1+n2+n3,3)*_r(n4+n5+n6+n7+n8,5)
	      - k120*_r(n2,1)*_r(n3,1)*_r(n1+n4+n5+n6+n7+n8,6)
	      + k120*_r(n2+n3,2)*_r(n1+n4+n5+n6+n7+n8,6)
	      - k120*_r(n1,1)*_r(n3,1)*_r(n2+n4+n5+n6+n7+n8,6)
	      + k120*_r(n1+n3,2)*_r(n2+n4+n5+n6+n7+n8,6)
	      + k720*_r(n3,1)*_r(n1+n2+n4+n5+n6+n7+n8,7)
	      - k120*_r(n1,1)*_r(n2,1)*_r(n3+n4+n5+n6+n7+n8,6)
	      + k120*_r(n1+n2,2)*_r(n3+n4+n5+n6+n7+n8,6)
	      + k720*_r(n2,1)*_r(n1+n3+n4+n5+n6+n7+n8,7)
	      + k720*_r(n1,1)*_r(n2+n3+n4+n5+n6+n7+n8,7)
	      - k5040*_r(n1+n2+n3+n4+n5+n6+n7+n8,8));
    }
    /**
     * Do the 8-particle calculation
     *
     * @param n1 1st Harmonic
     * @param n2 2nd Harmonic
     * @param n3 3rd Harmonic
     * @param n4 4th Harmonic
     * @param n5 5th Harmonic
     * @param n6 6th Harmonic
     * @param n7 7th Harmonic
     * @param n8 8th Harmonic
     *
     * @return The correlator
     */
    Complex c8(const HarmonicElement n1,
	       const HarmonicElement n2,
	       const HarmonicElement n3,
	       const HarmonicElement n4,
	       const HarmonicElement n5,
	       const HarmonicElement n6,
	       const HarmonicElement n7,
	       const HarmonicElement n8) const
    {
      return (c8P1(n1,n2,n3,n4,n5,n6,n7,n8)
	      + c8P2(n1,n2,n3,n4,n5,n6,n7,n8)
	      + c8P3(n1,n2,n3,n4,n5,n6,n7,n8)
	      + c8P4(n1,n2,n3,n4,n5,n6,n7,n8));
    }
  };
  /** 
   * Specialisation that disable higher order correlators 
   *
   * @{ 
   */
  /** 
   * Disable 8-particle correlator for maximum 7-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<HarmonicVector,7>::c8(const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement) const
  {
    throw std::runtime_error("8-particle correlator disabled");
    return Complex();
  }    
  /** 
   * Disable 8-particle correlator for maximum 6-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<HarmonicVector,6>::c8(const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement) const
  {
    throw std::runtime_error("8-particle correlator disabled");
    return Complex();
  }    
  /** 
   * Disable 7-particle correlator for maximum 6-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<HarmonicVector,6>::c7(const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement) const
  {
    throw std::runtime_error("7-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 8-particle correlator for maximum 5-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<HarmonicVector,5>::c8(const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement) const
  {
    throw std::runtime_error("8-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 7-particle correlator for maximum 5-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<HarmonicVector,5>::c7(const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement) const
  {
    throw std::runtime_error("7-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 6-particle correlator for maximum 5-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<HarmonicVector,5>::c6(const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement,
					const HarmonicElement) const
  {
    throw std::runtime_error("6-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 8-particle correlator for maximum 7-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<PartitionHarmonicVector,7>::c8(const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement) const
  {
    throw std::runtime_error("8-particle correlator disabled");
    return Complex();
  }    
  /** 
   * Disable 8-particle correlator for maximum 6-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<PartitionHarmonicVector,6>::c8(const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement) const
  {
    throw std::runtime_error("8-particle correlator disabled");
    return Complex();
  }    
  /** 
   * Disable 7-particle correlator for maximum 6-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<PartitionHarmonicVector,6>::c7(const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement) const
  {
    throw std::runtime_error("7-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 8-particle correlator for maximum 5-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<PartitionHarmonicVector,5>::c8(const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement) const
  {
    throw std::runtime_error("8-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 7-particle correlator for maximum 5-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<PartitionHarmonicVector,5>::c7(const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement) const
  {
    throw std::runtime_error("7-particle correlator disabled");
    return Complex();
  }
  /** 
   * Disable 6-particle correlator for maximum 5-particle class
   *
   * @throws std::runtime_error Always thrown 
   */
  template <>
  Complex ClosedM<PartitionHarmonicVector,5>::c6(const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement,
						 const HarmonicElement) const
  {
    throw std::runtime_error("6-particle correlator disabled");
    return Complex();
  }
  /* @} */
  /** Default closed form implementation */
  template <typename Harmonics=HarmonicVector> using Closed=ClosedM<Harmonics>;
}
#endif
// Local Variables:
//  mode: C++
// End:
